/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;

/**
 * An MBean attribute which is a name value pair combination.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.4 $
 */
public class Attribute implements Serializable
{
	private static final long serialVersionUID = 2484220110589082382L;

	private String name;
	private Object value;
	private transient int m_hash;


    /**
     * Creates an Attribute with the given name and value.
     *
     * <p>
     * Note: The value of name cant be null.
     * </p>
     *
     * @param name The name of the attribute
     * @param value The value of the attribute
     * @throws RuntimeOperationsException Wraps an IllegalArgumentException if the
     *              value of name is null.
     *
     */
	public Attribute(String name, Object value)
	{
		if (name == null) {throw new RuntimeOperationsException(new IllegalArgumentException("The name of an attribute cannot be null"));}
		this.name = name;
		this.value = value;

		m_hash = computeHash();
	}


    /**
     * Checks the given obj for equality with this <code>Attribute</code>.
     *
     * @param obj The object being checked against.
     * @return true If obj is equal to this <code>Attribute</code>, false otherwise.
     *
     */
	public boolean equals(Object obj)
	{
		if (obj == null) {return false;}
		if (obj == this) {return true;}
		try
		{
			Attribute other = (Attribute)obj;
			boolean namesEqual = name.equals(other.name);
			boolean valuesEqual = false;
			if (value == null) {valuesEqual = other.value == null;}
			else {valuesEqual = value.equals(other.value);}
			return namesEqual && valuesEqual;
		}
		catch (ClassCastException ignored) {}
		return false;
	}


    /**
     * Returns the hashCode
     *
     * @return int The hashCode
     *
     */
	public int hashCode() {return m_hash;}

    /**
     * Return the name of this <code>Attribute</code>.
     *
     * @return String The name of this Attribute
     *  
     */
	public String getName() {return name;}

    /**
     * Returns the value of this <code>Attribute</code>.
     *
     * @return Object The value of this Attribute
     *
     */
	public Object getValue() {return value;}


    /**
     * Returns a String representation of this <code>Attribtue</code>.
     *
     * <p>
     *  Representation is in the form: Attribute's name: n, value: v
     * </p>
     * 
     *
     */
	public String toString()
	{
        return new StringBuffer("Attribute's name: ").append(getName()).append(", value: ").append(getValue()).toString();
	}

	private int computeHash()
	{
		int hash = name.hashCode();
		if (value != null) {hash ^= value.hashCode();}
		return hash;
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
	{
		in.defaultReadObject();
		m_hash = computeHash();
	}
}
