/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectStreamField;
import java.util.HashSet;
import java.util.Vector;

/**
 * Provides a filter for the {@link AttributeChangeNotification}.
 *
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.7 $
 */
public class AttributeChangeNotificationFilter implements NotificationFilter
{
	private static final long serialVersionUID = -6347317584796410029L;
	private static final String serialName = "enabledAttributes";
	private static final ObjectStreamField[] serialPersistentFields =
	{
		new ObjectStreamField(serialName, Vector.class)
	};

    private HashSet m_enabledAttributes = new HashSet();


    /**
     * Returns the hashCode.
     *   
     * @return int The hashCode
     * 
     */
	public int hashCode()
	{
		return m_enabledAttributes.hashCode();
	}

    /**
     * Checks the given object for equality.
     *
     * @return boolean true if equal, false otherwise.
     *
     */
	public boolean equals(Object obj)
	{
		if (obj == null) {return false;}
		if (obj == this) {return true;}

		try
		{
			AttributeChangeNotificationFilter other = (AttributeChangeNotificationFilter)obj;
			return m_enabledAttributes.equals(other.m_enabledAttributes);
		}
		catch (ClassCastException x) {}
		return false;
	}


    /**
     * Enable the Notification for the given attribute name.
     *
     * @param name The name of the attribute to be enabled.
     * @throws RuntimeOperationException Wraps an IllegalArgumentException when 
     *          the value of the given name is null.
     */
	public void enableAttribute(String name)
	{
        if (name == null) {throw new RuntimeOperationsException(new IllegalArgumentException("Name cannot be null"));}
		synchronized (m_enabledAttributes)
		{
			if (!m_enabledAttributes.contains(name))
			{
				m_enabledAttributes.add(name);
			}
		}
	}

    /**
     * Disables the given attribute.
     *
     * <p>
     * Note: This ignores operation when the given name is null.
     * </p>
     *
     * @param name The attribute name to be disabled.
     *
     */
	public void disableAttribute(String name)
	{
		if (name != null)
		{
			synchronized (m_enabledAttributes)
			{
				m_enabledAttributes.remove(name);
			}
		}
	}

    /**
     * Disable all attribute notifications
     *
     */
	public void disableAllAttributes()
	{
		synchronized (m_enabledAttributes)
		{
			m_enabledAttributes.clear();
		}
	}

    /**
     * Returns the list of all enabled attributes for notification.
     *
     * @return Vector A new <code>Vector</code> containing the attributes for notifications
     *
     */
	public Vector getEnabledAttributes()
	{
		synchronized (m_enabledAttributes)
		{
			Vector v = new Vector(m_enabledAttributes);
			return v;
		}
	}


    /**
     * Checks if the given Notification is enabled.
     *
     * <p>
     * The notification needs to be an instance of {@link AttributeChangeNotification}
     * It checks the attribute name and checked it against the enabled attributes
     * contained in this filter.
     * </p>
     *
     * @return boolean if Notification is enabled for the given 
     *          {@link AttributeChangeNotification}, false otherwise.
     *
     *
     */
	public boolean isNotificationEnabled(Notification notification)
	{
		if (!(notification instanceof AttributeChangeNotification)) {return false;}
        AttributeChangeNotification n = (AttributeChangeNotification)notification;
		if (!AttributeChangeNotification.ATTRIBUTE_CHANGE.equals(n.getType())) {return false;}
        String attributeName = n.getAttributeName();
		if (attributeName != null)
		{
			synchronized (m_enabledAttributes)
			{
                if (m_enabledAttributes.contains(attributeName))
				{
					return true;
				}
			}
		}
		return false;
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
	{
		ObjectInputStream.GetField fields = in.readFields();

		Vector vector = (Vector)fields.get(serialName, null);
		if (fields.defaulted(serialName))
		{
            throw new IOException("Serialized stream corrupted: expecting a non-null Vector");
		}

		if (m_enabledAttributes == null) {m_enabledAttributes = new HashSet();}
		m_enabledAttributes.clear();
		m_enabledAttributes.addAll(vector);
	}

    private void writeObject(ObjectOutputStream out) throws IOException
	{
		ObjectOutputStream.PutField fields = out.putFields();

		Vector vector = getEnabledAttributes();
		fields.put(serialName, vector);
		out.writeFields();
	}
}
