/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management;

import java.io.Serializable;

/**
 * Descriptor is cloneable since represent a snapshot of what the client passed in to the model mbean.
 * This ensures that if the client nulls out some value, the descriptor is still valid, since it has been cloned
 * prior the client modification. The client can always re-set the descriptor on the model mbean.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.3 $
 */
public interface Descriptor extends Cloneable, Serializable
{

    /**
     * Returns the field value of a given field name.
     * 
     * @param fieldName The field name where the value being retrieved is for. 
     * @return Object The value for the given field name. Returns null if not
     *                found.
     * @throws RuntimeOperationException if the value for field name is illegal
     *
     */
	public Object getFieldValue(String fieldName) throws RuntimeOperationsException;

    /**
     * Sets a field name and field value. The field value will be checked 
     * before being set. This will either add a new field or update it if it
     * already exists.
     *
     * @param fieldName The name of the field to set againts
     * @param fieldValue The value for the given field name being set
     * @throws RuntimeOperationsException If values for fieldName or fieldValue
     *              are illegal or the description construction fails, then 
     *              this exception will be thrown 
     *
     */
	public void setField(String fieldName, Object fieldValue) throws RuntimeOperationsException;

    
    /**
     * Remove a field using the given field name.
     * <p>  
     * Note that this method will be silent if no such field is found.
     * </p>
     *
     * @param fieldName The field name to be removed.
     *
     */
	public void removeField(String fieldName);

    /**
     * Return all the field names as an array of String.
     * <p>
     * Note that this method will return an empty array if the 
     * <code>Descriptor</code> is empty.
     *
     * @return String[] The array of field names.
     *
     */
	public String[] getFieldNames();
    
    /**
     * Return all field values as an array using the given field names array.
     * <p>
     * Note that the order of the returned Object[] is the same as the 
     * order of the field name argument.
     * </p>
     * 
     * @param fieldNames The array of fieldnames
     * @return Object[] The array of all field values corresponding to the
     *                  given fieldNames array.
     *
     */
	public Object[] getFieldValues(String[] fieldNames);

    /**
     * Returns all the fields contained in this <code>Descriptor</code> as 
     * an array.
     *
     * @return String[] The String array in the format fieldName=fieldValue.An empty descriptor will result with an empty array.
     *
     */
	public String[] getFields();


    /**
     * Sets the given fieldNames against the givent fieldValues. The size
     * of both given array should match.
     *
     * @param The given array of fieldNames to set.
     * @param The given array of fieldValues.
     * @throws RuntimeOperationsException if fieldNames or fieldValues contains
     *          illegal values.
     *
     */
	public void setFields(String[] fieldNames, Object[] fieldValues) throws RuntimeOperationsException;

    /**
     * Returns a new Descriptor of which is a duplicate of this Descriptor
     *
     * @return Object The cloned Descriptor
     */
	public Object clone() throws RuntimeOperationsException;

    /**
     * Returns true if fieldValues are checked to be sure they are legal for the fieldNames.
     *
     * @throws RuntimeOperationsException If the validity checking fails.
     *
     */
	public boolean isValid() throws RuntimeOperationsException;
}
