/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management;

import java.util.ArrayList;

import javax.management.loading.ClassLoaderRepository;

import mx4j.server.MBeanServerImpl;
import mx4j.log.Log;
import mx4j.log.Logger;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.11 $
 */
public class MBeanServerFactory
{
	private static ArrayList m_mbeanServers = new ArrayList();

	private MBeanServerFactory() {}

	private static Logger getLogger()
	{
		return Log.getLogger(MBeanServerFactory.class.getName());
	}

	public static MBeanServer createMBeanServer()
	{
		return createMBeanServer(null);
	}

	public static MBeanServer createMBeanServer(String defaultDomain)
	{
		MBeanServer s = createMBeanServerImpl(defaultDomain, "createMBeanServer");
		synchronized (m_mbeanServers)
		{
			m_mbeanServers.add(s);
		}
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.INFO)) logger.info("MBeanServer " + s + " registered successfully");
		return s;
	}

	public static MBeanServer newMBeanServer()
	{
		return newMBeanServer(null);
	}

	public static MBeanServer newMBeanServer(String defaultDomain)
	{
		return createMBeanServerImpl(defaultDomain, "newMBeanServer");
	}

	public static void releaseMBeanServer(MBeanServer server)
	{
		Logger logger = getLogger();
		try
		{
			if (logger.isEnabledFor(Logger.INFO)) logger.info("Releasing MBeanServer " + server);

			if (server != null)
			{
				SecurityManager sm = System.getSecurityManager();
				if (sm != null)
				{
					sm.checkPermission(new MBeanServerPermission("releaseMBeanServer"));
				}

				boolean removed = false;
				synchronized (m_mbeanServers)
				{
					removed = m_mbeanServers.remove(server);
				}

				if (removed)
				{
					if (logger.isEnabledFor(Logger.INFO)) logger.info("MBeanServer " + server + " released successfully");
				}
				else
				{
					logger.warn("MBeanServer " + server + " not released, cannot find it");
				}
			}
			else
			{
				logger.warn("MBeanServer not released, it's null");
			}
		}
		catch (SecurityException x)
		{
			if (logger.isEnabledFor(Logger.INFO)) logger.info("Security Exception caught while releasing MBeanServer " + server, x);
			throw x;
		}
	}

	public static ArrayList findMBeanServer(String id)
	{
		Logger logger = getLogger();
		ArrayList list = null;
		try
		{
			if (logger.isEnabledFor(Logger.INFO)) {logger.info("Finding MBeanServer with ID: " + id);}

			SecurityManager sm = System.getSecurityManager();
			if (sm != null)
			{
				sm.checkPermission(new MBeanServerPermission("findMBeanServer"));
			}

			if (id == null)
			{
				list = (ArrayList)m_mbeanServers.clone();
			}
			else
			{
				list = new ArrayList();
				synchronized (m_mbeanServers)
				{
					for (int i = 0; i < m_mbeanServers.size(); ++i)
					{
						MBeanServer server = (MBeanServer)m_mbeanServers.get(i);
						try
						{
							String serverId = (String)server.getAttribute(new ObjectName("JMImplementation:type=MBeanServerDelegate"), "MBeanServerId");
							if (id.equals(serverId))
							{
								list.add(server);
								if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Found matching MBeanServer: " + server);}
							}
						}
						catch (Exception ignored)
						{
							logger.error("Delegate implementation is buggy", ignored);
						}
					}
				}
			}

			if (logger.isEnabledFor(Logger.INFO)) {logger.info("MBeanServer(s) found: " + list);}

			return list;
		}
		catch (SecurityException x)
		{
			if (logger.isEnabledFor(Logger.INFO)) logger.info("Security Exception caught while finding MBeanServer with ID: " + id, x);
			throw x;
		}
	}

	public static ClassLoaderRepository getClassLoaderRepository(MBeanServer server)
	{
        if (server instanceof MBeanServerImpl)
		{
			return ((MBeanServerImpl)server).getClassLoaderRepository();
		}
		else
		{
			return MBeanServerImpl.getStaticClassLoaderRepository();
		}
	}

	private static MBeanServer createMBeanServerImpl(String domain, String permission)
	{
		Logger logger = getLogger();
		try
		{
			if (logger.isEnabledFor(Logger.INFO)) {logger.info("Creating MBeanServer...");}

			SecurityManager sm = System.getSecurityManager();
			if (sm != null)
			{
				sm.checkPermission(new MBeanServerPermission(permission));
			}

			MBeanServer mbs = new MBeanServerImpl(domain);
			if (logger.isEnabledFor(Logger.INFO)) {logger.info("MBeanServer " + mbs + " created successfully");}
			return mbs;
		}
		catch (SecurityException x)
		{
			if (logger.isEnabledFor(Logger.INFO)) logger.info("Security Exception caught while creating an MBeanServer", x);
			throw x;
		}
	}
}
