/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management;

import java.security.BasicPermission;
import java.security.Permission;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.Collections;
import java.io.ObjectInputStream;
import java.io.IOException;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.6 $
 */
public final class MBeanServerPermission extends Permission
// Be careful in extending BasicPermission, since it provides a PermissionCollection
// that does not take in account the fact that createMBeanServer implies newMBeanServer
// (MBeanServerPermission.implies(Permission) is never called)
{
	private transient ArrayList targets;
	private transient boolean wildcard;

	public MBeanServerPermission(String name)
	{
		this(name, null);
	}

	public MBeanServerPermission(String name, String actions)
	{
		super(name);
		parseName(name);
	}

	public String getActions()
	{
		return null;
	}

	public int hashCode()
	{
		return targets.hashCode();
	}

	public boolean equals(Object obj)
	{
		if (obj == null) return false;
		if (obj == this) return true;

		try
		{
			MBeanServerPermission other = (MBeanServerPermission)obj;
			// We don't check wildcard: if targets is empty, means we have a wildcard (see parseName)
			return targets.equals(other.targets);
		}
		catch (ClassCastException x)
		{
		}
		return false;
	}

	public boolean implies(Permission p)
	{
		if (p == null) return false;
		if (getClass() != p.getClass()) return false;

		MBeanServerPermission other = (MBeanServerPermission)p;
		if (wildcard) return true;
		if (other.wildcard) return false;

		if (targets.containsAll(other.targets)) return true;

		// We have to manage the case where this contains createMBeanServer and other contains newMBeanServer
		if (other.targets.contains("newMBeanServer") && targets.contains("createMBeanServer"))
		{
			// Umpf, this is ugly, but not immediate: must care multithreading
			for (int i = 0; i < other.targets.size(); ++i)
			{
				Object perm = other.targets.get(i);
				if ("newMBeanServer".equals(perm)) continue;
				if (!targets.contains(perm)) return false;
			}
			return true;
		}

		return false;
	}

	private void parseName(String name)
	{
		if (name == null) throw new IllegalArgumentException("Permission name cannot be null");
		name = name.trim();
		if (name.length() == 0) throw new IllegalArgumentException("Permission name cannot be empty");

		targets = new ArrayList();
		StringTokenizer tokenizer = new StringTokenizer(name, ",");
		while (tokenizer.hasMoreTokens())
		{
			String target = tokenizer.nextToken().trim();
			if (target.length() == 0) continue;
			if ("*".equals(target))
			{
				targets.clear();
				wildcard = true;
				return;
			}
			else
			{
				targets.add(target);
			}
		}

		if (targets.size() < 1) throw new IllegalArgumentException("Permission name does not contain targets");

		// Important to provide same hashcode and equals to permission with names in different order
		Collections.sort(targets);
	}

	private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException
	{
        stream.defaultReadObject();
		parseName(getName());
	}
}
