/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import mx4j.log.Log;
import mx4j.log.Logger;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.7 $
 */
public class NotificationBroadcasterSupport implements NotificationBroadcaster
{
	private static final NotificationFilter NULL_FILTER = new NotificationFilter()
	{
		public boolean isNotificationEnabled(Notification notification) {return true;}
		public String toString() {return "null filter";}
	};

	private static final Object NULL_HANDBACK = new Object()
	{
		public String toString() {return "null handback";}
	};

	private HashMap m_listeners = new HashMap();

	public NotificationBroadcasterSupport() {}

	private Logger getLogger()
	{
		return Log.getLogger(getClass().getName());
	}

	public void addNotificationListener(NotificationListener listener, NotificationFilter filter, Object handback)
	{
		Logger logger = getLogger();

		if (logger.isEnabledFor(Logger.INFO)) logger.info("Adding notification listener: " + listener + ", filter: " + filter + ", handback: " + handback);

		if (listener == null) throw new RuntimeOperationsException(new IllegalArgumentException("Notification listener cannot be null"));

		// Normalize the arguments
		if (filter == null) filter = NULL_FILTER;
		if (handback == null) handback = NULL_HANDBACK;

		FilterHandbackPair pair = new FilterHandbackPair(filter, handback);

		synchronized (this)
		{
            ArrayList pairs = (ArrayList)m_listeners.get(listener);

			if (pairs == null)
			{
				// A new listener, register it
				pairs = new ArrayList();
				pairs.add(pair);
				m_listeners.put(listener, pairs);
			}
			else
			{
				// Check that the same triple (listener, filter, handback) is not already registered
				for (int i = 0; i < pairs.size(); ++i)
				{
					FilterHandbackPair other = (FilterHandbackPair)pairs.get(i);
					if (pair.getNotificationFilter().equals(other.getNotificationFilter()) && pair.getHandback().equals(other.getHandback()))
					{
						// Same filter and same handback for the same listener, it's already registered
						throw new RuntimeOperationsException(new IllegalArgumentException("Notification listener is already registered"));
					}
				}
				// Not yet registered, register.
				// Do not merge this call with the one in the if branch: like this is easier to debug
				// (I know the if-else branch from where I'm coming)
				pairs.add(pair);
			}

			if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Filters - Handbacks for this listener: " + pairs);
		}

		if (logger.isEnabledFor(Logger.INFO)) logger.info("Notification listener added successfully");
	}

	public void removeNotificationListener(NotificationListener listener) throws ListenerNotFoundException
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.INFO)) logger.info("Removing notification listener: " + listener);

		int removed = removeNotificationListenerImpl(listener, null, null);

		if (logger.isEnabledFor(Logger.INFO)) logger.info(removed + "notification listener(s) removed successfully");
	}

/*
	public void removeNotificationListener(NotificationListener listener, Object handback) throws ListenerNotFoundException
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.INFO)) {logger.info("Removing notification listener: " + listener + ", handback: " + handback);}

		// Normalize the argument if necessary
		if (handback == null) {handback = NULL_HANDBACK;}

		int removed = removeNotificationListenerImpl(listener, null, handback);

		if (logger.isEnabledFor(Logger.INFO))
		{
			logger.info(removed + "notification listener(s) removed successfully");
		}
	}

	public void removeNotificationListener(NotificationListener listener, NotificationFilter filter, Object handback) throws ListenerNotFoundException
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.INFO)) {logger.info("Removing notification listener: " + listener + ", filter: " + filter + ", handback: " + handback);}

		// Normalize the arguments if necessary
		if (filter == null) {filter = NULL_FILTER;}
		if (handback == null) {handback = NULL_HANDBACK;}

		int removed = removeNotificationListenerImpl(listener, filter, handback);

		if (logger.isEnabledFor(Logger.INFO))
		{
			logger.info(removed + "notification listener(s) removed successfully");
		}
	}
*/

	private int removeNotificationListenerImpl(NotificationListener listener, NotificationFilter filter, Object handback) throws ListenerNotFoundException
	{
		synchronized (this)
		{
			ArrayList pairs = (ArrayList)m_listeners.get(listener);

			if (pairs == null) throw new ListenerNotFoundException("Notification listener not found");

			if (filter == null)
			{
				if (handback == null)
				{
					// Means I want to remove all triplets for this listener

					ArrayList removed = (ArrayList)m_listeners.remove(listener);
					return removed.size();
				}
				else
				{
					// Means I want to remove all triplets with the given handback for this listener

					int count = 0;
					for (int i = 0; i < pairs.size(); ++i)
					{
						Object hand = ((FilterHandbackPair)pairs.get(i)).getHandback();
						if (handback.equals(hand))
						{
							pairs.remove(i);
							++count;
						}
					}

					// Check if it was the last listener
					if (pairs.isEmpty()) m_listeners.remove(listener);

					return count;
				}
			}
			else
			{
				if (handback == null)
				{
					// Means I want to remove all triplets with the given filter for this listener

					int count = 0;
					for (int i = 0; i < pairs.size(); ++i)
					{
						Object filt = ((FilterHandbackPair)pairs.get(i)).getNotificationFilter();
						if (filter.equals(filt))
						{
							pairs.remove(i);
							++count;
						}
					}

					// Check if it was the last listener
					if (pairs.isEmpty()) m_listeners.remove(listener);

					return count;
				}
				else
				{
					// Means I want to remove all triplets with the given filter and handback for this listener

					int count = 0;
					for (int i = 0; i < pairs.size(); ++i)
					{
						FilterHandbackPair pair = (FilterHandbackPair)pairs.get(i);
						if (filter.equals(pair.getNotificationFilter()) && handback.equals(pair.getHandback()))
						{
							pairs.remove(i);
							++count;
						}
					}

					// Check if it was the last listener
					if (pairs.isEmpty()) m_listeners.remove(listener);

					return count;
				}
			}
		}
	}

	public MBeanNotificationInfo[] getNotificationInfo()
	{
		// Subclasses should override returning more informations
		return new MBeanNotificationInfo[0];
	}

	private boolean hasListeners()
	{
		synchronized (this)
		{
			return m_listeners.size() > 0;
		}
	}

	public void sendNotification(final Notification n)
	{
		final Logger logger = getLogger();
		boolean debug = logger.isEnabledFor(Logger.DEBUG);
		final boolean info = logger.isEnabledFor(Logger.INFO);

		HashMap listeners = null;

		synchronized (this)
		{
			// Clone the listeners, so we can notify without holding any lock
			// It is a shallow copy, below we will clone the pairs as well
			// I don't care if in the middle someone else adds or remove other pairs
			listeners = (HashMap)m_listeners.clone();
		}

		// Loop over all listeners
		Iterator i = listeners.keySet().iterator();

		if (i.hasNext() && info) logger.info("Sending notifications...");

		while (i.hasNext())
		{
			final NotificationListener listener = (NotificationListener)i.next();
			if (info) logger.info("\tListener is: " + listener);

			// Clone again the pairs for this listener.
			// I freezed the listeners with the first clone, if someone removes a pair
			// in the middle of notifications I don't care: here I clone the actual pairs
			ArrayList pairs = null;
			synchronized (this)
			{
				pairs = (ArrayList)m_listeners.get(listener);
				pairs = (ArrayList)pairs.clone();
			}

			if (debug) logger.debug("\tFilters - Handback for this listener: " + pairs);

			// Loop over the same listener that registered many times with different filter / handbacks
			for (int j = 0; j < pairs.size(); ++j)
			{
				FilterHandbackPair pair = (FilterHandbackPair)pairs.get(j);

				final NotificationFilter filter = pair.getNotificationFilter();
				Object hb = pair.getHandback();

				// Denormalize the handback if necessary
				if (hb == NULL_HANDBACK) hb = null;

				final Object handback = hb;

				Runnable run = new Runnable()
				{
					public void run()
					{
						boolean enabled = false;
						try
						{
							enabled = filter != null && filter.isNotificationEnabled(n);
						}
						catch (RuntimeException x)
						{
							if (info) logger.info("\t\tRuntimeException caught from isNotificationEnabled", x);
							// And go on
						}

						if (info) logger.info("\t\tFilter is: " + filter + ", enabled: " + enabled);

						if (enabled)
						{
							if (info)
							{
								logger.info("\t\tHandback is: " + handback);
								logger.info("\t\tSending notification " + n);
							}
							try
							{
								listener.handleNotification(n, handback);
							}
							catch (RuntimeException x)
							{
								if (info) logger.info("\t\tRuntimeException caught from handleNotification", x);
								// And go on with next listener
							}
						}
					}
				};

				if (listener instanceof Serializable)
				{
					// It's remote, run in separate thread
					new Thread(run, "NotificationBroadcasterSupport Thread").start();
				}
				else
				{
					// Run in same thread, it's local
					run.run();
				}
			}
		}
	}

	private static class FilterHandbackPair
	{
		private NotificationFilter m_filter;
		private Object m_handback;

		public FilterHandbackPair(NotificationFilter filter, Object handback)
		{
			m_filter = filter;
			m_handback = handback;
		}

		public NotificationFilter getNotificationFilter()
		{
			return m_filter;
		}

		public Object getHandback()
		{
			return m_handback;
		}
	}
}
