/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management;

import java.util.Hashtable;
import java.util.Map;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.Iterator;
import java.io.Serializable;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.InvalidObjectException;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.8 $
 */
public class ObjectName implements Serializable
{
	private static final long serialVersionUID = 1081892073854801359L;

	private transient String m_domain;
	private transient Hashtable m_properties;
	private transient String m_propertiesString;
	private transient boolean m_isPropertyPattern;
	private transient boolean m_isDomainPattern;
	private transient int m_hashCode;

	private static HashMap m_cache = new HashMap();

	public ObjectName(String name) throws MalformedObjectNameException
	{
		parse(name);
	}

	public ObjectName(String domain, Hashtable table) throws MalformedObjectNameException
	{
		if (domain == null) {domain = "";}
		init(domain, table, null);
	}

	public ObjectName(String domain, String key, String value) throws MalformedObjectNameException
	{
		if (key == null || value == null || key.trim().length() == 0 || value.trim().length() == 0) {throw new MalformedObjectNameException("No property specified");}
		if (domain == null) {domain = "";}
		Hashtable table = new Hashtable();
		table.put(key, value);
		init(domain, table, null);
	}

	private void parse(String objectName) throws MalformedObjectNameException
	{
		boolean isSubclass = getClass() != ObjectName.class;
		boolean alreadyExists = false;

		// It is important from the security point of view to not cache subclasses.
		// An EvilObjectName may return an allowed domain when security checks are made, and
		// a prohibited domain when performing operations. Here we make sure that subclasses
		// are not cached.
		if (!isSubclass)
		{
			synchronized (m_cache)
			{
				ObjectName name = (ObjectName)m_cache.get(objectName);
				if (name != null)
				{
					// This ObjectName is already created, just copy it to avoid string parsing
					m_domain = name.getDomain();
					m_properties = name.getKeyPropertyList();
					m_propertiesString = name.getKeyPropertyListString();
					m_isDomainPattern = name.isDomainPattern();
					m_isPropertyPattern = name.isPropertyPattern();
					m_hashCode = name.hashCode();
					alreadyExists = true;
				}
			}
		}

		if (!alreadyExists)
		{
			String domain = parseDomain(objectName);
			String propertiesList = parsePropertiesList(objectName);
			Hashtable table = createPropertiesMap(propertiesList);
			init(domain, table, propertiesList);

			if (!isSubclass)
			{
				// Cache this ObjectName
				synchronized (m_cache)
				{
					// Overwrite if 2 threads computed the same ObjectName: we have been unlucky
					m_cache.put(objectName, this);
				}
			}
		}
	}

	private String parseDomain(String objectName) throws MalformedObjectNameException
	{
		int colon = objectName.indexOf(':');
		if (colon < 0) {throw new MalformedObjectNameException("Missing ':' character in ObjectName");}
		String domain = objectName.substring(0, colon);
		return domain;
	}

	private String parsePropertiesList(String objectName) throws MalformedObjectNameException
	{
		int colon = objectName.indexOf(':');
		if (colon < 0) {throw new MalformedObjectNameException("Missing ':' character in ObjectName");}
		String list = objectName.substring(colon + 1);
		return list;
	}

	private void init(String domain, Hashtable properties, String propertiesString) throws MalformedObjectNameException
	{
		if (properties == null || properties.size() < 1) {throw new MalformedObjectNameException("No property[ies] specified");}

		String dom = domain.trim();
		checkDomain(dom);
		m_domain = dom;

		checkProperties(properties);
		m_properties = properties;

		if (propertiesString == null)
		{
			m_propertiesString = convertPropertiesToString(properties);
		}
		else
		{
			// If the string is provided, I have to keep the property order, so I cannot use the Hashtable as I do above
			if (isPropertyPattern())
			{
				// I must remove the '*' from the string, just store the properties
				m_propertiesString = parseProperties(propertiesString);
			}
			else
			{
				m_propertiesString = propertiesString;
			}
		}

		// Calculate hash code for performance
		m_hashCode = computeHashCode();
	}

	private int computeHashCode()
	{
		int hash = getDomain().hashCode();
		hash ^= m_properties.hashCode();
		hash ^= new Boolean(isPropertyPattern()).hashCode();
		return hash;
	}

	private String parseProperties(String properties)
	{
		StringBuffer b = new StringBuffer();
		boolean firstTime = true;
		StringTokenizer tokenizer = new StringTokenizer(properties, ",");
		while (tokenizer.hasMoreTokens())
		{
			String property = tokenizer.nextToken().trim();
			// Keep only the properties, remove the asterisk
			if (!property.equals("*"))
			{
				if (!firstTime) {b.append(",");}
				else {firstTime = false;}
				b.append(property);
			}
		}
		return b.toString();
	}

	private void checkDomain(String domain) throws MalformedObjectNameException
	{
		if (domain.indexOf(':') >= 0) {throw new MalformedObjectNameException("Domain cannot contain ':' characters");}
		if (domain.indexOf(',') >= 0) {throw new MalformedObjectNameException("Domain cannot contain ',' characters");}
		if (domain.indexOf('=') >= 0) {throw new MalformedObjectNameException("Domain cannot contain '=' characters");}
		// Domain may contain '*' and '?' characters if it's a pattern
		if (domain.indexOf('*') >= 0 || domain.indexOf('?') >= 0) {m_isDomainPattern = true;}
	}

	private void checkProperties(Hashtable properties)
	{
		if (properties.containsKey("*"))
		{
			// The Hashtable will never contain the '*'
			properties.remove("*");
			m_isPropertyPattern = true;
		}
	}

	private Hashtable createPropertiesMap(String properties) throws MalformedObjectNameException
	{
		// At least '*' must be present
		if (properties.length() < 1) {throw new MalformedObjectNameException("Missing property");}
		if (properties.charAt(properties.length() - 1) == ',') {throw new MalformedObjectNameException("Missing property: ObjectName cannot end with ','");}

		Hashtable map = new Hashtable();
		StringTokenizer tokenizer = new StringTokenizer(properties, ",");
		while (tokenizer.hasMoreTokens())
		{
			String property = tokenizer.nextToken().trim();
			if (property.equals("*")) {map.put(property, property);}
			else
			{
				int equal = property.indexOf("=");
				if (equal < 0) {throw new MalformedObjectNameException("Missing '=' character in properties list");}
				String key = property.substring(0, equal).trim();
				if (key.length() < 1) {throw new MalformedObjectNameException("Missing key in properties list");}
				String value = property.substring(equal + 1).trim();
				if (value.length() < 1) {throw new MalformedObjectNameException("Missing value in properties list");}
				map.put(key, value);
			}
		}
		return map;
	}

	private String convertPropertiesToString(Map properties)
	{
		StringBuffer b = new StringBuffer();
		boolean firstTime = true;
		for (Iterator i = properties.entrySet().iterator(); i.hasNext();)
		{
			if (!firstTime) {b.append(",");}
			else {firstTime = false;}

			Map.Entry entry = (Map.Entry)i.next();
			b.append(entry.getKey());
			b.append("=");
			b.append(entry.getValue());
		}

		return b.toString();
	}

	public String getCanonicalKeyPropertyListString()
	{
		Map map = new TreeMap();
		map.putAll(m_properties);
		return convertPropertiesToString(map);
	}

	public String getCanonicalName() {return getDomain() + ":" + getCanonicalKeyPropertyListString();}

	public String getDomain() {return m_domain;}

	public String getKeyProperty(String property) {return (String)m_properties.get(property);}

	public Hashtable getKeyPropertyList() {return (Hashtable)m_properties.clone();}

	public String getKeyPropertyListString() {return m_propertiesString;}

	public boolean isPattern() {return isDomainPattern() || isPropertyPattern();}

	public boolean isPropertyPattern() {return m_isPropertyPattern;}

	private boolean isDomainPattern() {return m_isDomainPattern;}

	public int hashCode() {return m_hashCode;}

	public boolean equals(Object obj)
	{
		if (obj == null) {return false;}
		if (obj == this) {return true;}
		try
		{
			ObjectName other = (ObjectName)obj;
			return getDomain().equals(other.getDomain()) &&
				   m_properties.equals(other.m_properties) &&
				   isPattern() == other.isPattern() &&
				   isPropertyPattern() == other.isPropertyPattern();
		}
		catch (ClassCastException ignored) {}
		return false;
	}

	public String toString() {return getDomain() + ":" + getKeyPropertyListString();}

	private void writeObject(ObjectOutputStream out) throws IOException
	{
		StringBuffer b = new StringBuffer(getDomain());
		b.append(":");
		String properties = getKeyPropertyListString();
		// A bug in JMX RI forces MX4J to remove the spaces after the commas, sob :(
		if (properties.indexOf(" ") > 0)
		{
			StringTokenizer tokenizer = new StringTokenizer(properties, " ");
			while (tokenizer.hasMoreTokens()) {b.append(tokenizer.nextToken());}
		}
		else
		{
			b.append(properties);
		}

		if (isPropertyPattern())
		{
			if (m_properties.isEmpty()) {b.append("*");}
			else {b.append(",*");}
		}

		String name = b.toString();
		out.writeObject(name);
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
	{
		String objectName = (String)in.readObject();
		try
		{
			parse(objectName);
		}
		catch (MalformedObjectNameException x)
		{
			throw new InvalidObjectException("String representing the ObjectName is not a valid ObjectName: " + x.toString());
		}
	}
}
