/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management.loading;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLStreamHandlerFactory;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.ServiceNotFoundException;

import mx4j.loading.MLetParseException;
import mx4j.loading.MLetParser;
import mx4j.loading.MLetTag;
import mx4j.log.Log;
import mx4j.log.Logger;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @author <a href="mailto:shadow12@users.sourceforge.net">Bronwen Cassidy</a>
 * @version $Revision: 1.16 $
 */

/**
 * <p>Allows you to instantiate and register one or several MBeans in the MBean server coming from a remote URL. M-let is a shortcut for management applet.
 * The m-let service does this by loading an m-let text file, which specifies information on the MBeans to be obtained. The information on each MBean is
 * specified in a single instance of a tag, called the MLET tag. The location of the m-let text file is specified by a URL </p>
 * <p>The MLET tag has the following syntax:</p>
 * <pre>
 *  <MLET
 * 		CODE = class | OBJECT = serfile
 * 		ARCHIVE = "archiveList"
 * 		[CODEBASE = codebaseURL]
 * 		[NAME = mbeanname]
 * 		[VERSION = version] >
 * 	  [arglist]
 *  </MLET>
 * </pre>
 * <p>Where: </p>
 * <p>CODE = class<br/>
 * 			 This attribute specifies the full Java class name, including package name, of the MBean to be obtained. The compiled .class file of the MBean must be contained
 * 			 in one of the .jar files specified by the ARCHIVE attribute. Either CODE or OBJECT must be present.</p>
 * <p>OBJECT = serfile<br/>
 *			This attribute specifies the .ser file that contains a serialized representation of the MBean to be obtained. This file must be contained in one of the .jar files
 * 			specified by the ARCHIVE attribute. If the .jar file contains a directory hierarchy, specify the path of the file within this hierarchy.
 * 			Otherwise a match will not be found. Either CODE or OBJECT must be present.</p>
 * <p>ARCHIVE = <i>"archiveList"</i><br/>
 * 			 This mandatory attribute specifies one or more .jar files containing MBeans or other resources used by the MBean to be obtained.
 * 			 One of the .jar files must contain the file specified by the CODE or OBJECT attribute. If archivelist contains more than one file:<br/>
 * <ul>
 * 		<li>Each file must be separated from the one that follows it by a comma (,).</li>
 *      <li><i>archivelist</i> must be enclosed in double quote marks. </li>
 * </ul>
 * All .jar files in <i>archivelist</i> must be stored in the directory specified by the code base URL.</p>
 * <p>CODEBASE = <i>codebaseURL</i><br/>
 *            This optional attribute specifies the code base URL of the MBean to be obtained. It identifies the directory that contains the .jar files specified by the ARCHIVE attribute.
 *            Specify this attribute only if the .jar files are not in the same directory as the m-let text file. If this attribute is not specified, the base URL of the m-let text file is used.</p>
 * <p>NAME = <i>mbeanname</i><br/>
 *        This optional attribute specifies the object name to be assigned to the MBean instance when the m-let service registers it.
 *        If mbeanname starts with the colon character (:), the domain part of the object name is the domain of the agent.
 *        The m-let service invokes the getDomain() method of the Framework class to obtain this information.</p>
 * <p>VERSION = <i>version</i></br>
 * 			This optional attribute specifies the version number of the MBean and associated .jar files to be obtained. This version number can be used to specify that the .jar files are
 *          loaded from the server to update those stored locally in the cache the next time the m-let text file is loaded. version must be a series of non-negative decimal integers each
 *          separated by a period from the one that precedes it.</p>
 * <p><i>arglist</i><br/>
 * 			This optional attribute specifies a list of one or more parameters for the MBean to be instantiated.
 *          This list describes the parameters to be passed the MBean's constructor. Use the following syntax to specify each item in arglist:
 * <pre>
 *                      <ARG TYPE=argumentType VALUE=value>
 * </pre>
 *  where:
 * <ul>
 * <li><i>argumentType</i> is the type of the argument that will be passed as parameter to the MBean's constructor</li>
 * </ul>
 *
 * The arguments' type in the argument list should be a Java primitive type or a Java basic type
 * (<tt>java.lang.Boolean</tt>, <tt>java.lang.Byte</tt>, <tt>java.lang.Short</tt>, <tt>java.lang.Long</tt>, <tt>java.lang.Integer</tt>, <tt>java.lang.Float</tt>,
 * <tt>java.lang.Double</tt>, <tt>java.lang.String</tt>). </p>
 * <p>When an m-let text file is loaded, an instance of each MBean specified in the file is created and registered</p>
 * <p>The m-let Service extends the <tt>java.net.URLClassLoader</tt> and can be used to load remote classes and jar files in the VM of the agent</p>
 * <p><b>Note</b> - The MLet class loader uses the @see javax.management.MBeanServerFactory#getClassLoaderRepository to load classes that could not be found in the loaded jar files.</p>
 */
public class MLet extends URLClassLoader implements MLetMBean, MBeanRegistration
{
	private String m_libraryDirectory;
	private ObjectName m_objectName;
	private MBeanServer m_server;
	private String m_defaultCodeBase;
	private ThreadLocal m_loadingOnlyLocally = new ThreadLocal();
	private ThreadLocal m_loadingWithRepository = new ThreadLocal();
	private static ThreadLocal m_loadClassOriginator = new ThreadLocal();

	/**
	 * <p>Constructs a new MLet using the default delegation parent ClassLoader</p>
	 */
	public MLet()
	{
		this(new URL[0]);
	}

	/**
	 * <p>Constructs a new MLet for the specified URLs using the default delegation parent ClassLoader. The URLs will be searched in the order specified for classes and
	 * resources after first searching in the parent class loader.</p>
	 *
	 * @param urls - the urls from which to load classes and resources
	 */
	public MLet(URL[] urls)
	{
		super(urls);
		m_loadingWithRepository.set(Boolean.FALSE);
		m_loadingOnlyLocally.set(Boolean.FALSE);
	}

	/**
	 * <p>Constructs a new MLet for the given URLs. The URLs will be searched in the order specified for classes and resources after first searching in
	 * the specified parent class loader. The parent argument will be used as the parent class loader for delegation. The factory argument will be used
	 * as the stream handler factory to obtain protocol handlers when creating new URLs. </p>
	 *
	 * @param urls - the urls from which to load classes and resources
	 * @param parent - The parent class loader for delegation.
	 */
	public MLet(URL[] urls, ClassLoader parent)
	{
		super(urls, parent);
		m_loadingWithRepository.set(Boolean.FALSE);
		m_loadingOnlyLocally.set(Boolean.FALSE);
	}

	/**
	 * <p>Constructs a new MLet for the specified URLs, parent class loader, and URLStreamHandlerFactory. The parent argument will be used as the parent class loader for delegation.
	 * The factory argument will be used as the stream handler factory to obtain protocol handlers when creating new URLs.</p>
	 *
	 * @param urls - the urls from which to load classes and resources
	 * @param parent - The parent class loader for delegation.
	 * @param factory - The URLStreamHandlerFactory to use when creating URLs
	 */
	public MLet(URL[] urls, ClassLoader parent, URLStreamHandlerFactory factory)
	{
		super(urls, parent, factory);
		m_loadingWithRepository.set(Boolean.FALSE);
		m_loadingOnlyLocally.set(Boolean.FALSE);
	}

	/**
	 * <p>Allows the m-let to perform any operations it needs before being registered in the MBean server. If the ObjectName is null, the m-let provides a default name for its registration :type=MLet</p>
	 *
	 * @param server - the MBeanServer in which the MLet will be registered.
	 * @param name - the ObjectName of the MLet
	 *
	 * @throws Exception - This exception should be caught by the MBean server and re-thrown as an MBeanRegistrationException.
	 */
	public ObjectName preRegister(MBeanServer server, ObjectName name) throws Exception
	{
		m_server = server;
		m_objectName = name == null ? new ObjectName(m_server.getDefaultDomain(), "service", "MLet") : name;
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("MLet service " + m_objectName + " preRegistered successfully");
		return m_objectName;
	}

	/**
	 * <p>Allows the m-let to perform any operations needed after having been registered in the MBean server or after the registration has failed.</p>
	 *
	 * @param registrationDone - Indicates whether or not the m-let has been successfully registered in the MBean server. The value false means that either the registration phase has failed.
	 */
	public void postRegister(Boolean registrationDone)
	{
		Logger logger = getLogger();
		boolean done = registrationDone.booleanValue();
		if (!done)
		{
			m_server = null;
			logger.warn("MLet service " + m_objectName + " was not registered");
		}
		else
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("MLet service " + m_objectName+ " postRegistered successfully");
		}
	}

	/**
	 * <p>Allows the m-let to perform any operations it needs before being de-registered by the MBean server.</p>
	 *
	 * @throws Exception - This exception should be caught by the MBean server and re-thrown as an MBeanRegistrationException
	 */
	public void preDeregister() throws Exception
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) logger.debug("MLet service " + m_objectName + " preDeregistered successfully");
	}

	/**
	 * <p>Allows the m-let to perform any operations needed after having been de-registered in the MBean server</p>
	 */
	public void postDeregister()
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) logger.debug("MLet service " + m_objectName + " postDeregistered successfully");
	}

	/**
	 * <p>Appends the specified URL to the list of URLs to search for classes and resources.</p>
	 *
	 * @param url - the url represented as a string to add to the list of urls specified
	 * @throws - ServiceNotFoundException if the specified URL is malformed
	 */
	public void addURL(String url) throws ServiceNotFoundException
	{
		addURL(createURL(url));
	}

	/**
	 * <p>Appends the specified URL to the list of URLs to search for classes and resources</p>
	 *
	 * @param url - the url as a URL to be added to the list of urls specifed
	 */
	public void addURL(URL url)
	{
		Logger logger = getLogger();
		if (!Arrays.asList(getURLs()).contains(url))
		{
			if (logger.isEnabledFor(Logger.INFO)) logger.info("Adding URL to this MLet (" + m_objectName + ") classpath: " + url);
			super.addURL(url);
		}
		else
		{
			if (logger.isEnabledFor(Logger.INFO)) logger.info("URL already present in this MLet (" + m_objectName + ") classpath: " + url);
		}
	}

	/**
	 * <p>Load a class using the given ClassLoaderRepository, iterating through the list of class loaders. The given ClassLoaderRepository can be null, in which case it is equivalent
	 * to <tt>ClassLoader.loadClass(java.lang.String)</tt></p>
	 *
	 * @param name - The name of the class we want to load.
	 * @param repository - The ClassLoaderRepository that will be used to search for the given class, if it is not found in this ClassLoader
	 * @return - the loaded class object
	 * @throws - ClassNotFoundException if the specified class could not be found, neither in this ClassLoader nor in the given ClassLoaderRepository
	 */
	public Class loadClass(String name, ClassLoaderRepository repository) throws ClassNotFoundException
	{
		if (repository == null)
		{
			// This is a change to align the implementation of MLet to the one of the Reference Implementation.
			// By looking at the javadocs, the correct behavior is the one remarked, sigh :(
//			return loadClass(name);
			Class cls = loadClassLocally(name);
			return cls;
		}
		else
		{
			try
			{
				Class cls = loadClassLocally(name);
				return cls;
			}
			catch (ClassNotFoundException x)
			{
				// Not found locally, try the given repository
				Class cls = findClassInRepository(name, repository);
				return cls;
			}
		}
	}

	private Class loadClassLocally(String name) throws ClassNotFoundException
	{
		// Here I must call loadClass(name) but not delegate to the CLR when I arrive to findClass
		// I cannot call findClassLocally directly because otherwise the mechanism of parent delegation
		// implemented in loadClass() is skipped and the parent loaders do not have the chance to
		// load the given class

		try
		{
			m_loadingOnlyLocally.set(Boolean.TRUE);
			return loadClass(name);
		}
		finally
		{
			m_loadingOnlyLocally.set(Boolean.FALSE);
		}
	}

	/**
	 * <p>This is the main method for class loaders to load/find the class</p>
	 *
	 * @param name - the name of the class to load/find
	 * @throws ClassNotFoundException - if the class was not found
	 */
	protected Class findClass(String name) throws ClassNotFoundException
	{
		if (m_loadingWithRepository.get() == Boolean.TRUE) throw new ClassNotFoundException(name);

		MLet originator = (MLet)m_loadClassOriginator.get();
		if (originator == null)
		{
			originator = this;
			m_loadClassOriginator.set(originator);
		}

		try
		{
			Logger logger = getLogger();
			if (logger.isEnabledFor(Logger.INFO)) logger.info("Finding class " + name + "...");
			try
			{
				Class cls = findClassLocally(name);
				if (logger.isEnabledFor(Logger.INFO)) logger.info("Class " + name + " found in this MLet (" + m_objectName + ") classpath: " + this);
				return cls;
			}
			catch (ClassNotFoundException x)
			{
				if (m_loadingOnlyLocally.get() == Boolean.TRUE) throw x;

				if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Class " + name + " not found in this MLet (" + m_objectName + ") classpath: " + this + ", trying the ClassLoaderRepository...", x);
				try
				{
					m_loadingWithRepository.set(Boolean.TRUE);
					ClassLoaderRepository repository = MBeanServerFactory.getClassLoaderRepository(m_server);
					Class cls = findClassInRepository(name, repository);
					if (logger.isEnabledFor(Logger.INFO)) logger.info("Class " + name + " found with the ClassLoaderRepository " + repository);
					return cls;
				}
				catch (ClassNotFoundException xx)
				{
					if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Class " + name + " not found in the ClassLoaderRepository, trying the DefaultLoaderRepository...", xx);
					try
					{
						// Try with the static classloader repository
						ClassLoaderRepository repository = MBeanServerFactory.getClassLoaderRepository(null);
						Class cls = findClassInRepository(name, repository);
						if (logger.isEnabledFor(Logger.INFO)) logger.info("Class " + name + " found with the DefaultLoaderRepository " + repository);
						return cls;
					}
					catch (ClassNotFoundException xxx)
					{
						if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Class " + name + " not found in the DefaultLoaderRepository, giving up", xxx);

						// Ok, here I'm sure the class cannot be found.
						// There is no point in looping on other classloaders and redo the whole search
						// with a different MLet as starting point, since it will ask again the CLR, that
						// already said it cannot find it, and to the DLR, that already said the same.
						throw new MLetClassNotFoundError();
					}
				}
				finally
				{
					m_loadingWithRepository.set(Boolean.FALSE);
				}
			}
		}
		catch (MLetClassNotFoundError x)
		{
			if (originator == this) throw new ClassNotFoundException(name);
			throw x;
		}
		finally
		{
			if (originator == this) m_loadClassOriginator.set(null);
		}
	}

	private Class findClassLocally(String name) throws ClassNotFoundException
	{
		return super.findClass(name);
	}

	private Class findClassInRepository(String name, ClassLoaderRepository repository) throws ClassNotFoundException
	{
		// Try with the classloader repository of the same MBeanServer
		return repository.loadClassWithout(this, name);
	}

	/**
	 * <p>Loads a text file containing MLET tags that define the MBeans to be added to the agent. The location of the text file is specified by a URL.
	 * The MBeans specified in the MLET file will be instantiated and registered by the MBean server.</p>
	 *
	 * @param url - The URL of the text Mlet file to be loaded as String object
	 * @return A set containing one entry per MLET tag in the m-let text file loaded. Each entry specifies either the ObjectInstance for the created MBean,
	 * 			 or a throwable object (that is, an error or an exception) if the MBean could not be created.
	 * @throws ServiceNotFoundException - One of the following errors has occurred:
	 * 	<ul>
	 * 		<li>The m-let text file does not contain an MLET tag</li>
	 * 		<li>The m-let text file is not found</li>
	 *      <li>A mandatory attribute of the MLET tag is not specified</li>
	 *      <li>The url is malformed.</li>
	 * </ul>
	 * @throws IllegalStateException - MLet MBean is not registered with an MBeanServer
	 */
	public Set getMBeansFromURL(String url) throws ServiceNotFoundException
	{
		return getMBeansFromURL(createURL(url));
	}

	/**
	 * <p>Loads a text file containing MLET tags that define the MBeans to be added to the agent. The location of the text file is specified by a URL.
	 * The MBeans specified in the MLET file will be instantiated and registered by the MBean server </p>
	 *
	 * @param url - The URL of the file to loaded as a URL
	 * @return: A set containing one entry per MLET tag in the m-let text file loaded. Each entry specifies either the ObjectInstance for the created MBean,
	 * 			or a throwable object (that is, an error or an exception) if the MBean could not be created.
	 * @throws: ServiceNotFoundException - One of the following errors has occurred:
	 * 		<ul><li>The m-let text file does not contain an MLET tag</li>
	 * 			<li>The m-let text file is not found</li>
	 *          <li>A mandatory attribute of the MLET tag is not specified</li>
	 *          <li>The value of url is null.</li>
	 *      </ul>
	 * @throws IllegalStateException - MLet MBean is not registered with an MBeanServer.
	 */
	public Set getMBeansFromURL(URL url) throws ServiceNotFoundException
	{
		if (url == null) throw new ServiceNotFoundException("Cannot load MBeans from null URL");

		Logger logger = getLogger();
		String urlString = url.toString();
		m_defaultCodeBase = urlString.substring(0, urlString.lastIndexOf('/') + 1);

		InputStream is = null;
		try
		{
			is = url.openStream();
			if (logger.isEnabledFor(Logger.INFO)) logger.info("MLet " + m_objectName + ", parsing MLET URL " + url);
			return getMBeansFromURL(is);
		}
		catch (IOException x)
		{
			if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("MLet " + m_objectName + ", cannot open MLET URL", x);
			throw new ServiceNotFoundException(x.toString());
		}
		finally
		{
			if (is != null)
			{
				try
				{
					is.close();
				}
				catch (IOException ignored)
				{
				}
			}
		}
	}

	private Logger getLogger()
	{
		return Log.getLogger(getClass().getName());
	}

	private String getDefaultCodeBase()
	{
		if (m_defaultCodeBase == null)
		{
			String userDir = System.getProperty("user.dir");
			File f = new File(userDir);
			try
			{
				String url = f.toURL().toString();
				m_defaultCodeBase = url.substring(0, url.lastIndexOf('/') + 1);
			}
			catch (MalformedURLException ignored)
			{
				m_defaultCodeBase = "";
			}
		}
		return m_defaultCodeBase;
	}

	private Set getMBeansFromURL(InputStream is) throws ServiceNotFoundException
	{
		Logger logger = getLogger();
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		BufferedOutputStream os = new BufferedOutputStream(baos);
		try
		{
			readFromAndWriteTo(is, os);
		}
		catch (IOException x)
		{
			logger.error("Cannot read input stream", x);
			throw new ServiceNotFoundException(x.toString());
		}
		finally
		{
			try
			{
				os.close();
			}
			catch (IOException ignored)
			{
			}
		}

		try
		{
			HashSet mbeans = new HashSet();
			MLetParser parser = new MLetParser();
			List tags = parser.parse(new String(baos.toByteArray()));

			for (int i = 0; i < tags.size(); ++i)
			{
				MLetTag tag = (MLetTag)tags.get(i);

				// Add the MBean's codebase to the MLet classloader
				URL[] urls = tag.getURLs(getDefaultCodeBase());
				for (int j = 0; j < urls.length; ++j)
				{
					addURL(urls[j]);
				}

				// Create and register the MBean
				Object obj = createMBean(tag);
				mbeans.add(obj);
			}

			return mbeans;
		}
		catch (MLetParseException x)
		{
			throw new ServiceNotFoundException(x.toString());
		}
	}

	/**
	 * <p>Returns the absolute path name of a native library. The VM invokes this method to locate the native libraries that belong to classes loaded with this class loader.
	 * Libraries are searched in the JAR files using first just the native library name and if not found the native library name together with the architecture-specific path name
	 * (OSName/OSArch/OSVersion/lib/nativelibname), i.e. </p>
	 * <p>the library stat on Solaris SPARC 5.7 will be searched in the JAR file as:</p>
	 * <ol><li>libstat.so</li>
	 *     <li>SunOS/sparc/5.7/lib/libstat.so </li>
	 * </ol>
	 * <p>the library stat on Windows NT 4.0 will be searched in the JAR file as: </p>
	 * <ol><li>stat.dll </li>
	 *     <li>WindowsNT/x86/4.0/lib/stat.dll </li>
	 * </ol>
	 * If this method returns null, i.e. the libraries were not found in any of the JAR files loaded with this class loader,
	 * the VM searches the library along the path specified as the java.library.path property
	 *
	 * @param libraryName - the name of the library
	 *
	 * @return the absolute path of the native library
	 */
	protected String findLibrary(String libraryName)
	{
		// When asked to load a library, I should convert the library name to the system specific name
		// For example if libraryName == "stat", on Solaris the conversion yields libstat.so, on W2K yields stat.dll
		String sysLibraryName = System.mapLibraryName(libraryName);

		InputStream is = getResourceAsStream(sysLibraryName.replace(File.separatorChar, '/'));
		// try to load the native libaray from jar only using the library name
		String path = loadResource(sysLibraryName, is);
		if (path != null) return path;

		// library not found so try to load the library using the os specific architecture
		StringBuffer sysBuff = new StringBuffer();
		sysBuff.append(removeWhiteSpace(System.getProperty("os.name").trim())).append(File.separator);
		sysBuff.append(removeWhiteSpace(System.getProperty("os.arch").trim())).append(File.separator);
		sysBuff.append(removeWhiteSpace(System.getProperty("os.version").trim())).append(File.separator);
		sysBuff.append("lib").append(File.separator).append(sysLibraryName);

		path = loadResource(sysBuff.toString(), is);
		if (path != null) return path;

		// 3. Search the java.library.path
		StringTokenizer tokenizer = new StringTokenizer(System.getProperty("java.library.path").trim(), File.pathSeparator);
		while (tokenizer.hasMoreTokens())
		{
			String filePath = new StringBuffer(tokenizer.nextToken().trim()).append(File.separator).append(sysLibraryName).toString();
			try
			{
				path = loadResource(filePath, new FileInputStream(filePath));
				if (path != null) return path;
			}
			catch (FileNotFoundException ex)
			{
			}
		}
		// all paths exhausted return null
		return null;
	}

	private String loadResource(String sysLibraryName, InputStream is)
	{
		if (is != null)
		{
			// I should store the library in the given library directory
			try
			{
				File file = null;
				if (getLibraryDirectory() == null) file = new File(sysLibraryName);
				else file = new File(getLibraryDirectory(), sysLibraryName);

				OutputStream os = null;
				try
				{
					os = new BufferedOutputStream(new FileOutputStream(file));
					is = new BufferedInputStream(is);
					// Delete if already exists
					if (file.exists()) file.delete();

					readFromAndWriteTo(is, os);
					return file.getCanonicalPath();
				}
				finally
				{
					if (is != null) is.close();
					if (os != null) os.close();
				}
			}
			catch (IOException ignored)
			{
			}
		}
		return null;
	}

	private void readFromAndWriteTo(InputStream is, OutputStream os) throws IOException
	{
		byte[] buffer = new byte[65536];
		int read = -1;
		while ((read = is.read(buffer)) >= 0)
		{
			os.write(buffer, 0, read);
		}
		buffer = null;
	}

	private String removeWhiteSpace(String word)
	{
		int j = word.indexOf(' ');
		// no spaces in the word return it
		if (j < 0) return word;

		StringBuffer temp = new StringBuffer();
		int i = 0;
		while (j >= 0)
		{
			word = word.substring(i);
			j = word.indexOf(' ');
			// store the word up to the space in temp
			if (j >= 0) temp = temp.append(word.substring(0, j));
			else temp = temp.append(word.substring(0));
			i = j + 1;
		}
		return temp.toString();
	}

	/**
	 * <p>Gets the current directory used by the library loader for storing native libraries before they are loaded into memory</p>
	 *
	 * @return The current directory used by the library loader
	 */
	public String getLibraryDirectory()
	{
		return m_libraryDirectory;
	}

	/**
	 * <p>Sets the directory used by the library loader for storing native libraries before they are loaded into memory</p>
	 *
	 * @param libdir - the directory the library is stored in.
	 */
	public void setLibraryDirectory(String libdir)
	{
		m_libraryDirectory = libdir;
	}

	private Object createMBean(MLetTag tag) throws ServiceNotFoundException
	{
		if (m_server == null) throw new ServiceNotFoundException("MLet not registered on the MBeanServer");

		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.INFO)) logger.info("MLet " + m_objectName + ", creating MBean from\n" + tag);

		try
		{
			Object mbean = null;
			if (tag.getObject() != null)
			{
				// Read the file from the codebase URLs of this classloader
				String name = tag.getObject();
				InputStream is = getResourceAsStream(name);
				if (is == null) throw new ServiceNotFoundException("Cannot find MBean " + name + " in this MLet (" + m_objectName + ") classpath");

				InputStream bis = new BufferedInputStream(is);
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				OutputStream os = new BufferedOutputStream(baos);
				readFromAndWriteTo(bis, os);

				// Deserialize using the MLet classloader
				ObjectInputStream ois = m_server.deserialize(m_objectName, baos.toByteArray());
				mbean = ois.readObject();
			}
			else
			{
				// Instantiate using the MLet classloader
				String clsName = tag.getCode();
				Object[] args = tag.getArguments();
				String[] params = tag.getSignature();
				mbean = m_server.instantiate(clsName, m_objectName, args, params);
			}

			ObjectName objectName = tag.getObjectName();
			ObjectInstance instance = m_server.registerMBean(mbean, objectName);
			return instance;
		}
		catch (Throwable t)
		{
			return t;
		}
	}

	private URL createURL(String urlString) throws ServiceNotFoundException
	{
		try
		{
			URL url = new URL(urlString);
			return url;
		}
		catch (MalformedURLException x)
		{
			throw new ServiceNotFoundException(x.toString());
		}
	}

	private static class MLetClassNotFoundError extends Error
	{
	}
}
