/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management.modelmbean;

import java.util.Map;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Iterator;
import java.io.Serializable;

import javax.management.Descriptor;
import javax.management.RuntimeOperationsException;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.4 $
 */
public class DescriptorSupport implements Descriptor
{
	private static final long serialVersionUID = -6292969195866300415L;

	private HashMap descriptor = new HashMap();

	public DescriptorSupport() {}

	public DescriptorSupport(String[] pairs)
	{
		if (pairs == null) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid field pairs"));}
		for (int i = 0; i < pairs.length; ++i)
		{
			String pair = pairs[i];
			if (pair == null) {throw new RuntimeOperationsException(new IllegalArgumentException("Illegal pair: " + pair));}
			int equal = pair.indexOf('=');
			if (equal < 1) {throw new RuntimeOperationsException(new IllegalArgumentException("Illegal pair: " + pair));}
			else
			{
				String name = pair.substring(0, equal);
				Object value = null;
				if (equal < pair.length() - 1)
				{
					value = pair.substring(equal + 1);
				}
				setField(name, value);
			}
		}
	}


	public DescriptorSupport(String[] names, Object[] values)
	{
		setFields(names, values);
	}

	public Object clone() throws RuntimeOperationsException
	{
		try
		{
			return super.clone();
		}
		catch (CloneNotSupportedException x)
		{
			throw new RuntimeOperationsException("Clone operation not supported");
		}
	}

	public Object getFieldValue(String name) throws RuntimeOperationsException
	{
        if (name == null || name.trim().length() == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid field name"));}
		Object value = descriptor.get(name.toLowerCase());
		return value;
	}

	public void setField(String name, Object value) throws RuntimeOperationsException
	{
        checkField(name, value);
        descriptor.put(name.toLowerCase(), value);
	}

	public void removeField(String name)
	{
		descriptor.remove(name);
	}

	public String[] getFieldNames()
	{
		String[] names = (String[])descriptor.keySet().toArray(new String[0]);
		return names;
	}

	public Object[] getFieldValues(String[] names)
	{
		if (names == null)
		{
			// All values must be returned
			Object[] values = (Object[])descriptor.values().toArray(new Object[0]);
			return values;
		}

		ArrayList list = new ArrayList();
		for (int i = 0; i < names.length; ++i)
		{
			try
			{
				Object value = getFieldValue(names[i]);
				list.add(value);
			}
			catch (RuntimeOperationsException x)
			{
				list.add(null);
			}
		}
		Object[] values = (Object[])list.toArray(new Object[list.size()]);
		return values;
	}

	public String[] getFields()
	{
        ArrayList list = new ArrayList();
		StringBuffer buffer = new StringBuffer();
		for (Iterator i = descriptor.entrySet().iterator(); i.hasNext();)
		{
			Map.Entry entry = (Map.Entry)i.next();
			String key = (String)entry.getKey();
			Object value = entry.getValue();
			buffer.setLength(0);
			buffer.append(key);
			buffer.append("=");
			if (value != null)
			{
				if (value instanceof String)
				{
					buffer.append(value.toString());
				}
				else
				{
					buffer.append("(");
					buffer.append(value.toString());
					buffer.append(")");
				}
			}
			list.add(buffer.toString());
		}
		String[] fields = (String[])list.toArray(new String[list.size()]);
		return fields;
	}

	public void setFields(String[] names, Object[] values) throws RuntimeOperationsException
	{
		if (names == null || values == null || names.length != values.length) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid arguments"));}

		for (int i = 0; i < names.length; ++i)
		{
			setField(names[i], values[i]);
		}
	}

	public boolean isValid() throws RuntimeOperationsException
	{
		try
		{
			for (Iterator i = descriptor.entrySet().iterator(); i.hasNext();)
			{
				Map.Entry entry = (Map.Entry)i.next();
				String name = (String)entry.getKey();
				Object value = entry.getValue();
				checkField(name, value);
			}
			return true;
		}
		catch (RuntimeOperationsException x)
		{
			return false;
		}
	}

	private void checkField(String name, Object value)  throws RuntimeOperationsException
	{
        if (name == null || name.trim().length() == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Illegal field name"));}

		boolean isValid = true;

		if (name.equalsIgnoreCase("name"))
		{
			isValid = value != null;
		}
		else if (name.equalsIgnoreCase("descriptorType"))
		{
			isValid = value != null && (value.equals("MBean") ||
										value.equals("attribute") ||
										value.equals("operation") ||
										value.equals("notification"));
		}
        else if (name.equalsIgnoreCase("role"))
        {
            isValid = value != null && (value.equals("constructor") ||
										value.equals("operation") ||
										value.equals("getter") ||
										value.equals("setter"));
        }
        else if (name.equalsIgnoreCase("persistPolicy"))
		{
			isValid = value != null && (value.equals("Never") ||
										value.equals("OnTimer") ||
										value.equals("OnUpdate") ||
										value.equals("NoMoreOftenThan"));
		}
		else if (name.equalsIgnoreCase("persistPeriod"))
		{
			int v = objectToInt(value);
			isValid = v > 0;
		}
		else if (name.equalsIgnoreCase("currencyTimeLimit"))
		{
			int v = objectToInt(value);
			isValid = v >= -1;
		}
		else if (name.equalsIgnoreCase("export"))
		{
			isValid = value instanceof Serializable || value == null;
		}
		else if (name.equalsIgnoreCase("visibility"))
		{
			int v = objectToInt(value);
			isValid = v >= 1 && v <= 4;
		}
		else if (name.equalsIgnoreCase("getMethod") || name.equalsIgnoreCase("setMethod"))
		{
			isValid = value != null && value.toString().trim().length() > 0;
		}
		else if (name.equalsIgnoreCase("protocolMap"))
		{
			isValid = value instanceof Descriptor;
		}
		else if (name.equalsIgnoreCase("lastUpdatedTimeStamp"))
		{
			long v = objectToLong(value);
			isValid = v > 0;
		}
		else if (name.equalsIgnoreCase("severity"))
		{
			int v = objectToInt(value);
			isValid = v >= 0 && v <= 6;
		}
		else if (name.equalsIgnoreCase("messageId"))
		{
			isValid = value != null;
		}

		if (!isValid)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Invalid value '" + value + "' for field " + name));
		}
	}

	private int objectToInt(Object obj) throws RuntimeOperationsException
	{
		try
		{
			return Integer.parseInt(obj.toString());
		}
		catch (NumberFormatException x)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Illegal value '" + obj + "' for numeric field"));
		}
		catch (NullPointerException x)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Illegal value '" + obj + "' for numeric field"));
		}
	}
	private long objectToLong(Object obj) throws RuntimeOperationsException
	{
		try
		{
			return Long.parseLong(obj.toString());
		}
		catch (NumberFormatException x)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Illegal value '" + obj + "' for numeric field"));
		}
		catch (NullPointerException x)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Illegal value '" + obj + "' for numeric field"));
		}
	}
}
