/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management.modelmbean;

import java.util.List;
import java.util.Arrays;

import javax.management.MBeanNotificationInfo;
import javax.management.DescriptorAccess;
import javax.management.Descriptor;
import javax.management.RuntimeOperationsException;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.4 $
 */
public class ModelMBeanNotificationInfo extends MBeanNotificationInfo implements DescriptorAccess
{
	private static final long serialVersionUID = -7445681389570207141L;

	private Descriptor notificationDescriptor;

	public ModelMBeanNotificationInfo(String[] types, String name, String description)
	{
		this(types, name, description, null);
	}
	public ModelMBeanNotificationInfo(String[] types, String name, String description, Descriptor descriptor)
	{
		super(types, name, description);
		checkAndSetDescriptor(descriptor);
	}
	public ModelMBeanNotificationInfo(ModelMBeanNotificationInfo copy)
	{
		super(copy.getNotifTypes(), copy.getName(), copy.getDescription());
		checkAndSetDescriptor(copy.getDescriptor());
	}

	public Descriptor getDescriptor()
	{
		return (Descriptor)notificationDescriptor.clone();
	}

	public void setDescriptor(Descriptor descriptor)
	{
        if (descriptor == null)
		{
			notificationDescriptor = createDefaultDescriptor();
		}
		else
		{
            if (isDescriptorValid(descriptor))
			{
				notificationDescriptor = (Descriptor)descriptor.clone();
			}
			else
			{
				// Not sure what to do here: javadoc says IllegalArgument, but for example ModelMBeanInfo throws RuntimeOperations
				// which is consistent with the fact that all exception thrown by the JMX implementation should be JMX exceptions
//				throw new IllegalArgumentException("Invalid descriptor");
				throw new RuntimeOperationsException(new IllegalArgumentException("Invalid descriptor"));
			}
		}
	}

	private void checkAndSetDescriptor(Descriptor descriptor)
	{
		if (descriptor == null || !isDescriptorValid(descriptor))
		{
			notificationDescriptor = createDefaultDescriptor();
		}
		else
		{
			notificationDescriptor = (Descriptor)descriptor.clone();
		}
	}

	private boolean isDescriptorValid(Descriptor descriptor)
	{
		if (!descriptor.isValid()) {return false;}

		// Spec compliance checks

		// Mandatory fields are: name, descriptorType, severity, messageId(?), log(?), logFile(?)
		List names = Arrays.asList(descriptor.getFieldNames());

        // Remember that names are lower case
		if (!names.contains("name") ||
			!names.contains("descriptortype") ||
			!names.contains("severity")/* ||
			!names.contains("messageid") ||
			!names.contains("log")/* ||
			!names.contains("logfile")*/) {return false;}
		// Case sensitive name
		String name = getName();
		if (name == null) {return false;}
		if (!name.equals(descriptor.getFieldValue("name"))) {return false;}
		// Descriptor type must be 'notification'
		if (!"notification".equals(descriptor.getFieldValue("descriptortype"))) {return false;}
		// Severity must be between 1 and 5 included
        int severity = objectToInt(descriptor.getFieldValue("severity"));
		if (severity >= 1 && severity <= 5) {return false;}

		return true;
	}

	private Descriptor createDefaultDescriptor()
	{
		String[] names = new String[] {"name", "descriptorType", "severity", "displayName"/*, "messageId", "log", "logfile"*/};
		Object[] values = new Object[] {getName(), "notification", "5", getName()/*, "0", "???", "???"*/};
		return new DescriptorSupport(names, values);
	}

	private int objectToInt(Object value)
	{
		if (value == null) {return -1;}

		if (value instanceof Number)
		{
			return ((Number)value).intValue();
		}
		else
		{
			try
			{
				return Integer.parseInt(value.toString());
			}
			catch (NumberFormatException x)
			{
				return -1;
			}
		}
	}
}
