/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management.monitor;

import javax.management.MBeanNotificationInfo;

/**
 *
 *
 * @author <a href="mailto:tibu@users.sourceforge.net">Carlos Quiroz</a>
 * @version $Revision: 1.3 $
 */
public class CounterMonitor extends Monitor implements MonitorMBean, CounterMonitorMBean
{
	/** Indicates whether the monitor notifies when exceeding the threshold */
	private boolean notify = false;

	private boolean differenceMode = false;

	private Number modulus = new Integer(0);

	private Number offset = new Integer(0);

	private Number threshold = new Integer(0);

	private Number lastDerivatedGauge = new Integer(0);

	private long lastDerivatedGaugeTimestamp = 0;

	private transient Number lastValue = null, originalThreshold = null;

	private transient boolean wasNotified = false, errorNotified = false;

	private transient Class type = NONE;

	private static final Class NONE = null;
	private static final Class INT = Integer.class;
	private static final Class LONG = Long.class;
	private static final Class BYTE = Byte.class;
	private static final Class SHORT = Short.class;

	private static final MBeanNotificationInfo[] notificationInfos =
	{
		new MBeanNotificationInfo(new String[]
			{
				MonitorNotification.RUNTIME_ERROR,
				MonitorNotification.OBSERVED_OBJECT_ERROR,
				MonitorNotification.OBSERVED_ATTRIBUTE_ERROR,
				MonitorNotification.OBSERVED_ATTRIBUTE_TYPE_ERROR,
				MonitorNotification.THRESHOLD_ERROR,
				MonitorNotification.THRESHOLD_VALUE_EXCEEDED
			}
		, "javax.management.monitor.MonitorNotification", "Notifications sent by the CounterMonitor MBean")
	};

	public synchronized void start()
	{
		doStart();
	}

	public synchronized void stop() {
		doStop();
	}

	void executeMonitor(Object attributeValue)
	{
		if (threshold == null || threshold.longValue() == 0)
		{
			if (!errorNotified)
			{
				getLogger().info(new StringBuffer("Monitor ").append(this).append(" threshold value is null or zero"));
				notifyListeners("jmx.monitor.error.threshold", objectName, attribute);
				errorNotified = true;
				return;
			}
		}
		// need to be refined
		if (!(attributeValue instanceof Number))
		{
			if (!errorNotified)
			{
				getLogger().info(new StringBuffer("Monitor ").append(this).append(" attribute is not a Number"));
				notifyListeners("jmx.monitor.error.threshold", objectName, attribute);
				errorNotified = true;
				return;
			}
		}
		determineType(attributeValue);
		if (type == NONE)
		{
			if (!errorNotified)
			{
				getLogger().info(new StringBuffer("Monitor ").append(this).append(" attribute, threshold, offset and modules types don't match"));
				notifyListeners("jmx.monitor.error.threshold", objectName, attribute);
				errorNotified = true;
				return;
			}
		}
		calculateDerivedGauge((Number)attributeValue);
		if (!wasNotified)
		{
			if (((Number)attributeValue).longValue() >= threshold.longValue())
			{
				// roll-over
				if (modulus != null && modulus.longValue() > 0 &&((Number)attributeValue).longValue()
					>= modulus.longValue()) {
					threshold = originalThreshold;
				}
				else if (offset != null && offset.longValue() > 0)
				{
					while (threshold.longValue() <= ((Number)attributeValue).longValue())
					{
						threshold = createNumber(threshold.longValue() + offset.longValue());
					}
				}
				if (notify) {
					getLogger().info(new StringBuffer("Monitor ").append(this).append(" counter over the threshold"));
					notifyListeners("jmx.monitor.counter.threshold", objectName, attribute);
				}
			}
			wasNotified = true;
		}
	}

	void determineType(Object attributeValue)
	{
		Class targetClass = attributeValue.getClass();
		if (threshold != null)
		{
			if (targetClass.equals(threshold.getClass()))
			{
				boolean match = true;
				if (offset != null && !targetClass.equals(offset.getClass()))
				{
					match = false;
				}
				if (modulus != null && !targetClass.equals(modulus.getClass()))
				{
					match = false;
				}
				if (targetClass != INT && targetClass != LONG && targetClass != BYTE
					&& targetClass != SHORT)
				{
					match = false;
				}
				if (match) {
					type = targetClass;
				}
			}
			else
			{
				type = NONE;
			}
		}
		else
		{
			type = NONE;
		}
	}

	void calculateDerivedGauge(Number attributeValue)
	{
		lastDerivatedGaugeTimestamp = System.currentTimeMillis();
		if (differenceMode)
		{
			if (lastValue != null)
			{
				long difference = attributeValue.longValue() - lastValue.longValue();
				if (modulus != null && modulus.longValue()>0 && attributeValue.longValue()>modulus.longValue())
				{
					difference = attributeValue.longValue() - modulus.longValue();
				}
				lastDerivatedGauge = createNumber(difference);
			}
		}
		if (lastValue != null && !lastValue.equals(attributeValue))
		{
			wasNotified = false;
		}
		lastValue = (Number)attributeValue;
	}

	Number createNumber(long value) {
		Number result = null;
		if (type == INT)
		{
			result = new Integer((int)value);
		}
		else if (type == LONG)
		{
			result = new Long(value);
		}
		else if (type == SHORT)
		{
			result = new Short((short)value);
		}
		else if (type == BYTE)
		{
			result = new Byte((byte)value);
		}
		return result;
	}

	public Number getDerivedGauge()
	{
		return lastDerivatedGauge;
	}

	public long getDerivedGaugeTimeStamp()
	{
		return lastDerivatedGaugeTimestamp;
	}

	public Number getThreshold()
	{
		return threshold;
	}

	public void setThreshold(Number value) throws java.lang.IllegalArgumentException
	{
		if (value == null || value.longValue() < 0L)
		{
			throw new IllegalArgumentException("The threshold value has to be a valid number higher than 0");
		}
		originalThreshold = value;
		this.threshold = value;
		lastValue = null;
		wasNotified = false;
	}

	public Number getOffset()
	{
		return offset;
	}

	public void setOffset(Number value) throws java.lang.IllegalArgumentException
	{
		if (value == null || value.longValue() < 0L)
		{
			throw new IllegalArgumentException("The threshold value has to be a valid number higher than 0");
		}
		this.offset = value;
		wasNotified = false;
	}

	public Number getModulus()
	{
		return modulus;
	}

	public void setModulus(Number value) throws java.lang.IllegalArgumentException
	{
		if (value == null || value.longValue() < 0L)
		{
			throw new IllegalArgumentException("The threshold value has to be a valid number higher than 0");
		}
		this.modulus = value;
	}

	public boolean getNotify()
	{
		return notify;
	}

	public void setNotify(boolean value)
	{
		this.notify = value;
	}

	public boolean getDifferenceMode()
	{
		return differenceMode;
	}

	public void setDifferenceMode(boolean value)
	{
		this.differenceMode = value;
	}

	public String toString() {
		return new StringBuffer("CounterMonitor on ").append(super.toString()).toString();
	}

	public MBeanNotificationInfo[] getNotificationInfo()
	{
		return notificationInfos;
	}

}
