/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management.monitor;

import javax.management.MBeanNotificationInfo;

/**
 *
 *
 * @author <a href="mailto:tibu@users.sourceforge.net">Carlos Quiroz</a>
 * @version $Revision: 1.3 $
 */
public class GaugeMonitor extends Monitor
	implements MonitorMBean, GaugeMonitorMBean
{
	private Number highThreshold = new Integer(0);

	private Number lowThreshold = new Integer(0);

	private boolean notifyHigh = false, notifyLow = false;

	private boolean differenceMode = false;

	private transient boolean highNotified = false, lowNotified = false, errorNotified = false;

	private Number lastDerivatedGauge = new Integer(0);

	private long lastDerivatedGaugeTimestamp = 0;

	private transient Number lastValue = null;

	private transient Class type = NONE;

	private static final Class NONE = null;
	private static final Class INT = Integer.class;
	private static final Class LONG = Long.class;
	private static final Class BYTE = Byte.class;
	private static final Class SHORT = Short.class;
	private static final Class FLOAT = Float.class;
	private static final Class DOUBLE = Double.class;
	private transient boolean isLong = false;

	private static final MBeanNotificationInfo[] notificationInfos =
	{
		new MBeanNotificationInfo(new String[]
			{
				MonitorNotification.RUNTIME_ERROR,
				MonitorNotification.OBSERVED_OBJECT_ERROR,
				MonitorNotification.OBSERVED_ATTRIBUTE_ERROR,
				MonitorNotification.OBSERVED_ATTRIBUTE_TYPE_ERROR,
				MonitorNotification.THRESHOLD_ERROR,
				MonitorNotification.THRESHOLD_HIGH_VALUE_EXCEEDED,
				MonitorNotification.THRESHOLD_LOW_VALUE_EXCEEDED
			}
		, "javax.management.monitor.MonitorNotification", "Notifications sent by the GaugeMonitor MBean")
	};

	public synchronized void start()
	{
		doStart();
	}

	public synchronized void stop() {
		doStop();
	}

	void executeMonitor(Object attributeValue) {
		if (highThreshold == null || highThreshold.longValue() == 0)
		{
			if (!errorNotified)
			{
				getLogger().info(new StringBuffer("Monitor ").append(this).append(" threshold value is null or zero"));
				notifyListeners("jmx.monitor.error.threshold", objectName, attribute);
				errorNotified = true;
				return;
			}
		}
		if (lowThreshold == null || lowThreshold.longValue() == 0)
		{
			if (!errorNotified)
			{
				getLogger().info(new StringBuffer("Monitor ").append(this).append(" threshold value is null or zero"));
				notifyListeners("jmx.monitor.error.threshold", objectName, attribute);
				errorNotified = true;
				return;
			}
		}
		// need to be refined
		if (!(attributeValue instanceof Number))
		{
			if (!errorNotified)
			{
				getLogger().info(new StringBuffer("Monitor ").append(this).append(" attribute is not a Number"));
				notifyListeners("jmx.monitor.error.threshold", objectName, attribute);
				errorNotified = true;
				return;
			}
		}
		determineType(attributeValue);
		if (type == NONE)
		{
			if (!errorNotified)
			{
				getLogger().info(new StringBuffer("Monitor ").append(this).append(" attribute, threshold, offset and modules types don't match"));
				notifyListeners("jmx.monitor.error.threshold", objectName, attribute);
				errorNotified = true;
				return;
			}
		}
		calculateDerivedGauge((Number)attributeValue);
		if (isLong)
		{
			if (((Number)attributeValue).longValue() >= highThreshold.longValue())
			{
				if (!highNotified && notifyHigh)
				{
					getLogger().info(new StringBuffer("Monitor ").append(this).append(" counter over the threshold"));
					notifyListeners("jmx.monitor.gauge.high", objectName, attribute);
					highNotified = true;
					// hystersis
					lowNotified = false;
				}
			}
			if (((Number)attributeValue).longValue() <= lowThreshold.longValue())
			{
				if (!lowNotified && notifyLow)
				{
					getLogger().info(new StringBuffer("Monitor ").append(this).append(" counter over the threshold"));
					notifyListeners("jmx.monitor.gauge.low", objectName, attribute);
					lowNotified = true;
					// hystersis
					highNotified = false;
				}
			}
		}
		else
		{
			if (((Number)attributeValue).doubleValue() >= highThreshold.doubleValue())
			{
				if (!highNotified && notifyHigh)
				{
					getLogger().info(new StringBuffer("Monitor ").append(this).append(" counter over the threshold"));
					notifyListeners("jmx.monitor.gauge.high", objectName, attribute);
					highNotified = true;
					// hystersis
					lowNotified = false;
				}
			}
			if (((Number)attributeValue).doubleValue() <= lowThreshold.doubleValue())
			{
				if (!lowNotified && notifyLow)
				{
					getLogger().info(new StringBuffer("Monitor ").append(this).append(" counter over the threshold"));
					notifyListeners("jmx.monitor.gauge.low", objectName, attribute);
					lowNotified = true;
					// hystersis
					highNotified = false;
				}
			}
		}
	}

	void determineType(Object attributeValue)
	{
		Class targetClass = attributeValue.getClass();
		if (highThreshold != null && lowThreshold != null)
		{
			if (highThreshold.getClass().equals(lowThreshold.getClass()) && highThreshold.getClass().equals(targetClass))
			{
				boolean match = true;
				if (targetClass != INT && targetClass != LONG && targetClass != BYTE
					&& targetClass != SHORT && targetClass != FLOAT && targetClass != DOUBLE)
				{
					match = false;
				}
				if (match)
				{
					type = targetClass;
					if (targetClass.equals(FLOAT) || targetClass.equals(DOUBLE))
					{
						isLong = false;
					}
					else
					{
						isLong = true;
					}
				}
			}
			else
			{
				type = NONE;
			}
		}
		else
		{
			type = NONE;
		}
	}

	void calculateDerivedGauge(Number attributeValue)
	{
		lastDerivatedGaugeTimestamp = System.currentTimeMillis();
		if (differenceMode)
		{
			if (lastValue != null)
			{
				if (isLong)
				{
					long difference = attributeValue.longValue() - lastValue.longValue();
					lastDerivatedGauge = createNumber(difference);
				}
				else
				{
					double difference = attributeValue.doubleValue() - lastValue.doubleValue();
					lastDerivatedGauge = createNumber(difference);
				}
			}
		}
		lastValue = (Number)attributeValue;
	}

	Number createNumber(long value)
	{
		Number result = null;
		if (type == INT)
		{
			result = new Integer((int)value);
		}
		else if (type == LONG)
		{
			result = new Long(value);
		}
		else if (type == SHORT)
		{
			result = new Short((short)value);
		}
		else if (type == BYTE)
		{
			result = new Byte((byte)value);
		}
		return result;
	}

	Number createNumber(double value)
	{
		Number result = null;
		if (type == FLOAT)
		{
			result = new Float((float)value);
		}
		else if (type == DOUBLE)
		{
			result = new Double(value);
		}
		return result;
	}

	public Number getDerivedGauge()
	{
		return lastDerivatedGauge;
	}

	public long getDerivedGaugeTimeStamp()
	{
		return lastDerivatedGaugeTimestamp;
	}

	public Number getHighThreshold()
	{
		return highThreshold;
	}

	public Number getLowThreshold()
	{
		return lowThreshold;
	}

	public void setThresholds(Number highValue, Number lowValue) throws java.lang.IllegalArgumentException
	{
		if (highValue == null || lowValue == null)
		{
			throw new IllegalArgumentException("Threshold values cannot be null");
		}
		if (!highValue.getClass().equals(lowValue.getClass()))
		{
			throw new IllegalArgumentException("Threshold need to be of the same class");
		}
		if (highValue.doubleValue() < lowValue.doubleValue())
		{
			throw new IllegalArgumentException("Threshold need to be of the same class");
		}
		highThreshold = highValue;
		lowThreshold = lowValue;
	}

	public boolean getNotifyHigh()
	{
		return notifyHigh;
	}

	public void setNotifyHigh(boolean value)
	{
		this.notifyHigh = value;
	}

	public boolean getNotifyLow()
	{
		return notifyLow;
	}

	public void setNotifyLow(boolean value)
	{
		this.notifyLow = value;
	}

	public boolean getDifferenceMode()
	{
		return differenceMode;
	}

	public void setDifferenceMode(boolean value)
	{
		this.differenceMode = value;
	}

	public MBeanNotificationInfo[] getNotificationInfo()
	{
		return notificationInfos;
	}

	public String toString()
	{
		return new StringBuffer("GaugeMonitor on ").append(super.toString()).toString();
	}
}
