/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management.monitor;

import javax.management.ObjectName;
import javax.management.NotificationBroadcasterSupport;
import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.MBeanInfo;
import javax.management.MBeanAttributeInfo;
import java.util.Timer;
import java.util.TimerTask;

import mx4j.log.Logger;
import mx4j.log.Log;

/**
 * Class monitor. Parent class of all Monitoring classes
 *
 * @see MonitorMBean
 * @author <a href="mailto:tibu@users.sourceforge.net">Carlos Quiroz</a>
 * @version $Revision: 1.7 $
 */
public abstract class Monitor extends NotificationBroadcasterSupport
	implements MonitorMBean, MBeanRegistration
{
	/** Object name of the monitored MBean */
	ObjectName objectName;

	/** Monitored attribute */
	String attribute;

	/** Granularity of the monitoring in milliseconds */
	long granularity = 10000;

	/** Indicates whether the Monitor is active */
	boolean isActive = false;

	transient boolean errorNotified = false, mBeanFound = true;

	MBeanServer server;

	static Timer notificationTimer = new Timer();

	static long notificationID = 0;

	MonitorTask monitorTask = null;

	static synchronized long createNotificationID()
	{
		return ++notificationID;
	}

	abstract void executeMonitor(Object attribute);

	Logger getLogger()
	{
		return Log.getLogger(getClass().getName());
	}

	private class MonitorTask extends TimerTask {
		public void run() {
			try {
				if (isActive())
				{
					if (objectName == null) {
						getLogger().warn(new StringBuffer("Monitor ").append(Monitor.this.toString()).append(" object name ").append(objectName).append(" not found").toString());
						if (!errorNotified) {
							notifyListeners("jmx.monitor.error.mbean", objectName);
						}
						errorNotified = true;
						return;
					}
					getLogger().info("Execute monitor " + Monitor.this.toString());
					if (!server.isRegistered(objectName)) {
						getLogger().warn(new StringBuffer("Monitor ").append(Monitor.this.toString()).append(" object name ").append(objectName).append(" not found").toString());
						if (mBeanFound) {
							mBeanFound = false;
							notifyListeners("jmx.monitor.error.mbean", objectName);
						}
						return;
					}
					mBeanFound = true;
					if (errorNotified) {
						return;
					}
					MBeanInfo info = server.getMBeanInfo(objectName);
					MBeanAttributeInfo[] attributes =info.getAttributes();
					boolean found = false;
					if (attributes != null)
					{
						for (int i=0;i<attributes.length;i++) {
							if (attributes[i].getName().equals(attribute)) {
								found = true;
							}
						}
					}
					if (!found) {
						getLogger().warn(new StringBuffer("Monitor ").append(Monitor.this.toString()).append(" attribute ").append(attribute).append(" not found").toString());
						errorNotified = true;
						notifyListeners("jmx.monitor.error.attribute", objectName, attribute);
						return;
					}
					Object attributeValue = server.getAttribute(objectName, attribute);
					if (attributeValue == null) {
						getLogger().warn(new StringBuffer("Monitor ").append(Monitor.this.toString()).append(" attribute ").append(attribute).append(" is null").toString());
						errorNotified = true;
						notifyListeners("jmx.monitor.error.type", objectName, attribute);
						return;
					}
					executeMonitor(attributeValue);
				}
			} catch (Exception e) {
				getLogger().warn(new StringBuffer("Monitor ").append(Monitor.this.toString()).append(" object name ").append(objectName).append(" not found").toString());
				errorNotified = true;
				notifyListeners("jmx.monitor.error.runtime", objectName, attribute, e);
			}
		}
	}

	public abstract void start();

	synchronized void doStart()
	{
		if (!isActive) {
			getLogger().info(new StringBuffer("Starting monitor ").append(this).append(" every ").append(granularity).append(" milliseconds").toString());
			this.isActive = true;
			errorNotified = false;
			// updated to create a new MonitorTask every time
			monitorTask = new MonitorTask();
			notificationTimer.scheduleAtFixedRate(monitorTask, 0, granularity);
		}
	}

	public abstract void stop();

	synchronized void doStop()
	{
		if (isActive) {
			getLogger().info("Stopping monitor " + toString());
			this.isActive = false;
			monitorTask.cancel();
		}
	}

	public ObjectName getObservedObject() {
		return objectName;
	}

	public synchronized void setObservedObject(ObjectName objectName)  throws java.lang.IllegalArgumentException {
		if (objectName == null) {
			throw new IllegalArgumentException("The observed object name cannot be null");
		}
		errorNotified = false;
		this.objectName = objectName;
	}

	public String getObservedAttribute() {
		return attribute;
	}

	public synchronized void setObservedAttribute(String attribute) throws java.lang.IllegalArgumentException {
		if (attribute == null) {
			throw new IllegalArgumentException("The observed attribute cannot be null");
		}
		errorNotified = true;
		this.attribute = attribute;
	}

	public long getGranularityPeriod() {
		return granularity;
	}

	public void setGranularityPeriod(long period) throws java.lang.IllegalArgumentException {
		if (period <=0) {
			throw new IllegalArgumentException("The monitoring period can't be negative or zero");
		}
		this.granularity = period;
	}

	public synchronized boolean isActive() {
		return isActive;
	}

	public ObjectName preRegister(MBeanServer server, ObjectName name) throws Exception {
		getLogger().info("Pre register of monitor " + toString());
		this.server = server;
		errorNotified = false;
		return name;
	}

	public void postRegister(Boolean registrationDone) {}

	public void preDeregister() throws Exception {
		getLogger().info("Pre deregister of monitor " + toString());
		this.stop();
	}

	public void postDeregister() {}

	void notifyListeners(String type, ObjectName objectName) {
		MonitorNotification not = new MonitorNotification(type, this, createNotificationID(), System.currentTimeMillis(), "", objectName, null, null, null);
		sendNotification(not);
	}

	void notifyListeners(String type, ObjectName objectName, String attribute) {
		MonitorNotification not = new MonitorNotification(type, this, createNotificationID()
			, System.currentTimeMillis(), "", objectName, attribute, null, null);
		sendNotification(not);
	}

	void notifyListeners(String type, ObjectName objectName, String attribute, Throwable e) {
		MonitorNotification not = new MonitorNotification(type, this, createNotificationID()
			, System.currentTimeMillis(), e.toString(), objectName, attribute, null, null);
		sendNotification(not);
	}

	public String toString() {
		return new StringBuffer(objectName != null?objectName.toString():"?").append(" attribute ").append(attribute).toString();
	}
}
