/**
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management.openmbean;

import java.io.ObjectStreamException;
import java.io.Serializable;

/**
 * <p>The SimpleType class is the open type class whose instances describe all open data values which are neither arrays,
 * nor {@link CompositeData} values, nor {@link TabularData} values. It predefines all its possible instances as static fields,
 * and has no public constructor.</p>
 *
 * <p>Given a SimpleType instance describing values whose Java class name is className, the internal fields corresponding to the
 * TypeName and description of this SimpleType instance are also set to className. In other words, its methods getClassName, getTypeName
 * and getDescription all return the same string value className.</p>
 *
 * @author <a href="mailto:shadow12@users.sourceforge.net">Bronwen Cassidy</a>
 * @version $Revision: 1.8 $
 */
public class SimpleType extends OpenType implements Serializable
{
	private static final long serialVersionUID = 2215577471957694503L;

	// No non-transient fields allowed

	public static SimpleType BIGDECIMAL;
	public static SimpleType BIGINTEGER;
	public static SimpleType BOOLEAN;
	public static SimpleType BYTE;
	public static SimpleType CHARACTER;
	public static SimpleType DOUBLE;
	public static SimpleType FLOAT;
	public static SimpleType INTEGER;
	public static SimpleType LONG;
	public static SimpleType OBJECTNAME;
	public static SimpleType SHORT;
	public static SimpleType STRING;
	public static SimpleType VOID;

	private transient int m_hashCode = 0;

	static
	{
		try
		{
			SimpleType.BIGDECIMAL = new SimpleType("java.math.BigDecimal");
			SimpleType.BIGINTEGER = new SimpleType("java.math.BigInteger");
			SimpleType.BOOLEAN = new SimpleType("java.lang.Boolean");
			SimpleType.BYTE = new SimpleType("java.lang.Byte");
			SimpleType.CHARACTER = new SimpleType("java.lang.Character");
			SimpleType.DOUBLE = new SimpleType("java.lang.Double");
			SimpleType.FLOAT = new SimpleType("java.lang.Float");
			SimpleType.INTEGER = new SimpleType("java.lang.Integer");
			SimpleType.LONG = new SimpleType("java.lang.Long");
			SimpleType.OBJECTNAME = new SimpleType("javax.management.ObjectName");
			SimpleType.SHORT = new SimpleType("java.lang.Short");
			SimpleType.STRING = new SimpleType("java.lang.String");
			SimpleType.VOID = new SimpleType("java.lang.Void");
		}
		catch (OpenDataException ignored)
		{
			// as these are predefined data types there is no way they will be invalid hence this catch is ignored
		}
	}

	private SimpleType(String className) throws OpenDataException
	{
		super(className, className, className);
	}


	/**
	 * Checks if this <code>SimpleType</code> object is value of
	 * the given object
	 *
	 * @param object The object to check
	 * @return boolean
	 *
	 */
	public boolean isValue(Object object)
	{
		if (object == null) return false;
		return getClassName().equals(object.getClass().getName());
	}

	public Object readResolve() throws ObjectStreamException
	{
		//TODO: need a better way of doing this
		if ( getClassName().equals(String.class.getName()) )
			return SimpleType.STRING;
		if ( getClassName().equals(java.math.BigDecimal.class.getName()) )
			return SimpleType.BIGDECIMAL;
		if ( getClassName().equals(java.math.BigInteger.class.getName()) )
			return SimpleType.BIGINTEGER;
		if ( getClassName().equals(Boolean.class.getName()) )
			return SimpleType.BOOLEAN;
		if ( getClassName().equals(Byte.class.getName()) )
			return SimpleType.BYTE;
		if ( getClassName().equals(Character.class.getName()) )
			return SimpleType.CHARACTER;
		if ( getClassName().equals(Double.class.getName()) )
			return SimpleType.DOUBLE;
		if ( getClassName().equals(Float.class.getName()) )
			return SimpleType.FLOAT;
		if ( getClassName().equals(Integer.class.getName()) )
			return SimpleType.INTEGER;
		if ( getClassName().equals(Long.class.getName()) )
			return SimpleType.LONG;
		if ( getClassName().equals(javax.management.ObjectName.class.getName()) )
			return SimpleType.OBJECTNAME;
		if ( getClassName().equals(Short.class.getName()) )
			return SimpleType.SHORT;
		if ( getClassName().equals(Void.class.getName()) )
			return SimpleType.VOID;

		return null;
	}

	/**
	 * Check the given object for equality
	 *
	 * @return boolean if object is equal
	 *
	 */
	public boolean equals(Object object)
	{
		if (!(object instanceof SimpleType)) return false;
		SimpleType otherType = (SimpleType)object;
		return (this.getClassName().equals(otherType.getClassName()));
	}

	/**
	 * Retrieve the hashCode
	 *
	 * @return int The computed hasCode
	 *
	 */
	public int hashCode()
	{
		if (m_hashCode == 0)
		{
			int result = 17;
			result = (3 * 37 * result + getClassName().hashCode());
			m_hashCode = result;
		}
		return m_hashCode;
	}

	/**
	 * Returns a human readable representation of this SimpleType object
	 *
	 * @return String the String representation
	 *
	 */
	public String toString()
	{
		return (getClass().getName() + "(name = " + getTypeName() + ")");
	}
}
