/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management.timer;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import javax.management.InstanceNotFoundException;
import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.NotificationBroadcasterSupport;
import javax.management.ObjectName;
import javax.management.Notification;

import mx4j.log.Log;
import mx4j.log.Logger;
import mx4j.timer.TimerTask;
import mx4j.util.TimeQueue;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.7 $
 */
public class Timer extends NotificationBroadcasterSupport implements TimerMBean, MBeanRegistration
{
	public static final long ONE_SECOND = 1000;
	public static final long ONE_MINUTE = 60 * ONE_SECOND;
	public static final long ONE_HOUR = 60 * ONE_MINUTE;
	public static final long ONE_DAY = 24 * ONE_HOUR;
	public static final long ONE_WEEK = 7 * ONE_DAY;

	private TimeQueue m_queue = new TimeQueue();
	private boolean m_isActive;
	private int m_notificationID;
	private HashMap m_tasks = new HashMap();
	private boolean m_sendPastNotification;
	private ObjectName m_name;

	public Timer()
	{
	}

	private Logger getLogger()
	{
		return Log.getLogger(getClass().getName());
	}

	public ObjectName preRegister(MBeanServer server, ObjectName name) throws Exception
	{
		Logger logger = getLogger();
		m_name = name == null ? new ObjectName(server.getDefaultDomain(), "service", "Timer") : name;
		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Timer service " + m_name + " preRegistered successfully");
		return m_name;
	}

	public void postRegister(Boolean registrationDone)
	{
		Logger logger = getLogger();
		boolean done = registrationDone.booleanValue();
		if (!done)
		{
			logger.warn("Timer service " + m_name + " was not registered");
		}
		else
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Timer service " + m_name + " postRegistered successfully.");
		}
	}

	public void preDeregister() throws Exception
	{
		Logger logger = getLogger();
		stop();
		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Timer service " + m_name + " preDeregistered successfully");
	}

	public void postDeregister()
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Timer service " + m_name + " postDeregistered successfully");
	}

	public void start()
	{
		if (!isActive())
		{
			Logger logger = getLogger();
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Starting Timer service " + m_name);

			m_queue.clear();
			m_queue.start();

			ArrayList tasks = updateTasks();
			scheduleTasks(tasks);

			m_isActive = true;

			if (logger.isEnabledFor(Logger.TRACE))
			{
				logger.trace("Timer service " + m_name + " started successfully");
			}
		}
	}

	public void stop()
	{
		if (isActive())
		{
			Logger logger = getLogger();
			if (logger.isEnabledFor(Logger.TRACE))
			{
				logger.trace("Stopping Timer service " + m_name);
			}

			m_queue.stop();
			m_queue.clear();

			m_isActive = false;

			if (logger.isEnabledFor(Logger.TRACE))
			{
				logger.trace("Timer service " + m_name + " stopped successfully");
			}
		}
	}

	public boolean isActive()
	{
		return m_isActive;
	}

	public Integer addNotification(String type, String message, Object userData, Date date) throws IllegalArgumentException
	{
		return addNotification(type, message, userData, date, 0, 0);
	}

	public Integer addNotification(String type, String message, Object userData, Date date, long period) throws IllegalArgumentException
	{
		return addNotification(type, message, userData, date, period, 0);
	}

	public Integer addNotification(String type, String message, Object userData, Date date, long period, long occurrences) throws IllegalArgumentException
	{
		if (date == null) throw new IllegalArgumentException("Notification date cannot be null");
		if (period < 0) throw new IllegalArgumentException("Period cannot be negative");
		if (occurrences < 0) throw new IllegalArgumentException("Occurrences cannot be negative");

		long now = System.currentTimeMillis();

		if (isActive())
		{
			// Check for the validity of the notification times, as the Timer is active

			// One-shot notification in the past
			if (period == 0 && date.getTime() < now) throw new IllegalArgumentException("Notification date cannot be before current date");

			// Periodic limited, the last notification is in the past
			if (period > 0 && occurrences > 0)
			{
				long lastTime = date.getTime() + (occurrences - 1) * period;
				if (lastTime < now) throw new IllegalArgumentException("Notification date cannot be before current date");
			}
		}

		// Anyway, register the notification, no matter if the Timer is active
		Integer id = addNotificationImpl(type, message, userData, date, period, occurrences);

		// If the Timer is active, schedule the notification
		if (isActive())
		{
			TimerTask task = getTask(id);
			updateTask(task);
			if (!task.isFinished())
			{
				m_queue.schedule(task);
			}
		}
		return id;
	}

	private Integer addNotificationImpl(String type, String message, Object userData, Date date, long period, long occurrences)
	{
		Logger logger = getLogger();

		Integer id = createNotificationID();

		TimerNotification notification = new TimerNotification(type, this, message, id);
		notification.setUserData(userData);
		if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Adding timer notification: " + notification + " on Timer service " + m_name);

		TimerTask task = createTimerTask(notification, date, period, occurrences);

		synchronized (this)
		{
			m_tasks.put(id, task);
		}

		return id;
	}

	private TimerTask createTimerTask(TimerNotification notification, Date date, long period, long occurrences)
	{
		return new TimerTask(notification, date, period, occurrences)
		{
			public void run()
			{
				// Send the notification
				sendNotification(getNotification());
			}
		};
	}

	private ArrayList updateTasks()
	{
		ArrayList list = new ArrayList();
		boolean sendPast = getSendPastNotifications();

		synchronized (this)
		{
			for (Iterator i = m_tasks.entrySet().iterator(); i.hasNext();)
			{
				Map.Entry entry = getNextNonFinishedTaskEntry(i);
				if (entry == null) break;

				TimerTask task = (TimerTask)entry.getValue();

				if (!sendPast)
				{
					updateTask(task);
					if (task.isFinished()) continue;
				}
				list.add(task);
			}
			return list;
		}
	}

	private void updateTask(TimerTask task)
	{
		long now = System.currentTimeMillis();
		long time = task.getNextExecutionTime();
		while (time < now && !task.isFinished())
		{
			if (task.isPeriodic())
			{
				task.setNextExecutionTime(time + task.getPeriod());
				time = task.getNextExecutionTime();
			}
			else
			{
				task.setFinished(true);
			}
		}
	}

	private void scheduleTasks(ArrayList tasks)
	{
		synchronized (this)
		{
			for (int i = 0; i < tasks.size(); ++i)
			{
				TimerTask task = (TimerTask)tasks.get(i);
				m_queue.schedule(task);
			}
		}
	}

	public void removeNotification(Integer id) throws InstanceNotFoundException
	{
		Logger logger = getLogger();

		synchronized (this)
		{
			TimerTask t = getTask(id);
			if (t == null) throw new InstanceNotFoundException("Cannot find notification to remove with id: " + id);
			m_queue.unschedule(t);
			m_tasks.remove(id);
			if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Notification " + t.getNotification() + " removed successfully from Timer service " + m_name);
		}
	}

	public void removeNotifications(String type) throws InstanceNotFoundException
	{
		Logger logger = getLogger();

		boolean found = false;
		synchronized (this)
		{
			for (Iterator i = m_tasks.entrySet().iterator(); i.hasNext();)
			{
				Map.Entry entry = getNextNonFinishedTaskEntry(i);
				if (entry == null) break;

				TimerTask t = (TimerTask)entry.getValue();
				TimerNotification n = t.getNotification();
				if (n.getType().equals(type))
				{
					m_queue.unschedule(t);
					m_tasks.remove(entry.getKey());
					if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Notification " + n + " removed successfully from Timer service " + m_name);
					found = true;
				}
			}
		}

		if (!found) throw new InstanceNotFoundException("Cannot find timer notification to remove with type: " + type + " from Timer service " + m_name);
	}

	public void removeAllNotifications()
	{
		synchronized (this)
		{
			m_queue.clear();
			m_tasks.clear();
			m_notificationID = 0;
		}
	}

	public Vector getAllNotificationIDs()
	{
		Vector vector = new Vector();
		synchronized (this)
		{
			for (Iterator i = m_tasks.entrySet().iterator(); i.hasNext();)
			{
				Map.Entry entry = getNextNonFinishedTaskEntry(i);
				if (entry == null) break;
				vector.add(entry.getKey());
			}
		}
		return vector;
	}

	public Vector getNotificationIDs(String type)
	{
		Vector vector = new Vector();
		synchronized (this)
		{
			for (Iterator i = m_tasks.entrySet().iterator(); i.hasNext();)
			{
				Map.Entry entry = getNextNonFinishedTaskEntry(i);
				if (entry == null) break;
				TimerTask t = (TimerTask)entry.getValue();
				TimerNotification n = t.getNotification();
				if (n.getType().equals(type))
				{
					vector.add(entry.getKey());
				}
			}
		}
		return vector;
	}

	public boolean getSendPastNotifications()
	{
		return m_sendPastNotification;
	}

	public void setSendPastNotifications(boolean value)
	{
		m_sendPastNotification = value;
	}

	public int getNbNotifications()
	{
		int count = 0;
		synchronized (this)
		{
			for (Iterator i = m_tasks.entrySet().iterator(); i.hasNext();)
			{
				Map.Entry entry = getNextNonFinishedTaskEntry(i);
				if (entry == null) break;
				++count;
			}
			return count;
		}
	}

	public boolean isEmpty()
	{
		synchronized (this)
		{
			return getNbNotifications() == 0;
		}
	}

	public String getNotificationType(Integer id)
	{
		synchronized (this)
		{
			TimerTask t = getTask(id);
			return t == null ? null : t.getNotification().getType();
		}
	}

	public String getNotificationMessage(Integer id)
	{
		synchronized (this)
		{
			TimerTask t = getTask(id);
			return t == null ? null : t.getNotification().getMessage();
		}
	}

	public Object getNotificationUserData(Integer id)
	{
		synchronized (this)
		{
			TimerTask t = getTask(id);
			return t == null ? null : t.getNotification().getUserData();
		}
	}

	public Date getDate(Integer id)
	{
		synchronized (this)
		{
			TimerTask t = getTask(id);
			return t == null ? null : new Date(t.getDate());
		}
	}

	public Long getPeriod(Integer id)
	{
		synchronized (this)
		{
			TimerTask t = getTask(id);
			return t == null ? null : new Long(t.getPeriod());
		}
	}

	public Long getNbOccurences(Integer id)
	{
		synchronized (this)
		{
			TimerTask t = getTask(id);
			return t == null ? null : new Long(t.getOccurrences());
		}
	}

	private Integer createNotificationID()
	{
		// Must be sync'ed
		synchronized (this)
		{
			return new Integer(++m_notificationID);
		}
	}

	private TimerTask getTask(Integer id)
	{
		Logger logger = getLogger();

		synchronized (this)
		{
			TimerTask t = (TimerTask)m_tasks.get(id);

			if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Retrieving task with id " + id + ": " + t);

			if (t != null && t.isFinished())
			{
				if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Task with id " + id + " is expired, removing it");

				m_tasks.remove(id);
				t = null;
			}
			return t;
		}
	}

	private Map.Entry getNextNonFinishedTaskEntry(Iterator i)
	{
		Logger logger = getLogger();

		synchronized (this)
		{
			if (i.hasNext())
			{
				Map.Entry entry = (Map.Entry)i.next();
				TimerTask t = (TimerTask)entry.getValue();
				if (t.isFinished())
				{
					if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Found an expired notification, removing it: " + t);
					i.remove();
					return getNextNonFinishedTaskEntry(i);
				}
				return entry;
			}
			return null;
		}
	}

	public void sendNotification(Notification n)
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Start sending notifications from Timer service " + m_name);
		super.sendNotification(n);
		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Finished sending notifications from Timer service " + m_name);
	}
}
