/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.loading;

import java.io.ObjectInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.StreamCorruptedException;
import java.io.ObjectStreamClass;
import java.lang.reflect.Proxy;

import javax.management.MBeanServerFactory;

/**
 * Load classes with the specified classloader, or with the
 * {@link javax.management.loading.DefaultLoaderRepository DefaultLoaderRepository}
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.2 $
 */
public class ClassLoaderObjectInputStream extends ObjectInputStream
{
	private ClassLoader m_classLoader;

	/**
	 * Creates an instance of this ObjectInputStream that will load classes with the DefaultLoaderRepository
	 */
	public ClassLoaderObjectInputStream(InputStream stream) throws IOException, StreamCorruptedException
	{
		super(stream);
	}
	/**
	 * Creates an instance of this ObjectInputStream that will load classes with the specified classloader
	 */
	public ClassLoaderObjectInputStream(InputStream stream, ClassLoader classLoader) throws IOException, StreamCorruptedException
	{
		super(stream);
		m_classLoader = classLoader;
	}

	protected Class resolveClass(ObjectStreamClass osc) throws IOException, ClassNotFoundException
	{
		String name = osc.getName();
		return loadClass(name);
	}

	protected Class resolveProxyClass(String[] interfaces) throws IOException, ClassNotFoundException
	{
		Class[] classes = new Class[interfaces.length];
		for (int i = 0; i < interfaces.length; ++i) {classes[i] = loadClass(interfaces[i]);}

		if (m_classLoader != null)
		{
			return Proxy.getProxyClass(m_classLoader, classes);
 		}
		else
		{
			// I define the resulting proxy class to be in the space of the context classloader, as I may
			// load the interfaces from different classloaders, and I cannot choose the one that will define
			// the proxy class (which one ?).
			return Proxy.getProxyClass(Thread.currentThread().getContextClassLoader(), classes);
		}
	}

	private Class loadClass(String name) throws ClassNotFoundException
	{
		if (m_classLoader != null)
		{
            return m_classLoader.loadClass(name);
		}
		else
		{
            return MBeanServerFactory.getClassLoaderRepository(null).loadClass(name);
		}
	}
}
