/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedExceptionAction;
import java.security.PrivilegedActionException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.BadAttributeValueExpException;
import javax.management.BadBinaryOpValueExpException;
import javax.management.BadStringOperationException;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.InvalidApplicationException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistration;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.MBeanServerDelegate;
import javax.management.MBeanServerNotification;
import javax.management.MalformedObjectNameException;
import javax.management.NotCompliantMBeanException;
import javax.management.NotificationBroadcaster;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.OperationsException;
import javax.management.QueryExp;
import javax.management.ReflectionException;
import javax.management.RuntimeErrorException;
import javax.management.RuntimeOperationsException;
import javax.management.loading.ClassLoaderRepository;

import mx4j.ImplementationException;
import mx4j.loading.ClassLoaderObjectInputStream;
import mx4j.log.Log;
import mx4j.log.Logger;
import mx4j.server.interceptor.ContextClassLoaderMBeanServerInterceptor;
import mx4j.server.interceptor.MBeanServerInterceptor;
import mx4j.server.interceptor.MBeanServerInterceptorConfigurator;
import mx4j.server.interceptor.NotificationListenerMBeanServerInterceptor;
import mx4j.server.interceptor.SecurityMBeanServerInterceptor;
import mx4j.server.interceptor.InvokerMBeanServerInterceptor;
import mx4j.util.Utils;

/**
 * MBeanServer implementation. <p>
 * The MBeanServer accomplishes these roles:
 * <ul>
 * <li> Returns information about the Agent
 * <li> Acts as a repository for MBeans
 * <li> Acts as an invoker, on behalf of the user, on MBeans
 * </ul>
 * <br>
 * The repository function is delegated to instances of {@link MBeanRepository} classes.
 * This class acts as a factory for MBeanRepository instances, that can be controlled via the system property
 * mx4j.mbeanserver.repository set to the qualified name of the implementation class. <p>
 *
 * This class also acts as an invoker on MBeans. The architecture is interceptor-based, that is whenever you call
 * from a client an MBeanServer method that will end up to call the MBean instance, the call is dispatched to
 * the interceptor chain and eventually to the MBean. <br>
 * The interceptors are configurable via the MBean {@link MBeanServerInterceptorConfigurator}.
 * When the call is about to arrive to the MBean instance, the last interceptor dispatches the call depending on
 * the MBean type: if the MBean is a dynamic MBean, the call is dispatched directly; if the MBean is a standard
 * MBean an {@link MBeanInvoker} is delegated to invoke on the MBean instance.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.32 $
 */
public class MBeanServerImpl implements MBeanServer
{
	private String m_defaultDomain;
	private MBeanRepository m_repository;
	private MBeanServerDelegate m_delegate;
	private ObjectName m_delegateName;
	private MBeanIntrospector m_introspector;
	private MBeanServerInterceptorConfigurator m_invoker;
	private static long m_notifications;
	private BaseClassLoaderRepository m_classLoaderRepository;

	private static final String[] EMPTY_PARAMS = new String[0];
	private static final Object[] EMPTY_ARGS = new Object[0];
	private static BaseClassLoaderRepository m_staticClassLoaderRepository = new DefaultClassLoaderRepository();

	/**
	 * Create a new MBeanServerImpl with the specified default domain.
	 * Also registers interceptors to their respective reserved ObjectName
	 *
	 * @param defaultDomain The default domain to be used
	 * @throws java.lang.SecurityException If access is not granted to
	 *         create an MBeanServer instance
	 *
	 */
	public MBeanServerImpl(String defaultDomain)
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Creating MBeanServer instance...");

		SecurityManager sm = System.getSecurityManager();
		if (sm != null)
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Checking permission to create MBeanServer...");
			sm.checkPermission(new javax.management.MBeanServerPermission("newMBeanServer"));
		}

		if (defaultDomain == null) defaultDomain = "";
		m_defaultDomain = defaultDomain;
		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("MBeanServer default domain is: '" + m_defaultDomain + "'");

		m_repository = createMBeanRepository();

		m_classLoaderRepository = createClassLoaderRepository();

		m_introspector = new MBeanIntrospector();

		// This is the official name of the delegate, it is used as a source for MBeanServerNotifications
		try
		{
			m_delegateName = new ObjectName("JMImplementation", "type", "MBeanServerDelegate");
		}
		catch (MalformedObjectNameException ignored)
		{
		}

		// The delegate must be created as first since it must send notifications of registration of internal MBeans
		// We cannot register it though since the interceptor stack is not in place yet.
		m_delegate = new MBeanServerDelegate();

		try
		{
			ObjectName invokerName = new ObjectName(MBeanServerInterceptorConfigurator.OBJECT_NAME);
			m_invoker = new MBeanServerInterceptorConfigurator(this);

			ContextClassLoaderMBeanServerInterceptor ccl = new ContextClassLoaderMBeanServerInterceptor();
			NotificationListenerMBeanServerInterceptor notif = new NotificationListenerMBeanServerInterceptor();
			SecurityMBeanServerInterceptor sec = new SecurityMBeanServerInterceptor();
			InvokerMBeanServerInterceptor inv = new InvokerMBeanServerInterceptor(this);

			m_invoker.addPreInterceptor(ccl);
			m_invoker.addPreInterceptor(notif);
			m_invoker.addPostInterceptor(sec);
			m_invoker.addPostInterceptor(inv);
			m_invoker.start();

			// The interceptor stack is in place, register the configurator and all interceptors
			privilegedRegisterMBean(m_invoker, invokerName);

			ObjectName cclName = new ObjectName("JMImplementation", "interceptor", "contextclassloader");
			ObjectName notifName = new ObjectName("JMImplementation", "interceptor", "notificationwrapper");
			ObjectName secName = new ObjectName("JMImplementation", "interceptor", "security");
			ObjectName invName = new ObjectName("JMImplementation", "interceptor", "invoker");

			privilegedRegisterMBean(ccl, cclName);
			privilegedRegisterMBean(notif, notifName);
			privilegedRegisterMBean(sec, secName);
			privilegedRegisterMBean(inv, invName);
		}
		catch (Exception x)
		{
			logger.error("MBeanServerInterceptorConfigurator cannot be registered", x);
			throw new ImplementationException();
		}

		// Now register the delegate
		try
		{
			privilegedRegisterMBean(m_delegate, m_delegateName);
		}
		catch (Exception x)
		{
			logger.error("MBeanServerDelegate cannot be registered", x);
			throw new ImplementationException();
		}

		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("MBeanServer instance created successfully");
	}

	/**
	 * Returns the ClassLoaderRepository that maps ClassLoaderRepositories of all MBeanServer in a JVM.
	 *
	 * @see javax.management.loading.ClassLoaderRepository
	 */
	public static ClassLoaderRepository getStaticClassLoaderRepository()
	{
		return m_staticClassLoaderRepository;
	}

	/**
	 * Returns the ClassLoaderRepository for this MBeanServer.
	 * When first the ClassLoaderRepository is created in the constructor, the system property
	 * <i>mx4j.mbeanserver.classloader.repository</i> it tested; if it is non-null and defines a subclass of
	 * {@link BaseClassLoaderRepository}, then that class is used instead of the default one.
	 */
	public ClassLoaderRepository getClassLoaderRepository()
	{
        SecurityManager sm = System.getSecurityManager();
        if (sm != null)
        {
            sm.checkPermission(new MBeanPermission("-#-[-]", "getClassLoaderRepository"));
        }

		return getBaseClassLoaderRepository();
	}

	private BaseClassLoaderRepository getBaseClassLoaderRepository()
	{
		return m_classLoaderRepository;
	}

	/**
	 * Returns a ClassLoader using the ObjectName.
	 * Returns null if the MBean found is not an instance of java.lang.ClassLoader
	 *
	 * @param name The ObjectName to be used
	 * @return ClassLoader The ClassLoader
	 * @throws InstanceNotFoundException If no instance was found from the
	 *          MBeanRepository
	 * @throws SecurityException If the ObjectName doesn't have
	 *          permission to getClassLoader
	 *
	 */
	public ClassLoader getClassLoader(ObjectName name) throws InstanceNotFoundException
	{
		SecurityManager sm = System.getSecurityManager();
		if (sm != null)
		{
			name = secureObjectName(name);
			sm.checkPermission(new MBeanPermission(Utils.createPermissionTargetName("-", "-", name), "getClassLoader"));
		}

		return findClassLoaderMBean(name);
	}

	/**
	 * Returns the ClassLoader used by the MBean for the given ObjectName
	 *
	 * @param name The ObjectName for the MBean
	 * @throws InstanceNotFoundException If no instance was found from the
	 *          MBeanRepository
	 * @throws SecurityException If the ObjectName doesn't have
	 *          permission to getClassLoaderFor
	 *
	 */
	public ClassLoader getClassLoaderFor(ObjectName name) throws InstanceNotFoundException
	{
		SecurityManager sm = System.getSecurityManager();
		if (sm != null)
		{
			name = secureObjectName(name);
			sm.checkPermission(new MBeanPermission(Utils.createPermissionTargetName("-", "-", name), "getClassLoaderFor"));
		}

		MBeanMetaData metadata = findMBeanMetaData(name);
		return metadata.classloader;
	}

	/**
	 * Returns the MBean classloader corrispondent to the given ObjectName.
	 * If <code>name</code> is null, the classloader of this class is returned,
	 * and if it is null as well, the system classloader is returned.
	 * Guaranteed to never return null.
	 */
	private ClassLoader findClassLoader(ObjectName name) throws InstanceNotFoundException
	{
		if (name == null)
		{
			ClassLoader cl = getClass().getClassLoader();
			if (cl != null) return cl;

			// If the classloader is null, then the JMX classes are in the boot classpath;
			// every classloader will do the job, because of delegation. Just avoid to return null.
			return ClassLoader.getSystemClassLoader();
		}
		else
		{
			return findClassLoaderMBean(name);
		}
	}

	private ClassLoader findClassLoaderMBean(ObjectName name) throws InstanceNotFoundException
	{
		MBeanMetaData metadata = findMBeanMetaData(name);
		if (metadata.mbean instanceof ClassLoader)
		{
			return (ClassLoader)metadata.mbean;
		}
		else
		{
			throw new InstanceNotFoundException(name.getCanonicalName());
		}
	}

	/**
	 * Deserialize the given bytes using the classloader that can actually load className using the classloader
	 * MBean with the specified ObjectName as initiating classloader.
	 *
	 * @param className The className to load
	 * @param loaderName The ObjectName of the classloader MBean that initiates the loading of className
	 * @param bytes The data to de-serialize
	 * @return ObjectInputStream The de-serialized stream
	 * @throws InstanceNotFoundException If the specified ObjectName does not represent a registered MBean
	 * @throws OperationsException I/O Related Error
	 * @throws ReflectionException If className can't be loaded
	 *
	 */
	public ObjectInputStream deserialize(String className, ObjectName loaderName, byte[] bytes)
			throws InstanceNotFoundException, OperationsException, ReflectionException
	{
		if (className == null || className.trim().length() == 0)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Invalid class name"));
		}

		// TODO: Missing security check (JMX 1.2); beware that if loaderName is null this class classloader
		// should be used (but retrieving this classloader has no security checks). Refer also to the
		// behavior of getClassLoader(ObjectName name): what does it return if name == null ?

		ClassLoader cl = findClassLoader(loaderName);
		try
		{
			Class cls = cl.loadClass(className);
			return deserializeImpl(cls.getClassLoader(), bytes);
		}
		catch (ClassNotFoundException x)
		{
			throw new ReflectionException(x);
		}
	}

	/**
	 * Deserialize the specified bytes using the classloader that is able to load the specified class.
	 * The ClassLoaderRepository of this MBeanServer is used to find the classloader.
	 *
	 * @param className The className used to find a suitable classloader
	 * @param bytes The data to de-serialize
	 * @return ObjectInputStream The de-serialized stream
	 * @throws OperationsException I/O Related Error
	 * @throws ReflectionException If the className cannot be loaded using this MBeanServer's ClassLoaderRepository
	 *
	 */
	public ObjectInputStream deserialize(String className, byte[] bytes)
			throws OperationsException, ReflectionException
	{
		if (className == null || className.trim().length() == 0)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Invalid class name"));
		}

		// Find the classloader that can load the given className using the ClassLoaderRepository
		try
		{
			Class cls = getClassLoaderRepository().loadClass(className);
			return deserializeImpl(cls.getClassLoader(), bytes);
		}
		catch (ClassNotFoundException x)
		{
			throw new ReflectionException(x);
		}
	}

	/**
	 * Deserialize the specified bytes using the classloader of the MBean registered with the given ObjectName
	 *
	 * @param objectName  The ObjectName of the MBean whose classloader is used to deserialize
	 * @param bytes The data to de-serialize
	 * @return ObjectInputStream The de-serialized stream
	 * @throws OperationsException I/O Related Error
	 * @throws InstanceNotFoundException If the specified ObjectName does not represent a registered MBean
	 */
	public ObjectInputStream deserialize(ObjectName objectName, byte[] bytes)
			throws InstanceNotFoundException, OperationsException
	{
		ClassLoader cl = getClassLoaderFor(objectName);
		return deserializeImpl(cl, bytes);
	}

	/**
	 * Deserialized the given bytes using the specified classloader.
	 */
	private ObjectInputStream deserializeImpl(ClassLoader classloader, byte[] bytes) throws OperationsException
	{
		if (bytes == null || bytes.length == 0)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Invalid bytes"));
		}

		ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
		try
		{
			return new ClassLoaderObjectInputStream(bais, classloader);
		}
		catch (IOException x)
		{
			throw new OperationsException(x.toString());
		}
	}

	private MBeanServerInterceptor getHeadInterceptor()
	{
		MBeanServerInterceptor head = m_invoker.getHeadInterceptor();

		if (head == null) throw new IllegalStateException("No MBeanServer interceptor, probably the configurator has been stopped");

		return head;
	}

	private Logger getLogger()
	{
		return Log.getLogger(getClass().getName());
	}

	private MBeanRepository createMBeanRepository()
	{
		Logger logger = getLogger();

		final String property = "mx4j.mbeanserver.repository";
		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Checking for system property " + property);

		String value = (String)AccessController.doPrivileged(new PrivilegedAction()
		{
			public Object run()
			{
				return System.getProperty(property);
			}
		});

		if (value != null)
		{
			if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Property found for custom MBeanServer registry; class is: " + value);

			try
			{
				MBeanRepository registry = (MBeanRepository)Thread.currentThread().getContextClassLoader().loadClass(value).newInstance();
				if (logger.isEnabledFor(Logger.TRACE))
				{
					logger.trace("Custom MBeanServer registry created successfully");
				}
				return registry;
			}
			catch (Exception x)
			{
				if (logger.isEnabledFor(Logger.TRACE))
				{
					logger.trace("Custom MBeanServer registry could not be created", x);
				}
			}
		}

		return new DefaultMBeanRepository();
	}

	private BaseClassLoaderRepository createClassLoaderRepository()
	{
		Logger logger = getLogger();

		final String property = "mx4j.mbeanserver.classloader.repository";
		if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Checking for system property " + property);

		String value = (String)AccessController.doPrivileged(new PrivilegedAction()
		{
			public Object run()
			{
				return System.getProperty(property);
			}
		});

		if (value != null)
		{
			if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Property found for custom ClassLoaderRepository; class is: " + value);

			try
			{
				BaseClassLoaderRepository repository = (BaseClassLoaderRepository)Thread.currentThread().getContextClassLoader().loadClass(value).newInstance();
				if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Custom ClassLoaderRepository created successfully");
				return repository;
			}
			catch (Exception x)
			{
				if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Custom ClassLoaderRepository could not be created", x);
			}
		}
		return new DefaultClassLoaderRepository();
	}

	private MBeanRepository getMBeanRepository()
	{
		return m_repository;
	}

	private MBeanMetaData findMBeanMetaData(ObjectName objectName) throws InstanceNotFoundException
	{
		MBeanMetaData metadata = null;
		if (objectName != null)
		{
			objectName = normalizeObjectName(objectName);

			MBeanRepository repository = getMBeanRepository();
			synchronized (repository)
			{
				metadata = repository.get(objectName);
			}
		}
		if (metadata == null)
		{
			throw new InstanceNotFoundException("MBeanServer cannot find MBean with ObjectName " + objectName);
		}
		return metadata;
	}

	/**
	 * Add a NotificationListener to a registered MBean
	 *
	 * Register an MBean listener to allow broadcast of management events(notifications).
	 * The listener should be an instance of {@link NotificationListener}.
	 * See Page 47 (JMX Notification Model) of the JMX1.1 Agent Specification
	 * for details.
	 *
	 *
	 * @param observed The observed MBean where the listener is added
	 * @param listener Handles notification sent by the observed MBean
	 * @param filter  Implementation which provides filter to be applied on
	 *              notifications
	 * @param handback  Object sent when to the listener
	 * @throws InstanceNotFoundException If the MBean pointed by "listener"
	 *          is not an instance of NotificationListener
	 * @see mx4j.server.interceptor.NotificationListenerMBeanServerInterceptor
	 *
	 */
	public void addNotificationListener(ObjectName observed, ObjectName listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException
	{
		listener = secureObjectName(listener);

		Object mbean = findMBeanMetaData(listener).mbean;
		if (!(mbean instanceof NotificationListener))
		{
			throw new InstanceNotFoundException("ObjectName " + listener + " is not a NotificationListener");
		}
		addNotificationListener(observed, (NotificationListener)mbean, filter, handback);
	}

	/**
	 * Add a NotificationListener to a registered MBean
	 *
	 * Register an MBean listener to allow broadcast of management events(notifications).
	 * The listener should be an instance of {@link NotificationListener}.
	 * See Page 47 (JMX Notification Model) of the JMX1.1 Agent Specification
	 * for details.
	 *
	 *
	 * @param observed The observed MBean where the listener is added
	 * @param listener Handles notification sent by the observed MBean
	 * @param filter  Implementation which provides filter to be applied on
	 *              notifications
	 * @param handback  Object sent when to the listener
	 * @throws InstanceNotFoundException If the MBean pointed by "listener"
	 *          is not an instance of NotificationBroadCaster
	 * @see mx4j.server.interceptor.NotificationListenerMBeanServerInterceptor
	 *
	 */
	public void addNotificationListener(ObjectName observed, NotificationListener listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException
	{
		if (listener == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("NotificationListener cannot be null"));
		}

		observed = secureObjectName(observed);

		MBeanMetaData metadata = findMBeanMetaData(observed);

		Object mbean = metadata.mbean;

		if (!(mbean instanceof NotificationBroadcaster))
		{
			throw new InstanceNotFoundException("ObjectName " + observed + " is not a NotificationBroadcaster");
		}

		addNotificationListenerImpl(metadata, listener, filter, handback);
	}

	private void addNotificationListenerImpl(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException
	{
		getHeadInterceptor().addNotificationListener(metadata, listener, filter, handback);
	}

	/**
	 * Remove a notificationListener.
	 *
	 * @param observed The observed MBean where the listener is removed
	 * @param listener The listener to removed
	 * @throws InstanceNotFoundException If the MBean pointed by "listener"
	 *          is not an instance of NotificationListener
	 * @throws ListnerNotFoundException If no specified listener was found
	 *
	 */
	public void removeNotificationListener(ObjectName observed, ObjectName listener)
			throws InstanceNotFoundException, ListenerNotFoundException
	{
		listener = secureObjectName(listener);

		Object mbean = findMBeanMetaData(listener).mbean;
		if (!(mbean instanceof NotificationListener))
		{
			throw new InstanceNotFoundException("ObjectName " + listener + " is not a NotificationListener");
		}
		removeNotificationListener(observed, (NotificationListener)mbean);
	}

	/**
	 * Removed a notificationListener.
	 *
	 * @param observed The observed MBean where the listener is removed
	 * @param listener The listener to removed
	 * @throws InstanceNotFoundException If the MBean pointed by "listener"
	 *          is not an instance of NotificationListener
	 * @throws ListnerNotFoundException If no specified listener was found
	 *
	 */
	public void removeNotificationListener(ObjectName observed, NotificationListener listener)
			throws InstanceNotFoundException, ListenerNotFoundException
	{
		if (listener == null)
		{
			throw new ListenerNotFoundException("NotificationListener cannot be null");
		}

		observed = secureObjectName(observed);

		MBeanMetaData metadata = findMBeanMetaData(observed);
		Object mbean = metadata.mbean;

		if (!(mbean instanceof NotificationBroadcaster))
		{
			throw new InstanceNotFoundException("ObjectName " + observed + " is not a NotificationBroadcaster");
		}

		removeNotificationListenerImpl(metadata, listener, null, null);
	}

	/**
	 * Not present in the specification but needed
	 *
	 */
	public void removeNotificationListener(ObjectName observed, ObjectName listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException, ListenerNotFoundException
	{
		// TODO for JMX 1.2: secure the ObjectNames

		Object mbean = findMBeanMetaData(listener).mbean;
		if (!(mbean instanceof NotificationListener))
		{
			throw new InstanceNotFoundException("ObjectName " + listener + " is not a NotificationListener");
		}
		removeNotificationListener(observed, (NotificationListener)mbean, filter, handback);
	}

	/**
	 * Not present in the specification but needed
	 *
	 *
	 */
	public void removeNotificationListener(ObjectName observed, NotificationListener listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException, ListenerNotFoundException
	{
		// TODO for JMX 1.2: secure the ObjectNames

		if (listener == null)
		{
			throw new ListenerNotFoundException("NotificationListener cannot be null");
		}

		MBeanMetaData metadata = findMBeanMetaData(observed);
		Object mbean = metadata.mbean;

		if (!(mbean instanceof NotificationBroadcaster))
		{
			throw new InstanceNotFoundException("ObjectName " + observed + " is not a NotificationBroadcaster");
		}

		removeNotificationListenerImpl(metadata, listener, filter, handback);
	}

	private void removeNotificationListenerImpl(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback)
			throws ListenerNotFoundException
	{
		getHeadInterceptor().removeNotificationListener(metadata, listener, filter, handback);
	}

	/**
	 * Instantiate the given className using the default {@link javax.management.loading.ClassLoaderRepository ClassLoaderRepository}.
	 * <p>
	 * The class should have a public constructor. Note that the instance
	 * created isn't registered with the MBeanServer
	 * </p>
	 *
	 * @param className The className to instantiate
	 * @return Object The created instance
	 * @throws ReflectionException
	 * @throws MBeanException
	 * @throws IllegalArgumentException If the className is null or 0 length
	 *
	 */
	public Object instantiate(String className)
			throws ReflectionException, MBeanException
	{
		return instantiate(className, null, null);
	}

	/**
	 * Instantiate the given className using the given args and parameters
	 * using the {@link javax.management.loading.ClassLoaderRepository ClassLoaderRepository}
	 * <p>
	 * The class should have a public constructor. Note that the instance
	 * created isn't registered with the MBeanServer
	 * </p>
	 *
	 * @param className The className to instantiate
	 * @param args The argument as an array
	 * @param parameters The parameters as an array
	 * @return Object The created instance
	 * @throws ReflectionException
	 * @throws MBeanException
	 * @throws IllegalArgumentException If the className is null or 0 length
	 *
	 */
	public Object instantiate(String className, Object[] args, String[] parameters)
			throws ReflectionException, MBeanException
	{
		if (className == null || className.trim().length() == 0)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Class name cannot be null or empty"));
		}

		try
		{
			Class cls = getBaseClassLoaderRepository().loadClass(className);
			return instantiateImpl(className, cls.getClassLoader(), null, parameters, args).mbean;
		}
		catch (ClassNotFoundException x)
		{
			throw new ReflectionException(x);
		}
	}

	/**
	 * Instantiate the given className using the loaderName as the ClassLoader
	 * <p>
	 * The class should have a public constructor. Note that the instance
	 * created isn't registered with the MBeanServer
	 * </p>
	 *
	 * @param className The className to instantiate
	 * @param loaderName The MBean loaderName to use
	 * @return Object The created instance
	 * @throws ReflectionException
	 * @throws MBeanException
	 * @throws InstanceNotFoundException If the className is null or 0 length
	 *
	 */
	public Object instantiate(String className, ObjectName loaderName)
			throws ReflectionException, MBeanException, InstanceNotFoundException
	{
		return instantiate(className, loaderName, null, null);
	}

	/**
	 *
	 * Instantiate the given className using the ClassLoader (loaderName),
	 * args and parameters
	 *
	 *
	 * @param className The className to instantiate
	 * @param loaderName The MBean loaderName to use
	 * @param args The argument as an array
	 * @param parameters The parameters as an array
	 * @return Object The created instance
	 * @throws ReflectionException
	 * @throws MBeanException
	 * @throws InstanceNotFoundException If the className is null or 0 length
	 *
	 *
	 */
	public Object instantiate(String className, ObjectName loaderName, Object[] args, String[] parameters)
			throws ReflectionException, MBeanException, InstanceNotFoundException
	{
		if (className == null || className.trim().length() == 0)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Class name cannot be null or empty"));
		}

		// loaderName can be null: means using this class' ClassLoader

		loaderName = secureObjectName(loaderName);
		if (loaderName != null && loaderName.isPattern())
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("ObjectName for the ClassLoader cannot be a pattern ObjectName"));
		}

		ClassLoader cl = findClassLoader(loaderName);
		return instantiateImpl(className, cl, null, parameters, args).mbean;
	}

	private MBeanMetaData instantiateImpl(String className, ClassLoader classloader, ObjectName name, String[] params, Object[] args)
			throws ReflectionException, MBeanException
	{
		if (params == null) params = EMPTY_PARAMS;
		if (args == null) args = EMPTY_ARGS;
		if (classloader == null) classloader = ClassLoader.getSystemClassLoader();

		MBeanMetaData metadata = createMBeanMetaData();
		metadata.classloader = classloader;
		metadata.name = secureObjectName(name);

		getHeadInterceptor().instantiate(metadata, className, params, args);

		return metadata;
	}


	/**
	 * Registers an MBean and returns an ObjectInstance.
	 *
	 * <p>
	 * The MBeanServer will use the default ClassLoaderRepository
	 * </p>
	 *
	 * @param className The classname of the MBean
	 * @param objectName The given objectName
	 * @return ObjectInstance The ObjectInstance created
	 * @throws ReflectionException
	 * @throws InstanceAlreadyExistsException If an instance already exists
	 *              in the MBeanServer
	 * @throws MBeanRegistrationException
	 * @throws MBeanException
	 * @throws NotCompliantMBeanException If not an MBean
	 */
	public ObjectInstance createMBean(String className, ObjectName objectName)
			throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException
	{
		return createMBean(className, objectName, null, null);
	}

	/**
	 *
	 * Registers an MBean and returns an ObjectInstance
	 * <p>
	 * The MBeanServer will use the default ClassLoaderRepository
	 * </p>
	 *
	 * @param className The classname of the MBean
	 * @param objectName The given objectName
	 * @param args The arguments as array
	 * @param parameters The parameters as array
	 * @return ObjectInstance The ObjectInstance created
	 * @throws ReflectionException
	 * @throws InstanceAlreadyExistsException
	 * @throws MBeanRegistrationException
	 * @throws MBeanException
	 * @throws NotCompliantMBeanException
	 *
	 */
	public ObjectInstance createMBean(String className, ObjectName objectName, Object[] args, String[] parameters)
			throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException
	{
		try
		{
			Class cls = getBaseClassLoaderRepository().loadClass(className);

			MBeanMetaData metadata = instantiateImpl(className, cls.getClassLoader(), objectName, parameters, args);

			registerImpl(metadata, false);

			return metadata.instance;
		}
		catch (ClassNotFoundException x)
		{
			throw new ReflectionException(x);
		}
	}

	/**
	 *
	 * Registers an MBean and returns an ObjectInstance, and use the specified
	 * loaderName.
	 *
	 * @param className The classname of the MBean
	 * @param objectName The objectname
	 * @param loaderName The ObjectName of the loader
	 * @return ObjectInstance The ObjectInstance created
	 * @throws ReflectionException
	 * @throws InstanceAlreadyExistsException
	 * @throws MBeanRegistrationException
	 * @throws NotCompliantMBeanException
	 * @throws InstanceNotFoundException
	 *
	 */
	public ObjectInstance createMBean(String className, ObjectName objectName, ObjectName loaderName)
			throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException, InstanceNotFoundException
	{
		return createMBean(className, objectName, loaderName, null, null);
	}


	/**
	 *
	 * Registers an MBean and returns an ObjectInstance using the specified
	 * loaderName for the ClassLoader
	 * <p>
	 * The MBeanServer will use the default ClassLoaderRepository
	 * </p>
	 *
	 * @param className The classname of the MBean
	 * @param objectName The given objectName
	 * @param args The arguments as array
	 * @param parameters The parameters as array
	 * @return ObjectInstance The ObjectInstance created
	 * @throws ReflectionException
	 * @throws InstanceAlreadyExistsException
	 * @throws MBeanRegistrationException
	 * @throws MBeanException If error on instantiating the MBean
	 * @throws InstanceNotFoundException If the loaderName wasn't found
	 * @throws NotCompliantMBeanException
	 *
	 */
	public ObjectInstance createMBean(String className, ObjectName objectName, ObjectName loaderName, Object[] args, String[] parameters)
			throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException, InstanceNotFoundException
	{
		loaderName = secureObjectName(loaderName);

		ClassLoader cl = findClassLoader(loaderName);

		MBeanMetaData metadata = instantiateImpl(className, cl, objectName, parameters, args);

		registerImpl(metadata, false);

		return metadata.instance;
	}

	/**
	 * Register an already created MBean using the specified <code>ObjectName</code>
	 * @param mbean The instance of the MBean to register
	 * @param objectName The <code>ObjectName</code> to use
	 * @throws InstanceAlreadyExistsException If the MBean already exists
	 * @throws MBeanRegistrationException Error in registering the MBean
	 * @throws NotCompliantMBeanException If not a compliant MBean
	 * @throws IllegalArgumentException if the MBean object is null
	 *
	 */
	public ObjectInstance registerMBean(Object mbean, ObjectName objectName)
			throws InstanceAlreadyExistsException, MBeanRegistrationException, NotCompliantMBeanException
	{
		return registerMBeanImpl(mbean, objectName, false);
	}

	private ObjectInstance registerMBeanImpl(Object mbean, ObjectName objectName, boolean privileged)
			throws InstanceAlreadyExistsException, MBeanRegistrationException, NotCompliantMBeanException
	{
		if (mbean == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("MBean cannot be null"));
		}

		MBeanMetaData metadata = createMBeanMetaData();
		metadata.mbean = mbean;
		ClassLoader cl = mbean.getClass().getClassLoader();
		if (cl == null)
		{
			// MBean class is in boot classpath, use system classloader
			// For delegation, it will be found anyway,
			// I just don't want to set it to null
			cl = ClassLoader.getSystemClassLoader();
		}
		metadata.classloader = cl;
		metadata.name = secureObjectName(objectName);

		registerImpl(metadata, privileged);

		return metadata.instance;
	}

	private MBeanMetaData createMBeanMetaData()
	{
		return new MBeanMetaData();
	}

	/**
	 * This is called only to register implementation MBeans from the constructor.
	 * Since to create an instance of this class already requires a permission, here we hide the registration
	 * of implementation MBeans to the client that thus need no further permissions.
	 */
	private ObjectInstance privilegedRegisterMBean(final Object mbean, final ObjectName name)
			throws InstanceAlreadyExistsException, MBeanRegistrationException, NotCompliantMBeanException
	{
		try
		{
			return (ObjectInstance)AccessController.doPrivileged(new PrivilegedExceptionAction()
			{
				public Object run() throws Exception
				{
					return registerMBeanImpl(mbean, name, true);
				}
			});
		}
		catch (PrivilegedActionException x)
		{
			Exception xx = x.getException();
			if (xx instanceof InstanceAlreadyExistsException) throw (InstanceAlreadyExistsException)xx;
			else if (xx instanceof MBeanRegistrationException) throw (MBeanRegistrationException)xx;
			else if (xx instanceof NotCompliantMBeanException) throw (NotCompliantMBeanException)xx;
			else throw new MBeanRegistrationException(xx);
		}
	}

	private void registerImpl(MBeanMetaData metadata, boolean privileged) throws InstanceAlreadyExistsException, MBeanRegistrationException, NotCompliantMBeanException
	{
		m_introspector.introspect(metadata);

		if (!m_introspector.isMBeanCompliant(metadata)) throw new NotCompliantMBeanException("MBean is not compliant");

		MBeanServerInterceptor head = getHeadInterceptor();

		try
		{
			// With this call, the MBean implementor can replace the ObjectName with a subclass that is not secure
			head.registration(metadata, MBeanServerInterceptor.PRE_REGISTER);
			metadata.name = secureObjectName(metadata.name);

			register(metadata, privileged);
			head.registration(metadata, MBeanServerInterceptor.POST_REGISTER_TRUE);
		}
		catch (Throwable x)
		{
			try
			{
				head.registration(metadata, MBeanServerInterceptor.POST_REGISTER_FALSE);
			}
			catch (MBeanRegistrationException ignored)
			{/* Ignore this one to rethrow the other one */
			}

			if (x instanceof SecurityException)
            {
                throw (SecurityException)x;
            }
            else if (x instanceof InstanceAlreadyExistsException)
			{
				throw (InstanceAlreadyExistsException)x;
			}
			else if (x instanceof MBeanRegistrationException)
			{
				throw (MBeanRegistrationException)x;
			}
			else if (x instanceof Exception)
			{
				throw new MBeanRegistrationException((Exception)x);
			}
			else if (x instanceof Error)
			{
				throw new MBeanRegistrationException(new RuntimeErrorException((Error)x));
			}
			else
			{
				throw new ImplementationException();
			}
		}

		if (metadata.mbean instanceof ClassLoader)
		{
			ClassLoader cl = (ClassLoader)metadata.mbean;
			getBaseClassLoaderRepository().addClassLoader(cl);
			BaseClassLoaderRepository dlr = (BaseClassLoaderRepository)getStaticClassLoaderRepository();
			dlr.addClassLoader(cl);
		}

		metadata.instance = new ObjectInstance(metadata.name, metadata.info.getClassName());
	}

	private void register(MBeanMetaData metadata, boolean privileged) throws InstanceAlreadyExistsException
	{
		metadata.name = normalizeObjectName(metadata.name);

		ObjectName objectName = metadata.name;
		if (objectName == null || objectName.isPattern())
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("ObjectName cannot be null or a pattern ObjectName"));
		}
		if (objectName.getDomain().equals("JMImplementation") && !privileged)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Domain 'JMImplementation' is reserved for the JMX Agent"));
		}

		MBeanRepository repository = getMBeanRepository();
		synchronized (repository)
		{
			if (repository.get(objectName) != null)
			{
				throw new InstanceAlreadyExistsException(objectName.toString());
			}
			repository.put(objectName, metadata);
		}

		notify(objectName, MBeanServerNotification.REGISTRATION_NOTIFICATION);
	}

	private void notify(ObjectName objectName, String notificationType)
	{
		long sequenceNumber = 0;
		synchronized (MBeanServerImpl.class)
		{
			sequenceNumber = m_notifications;
			++m_notifications;
		}

		m_delegate.sendNotification(new MBeanServerNotification(notificationType, m_delegateName, sequenceNumber, objectName));
	}


	/**
	 * Unregister a registered MBean
	 *
	 * @param objectName The objectName where the MBean to unregister is bound
	 * @throws InstanceNotFoundException If no instance was found
	 * @throws MBeanRegistrationException
	 */
	public void unregisterMBean(ObjectName objectName)
			throws InstanceNotFoundException, MBeanRegistrationException
	{
		objectName = secureObjectName(objectName);

		if (objectName == null || objectName.isPattern())
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("ObjectName cannot be null or a pattern ObjectName"));
		}
		if (objectName.getDomain().equals("JMImplementation"))
		{
			throw new MBeanRegistrationException(new RuntimeOperationsException(new IllegalArgumentException("Domain 'JMImplementation' is reserved for the JMX Agent")));
		}

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		try
		{
			MBeanServerInterceptor head = getHeadInterceptor();
			head.registration(metadata, MBeanServerInterceptor.PRE_DEREGISTER);

			unregister(metadata);

			getHeadInterceptor().registration(metadata, MBeanServerInterceptor.POST_DEREGISTER);

			if (metadata.mbean instanceof ClassLoader)
			{
				getBaseClassLoaderRepository().removeClassLoader((ClassLoader)metadata.mbean);
				BaseClassLoaderRepository dlr = (BaseClassLoaderRepository)getStaticClassLoaderRepository();
				dlr.removeClassLoader((ClassLoader)metadata.mbean);
			}
		}
		catch (MBeanRegistrationException x)
		{
			throw x;
		}
		catch (SecurityException x)
		{
			throw x;
		}
		catch (Exception x)
		{
			throw new MBeanRegistrationException(x);
		}
		catch (Error x)
		{
			throw new MBeanRegistrationException(new RuntimeErrorException(x));
		}
	}

	private void unregister(MBeanMetaData metadata)
	{
		ObjectName objectName = metadata.name;

		MBeanRepository repository = getMBeanRepository();
		synchronized (repository)
		{
			repository.remove(objectName);
		}

		notify(objectName, MBeanServerNotification.UNREGISTRATION_NOTIFICATION);
	}

	/**
	 * Retrieve an attribute using the given objectName
	 *
	 * @param objectName The objectName of the registered MBean
	 * @param attribute The name of the attribute to retrieve
	 * @return Object The attribute
	 * @throws InstanceNotFoundException If no instance was found
	 * @throws MBeanException
	 * @throws AttributeNotFoundException If the attribute wasn't found
	 * @throws ReflectionException
	 */
	public Object getAttribute(ObjectName objectName, String attribute)
			throws InstanceNotFoundException, MBeanException, AttributeNotFoundException, ReflectionException
	{
		if (attribute == null || attribute.trim().length() == 0)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Invalid attribute"));
		}

		objectName = secureObjectName(objectName);

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		return getHeadInterceptor().getAttribute(metadata, attribute);
	}


	/**
	 * Sets the vallue of the Attribute of the given MBean
	 *
	 * @param objectName the objectName to use
	 * @param attribute the attribute to set
	 * @throws InstanceNotFoundException No instance was found
	 * @throws InvalidAttributeValueException If the attribute was invalid
	 * @throws MBeanException The exception thrown
	 * @throws ReflectionException
	 *
	 */
	public void setAttribute(ObjectName objectName, Attribute attribute)
			throws InstanceNotFoundException, AttributeNotFoundException, InvalidAttributeValueException, MBeanException, ReflectionException
	{
		if (attribute == null || attribute.getName().trim().length() == 0)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Invalid attribute"));
		}

		objectName = secureObjectName(objectName);

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		getHeadInterceptor().setAttribute(metadata, attribute);
	}

	public AttributeList getAttributes(ObjectName objectName, String[] attributes)
			throws InstanceNotFoundException, ReflectionException
	{
		if (attributes == null || attributes.length == 0)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Invalid attribute list"));
		}

		objectName = secureObjectName(objectName);

		MBeanMetaData metadata = findMBeanMetaData(objectName);

        SecurityManager sm = System.getSecurityManager();
        if (sm != null)
        {
            // Must check if the user has the right to call this method,
            // regardless of the attributes
            sm.checkPermission(new MBeanPermission(Utils.createPermissionTargetName(metadata.info.getClassName(), "-", objectName), "getAttribute"));
        }

		return getHeadInterceptor().getAttributes(metadata, attributes);
	}

	/**
	 * Sets the attribute of an MBean
	 *
	 * @param objectName the objectName
	 * @param attributes The attributes to set
	 * @return AttributeList the attributelist that were set with new values
	 * @throws InstanceNotFoundException If the MBean wasn't found
	 * @throws ReflectionException
	 *
	 *
	 */
	public AttributeList setAttributes(ObjectName objectName, AttributeList attributes)
			throws InstanceNotFoundException, ReflectionException
	{
		if (attributes == null)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Invalid attribute list"));
		}

		objectName = secureObjectName(objectName);

        MBeanMetaData metadata = findMBeanMetaData(objectName);

        SecurityManager sm = System.getSecurityManager();
        if (sm != null)
        {
            // Must check if the user has the right to call this method,
            // regardless of the attributes
            sm.checkPermission(new MBeanPermission(Utils.createPermissionTargetName(metadata.info.getClassName(), "-", objectName), "setAttribute"));
        }

		return getHeadInterceptor().setAttributes(metadata, attributes);
	}

	/**
	 * Invokes/Execute an operation on an MBean
	 *
	 * @param objectName The objectName to use
	 * @param methodName the name of the method to invoke
	 * @param args the arguments
	 * @param parameters the parameters
	 * @return Object The result of invocation
	 * @throws InstanceNotFoundException If the MBean wasn't found
	 * @throws MBeanException
	 * @throws ReflectionException
	 *
	 *
	 */
	public Object invoke(ObjectName objectName, String methodName, Object[] args, String[] parameters)
			throws InstanceNotFoundException, MBeanException, ReflectionException
	{
		if (methodName == null || methodName.trim().length() == 0)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Invalid operation name"));
		}

		if (args == null) args = EMPTY_ARGS;
		if (parameters == null) parameters = EMPTY_PARAMS;

		objectName = secureObjectName(objectName);

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		return getHeadInterceptor().invoke(metadata, methodName, parameters, args);
	}

	/**
	 * Retrieve the default domain of this MBean Server
	 *
	 * @return String the default domain
	 *
	 */
	public String getDefaultDomain()
	{
		return m_defaultDomain;
	}

	/**
	 * Returns the number of MBeans
	 *
	 * @return Integer The number of MBeans registered
	 *
	 */
	public Integer getMBeanCount()
	{
		MBeanRepository repository = getMBeanRepository();
		synchronized (repository)
		{
			return new Integer(repository.size());
		}
	}

	/**
	 *  Checks if the specified <code>ObjectName</code> is registered
	 * with this MBean Server
	 *
	 * @param objectName The objectName to check
	 * @return boolean true if registered
	 *
	 */
	public boolean isRegistered(ObjectName objectName)
	{
		try
		{
			return findMBeanMetaData(objectName) != null;
		}
		catch (InstanceNotFoundException x)
		{
			return false;
		}
	}

	/**
	 * Returns an MBeanInfo using the specified objectName
	 *
	 * @param objectName The objectName to use for the MBeanInfo
	 * @return MBeanInfo The MBeanInfo object
	 * @throws InstanceNotFoundException If no instance was found
	 * @throws IntrospectionException
	 * @throws ReflectionException
	 *
	 *
	 */
	public MBeanInfo getMBeanInfo(ObjectName objectName)
			throws InstanceNotFoundException, IntrospectionException, ReflectionException
	{
		objectName = secureObjectName(objectName);

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		MBeanInfo info = getHeadInterceptor().getMBeanInfo(metadata);
		if (info == null) throw new IntrospectionException("MBeanInfo returned by the DynamicMBean " + objectName + " is null");
		return info;
	}

	/**
	 *
	 * Returns an ObjectInstance using the specified objectName
	 *
	 * @param objectName The object where the ObjectInstance is registered.
	 * @return ObjectInstance The objectinstance
	 * @throws InstanceNotFoundException If no MBean is registered with the
	 *          given objectName
	 *
	 */
	public ObjectInstance getObjectInstance(ObjectName objectName)
			throws InstanceNotFoundException
	{
		SecurityManager sm = System.getSecurityManager();
        if (sm != null)
        {
			objectName = secureObjectName(objectName);
            sm.checkPermission(new MBeanPermission(Utils.createPermissionTargetName("-", "-", objectName), "getObjectInstance"));
        }

		MBeanMetaData metadata = findMBeanMetaData(objectName);
		return metadata.instance;
	}

	/**
	 *
	 * Checks if the MBean denoted by objectName is of type className
	 *
	 * @param objectName The objectName to check
	 * @param className the className as String
	 * @return boolean The result
	 * @throws InstanceNotFoundException If no MBean was found using <code>ObjectName</code>
	 *
	 */
	public boolean isInstanceOf(ObjectName objectName, String className)
			throws InstanceNotFoundException
	{
		if (className == null || className.trim().length() == 0)
		{
			throw new RuntimeOperationsException(new IllegalArgumentException("Invalid class name"));
		}

		objectName = secureObjectName(objectName);

		MBeanMetaData metadata = findMBeanMetaData(objectName);

        SecurityManager sm = System.getSecurityManager();
		if (sm != null)
		{
			sm.checkPermission(new MBeanPermission(Utils.createPermissionTargetName(metadata.info.getClassName(), "-", objectName), "isInstanceOf"));
		}

		try
		{
			Class cls = metadata.classloader.loadClass(className);
			return cls.isInstance(metadata.mbean);
		}
		catch (ClassNotFoundException x)
		{
			return false;
		}
	}

	/**
	 *
	 * Returns the names of all MBeans controlled by the MBean Server using
	 * the given patternName and filter.
	 *
	 * @param patternName The pattern to use
	 * @param filter The filter to use
	 * @return Set A set containing <code>ObjectInstance</code>. Empty if
	 *          no ObjectInstance was found
	 *
	 *
	 *
	 */
	public Set queryMBeans(ObjectName patternName, QueryExp filter)
	{
		SecurityManager sm = System.getSecurityManager();
		if (sm != null)
		{
			patternName = secureObjectName(patternName);
			// Must check if the user has the right to call this method,
			// no matter which ObjectName has been passed.
			sm.checkPermission(new MBeanPermission("-#-[-]", "queryMBeans"));
		}

		Set match = queryObjectNames(patternName, filter, true);

		Set set = new HashSet();
		for (Iterator i = match.iterator(); i.hasNext();)
		{
			ObjectName name = (ObjectName)i.next();
			try
			{
				MBeanMetaData metadata = findMBeanMetaData(name);
                set.add(metadata.instance);
			}
			catch (InstanceNotFoundException ignored)
			{
				// A concurrent thread removed the MBean after queryNames, ignore
			}
		}
        return set;
	}


	/**
	 * Returns the names of all MBeans controlled by the MBean Server using
	 * the given patternName and filter.
	 *
	 * @param patternName The pattern to use
	 * @param filter the filter to use
	 * @return Set A set containing <code>ObjectNames</code>. Empty if no
	 *          <code>ObjectNames</code> was found
	 *
	 *
	 */
	public Set queryNames(ObjectName patternName, QueryExp filter)
	{
		SecurityManager sm = System.getSecurityManager();
		if (sm != null)
		{
			patternName = secureObjectName(patternName);
			// Must check if the user has the right to call this method,
			// no matter which ObjectName has been passed.
			sm.checkPermission(new MBeanPermission("-#-[-]", "queryNames"));
		}

		return queryObjectNames(patternName, filter, false);
	}

	/**
	 * Utility method for queryNames and queryMBeans that returns a set of ObjectNames.
	 * It does 3 things:
	 * 1) filter the MBeans following the given ObjectName pattern
	 * 2) filter the MBeans following the permissions that client code has
	 * 3) filter the MBeans following the given QueryExp
	 * It is important that these 3 operations are done in this order
	 */
	private Set queryObjectNames(ObjectName patternName, QueryExp filter, boolean instances)
	{
		// First, retrieve the scope of the query: all mbeans matching the patternName
		Set scope = findMBeansByPattern(patternName);

		// Second, filter the scope by checking the caller's permissions
		Set secureScope = filterMBeansBySecurity(scope, instances);

		// Third, filter the scope using the given QueryExp
		Set match = filterMBeansByQuery(secureScope, filter);

		return match;
	}

	/**
	 * Returns a set of ObjectNames of the registered MBeans that match the given ObjectName pattern
	 */
	private Set findMBeansByPattern(ObjectName pattern)
	{
		if (pattern == null)
		{
			try
			{
				pattern = new ObjectName("*:*");
			}
			catch (MalformedObjectNameException ignored)
			{
			}
		}

		pattern = normalizeObjectName(pattern);

		String patternDomain = pattern.getDomain();
		Hashtable patternProps = pattern.getKeyPropertyList();

		Set set = new HashSet();

		// Clone the repository, we are faster than holding the lock while iterating
		MBeanRepository repository = (MBeanRepository)getMBeanRepository().clone();

		for (Iterator i = repository.iterator(); i.hasNext();)
		{
			MBeanMetaData metadata = (MBeanMetaData)i.next();
			ObjectName name = metadata.name;
			Hashtable props = name.getKeyPropertyList();

			String domain = name.getDomain();
			if (Utils.wildcardMatch(patternDomain, domain))
			{
				// Domain matches, now check properties
				if (pattern.isPropertyPattern())
				{
					// A property pattern with no entries, can only be '*'
					if (patternProps.size() == 0)
					{
						// User wants all properties
						set.add(name);
					}
					else
					{
						// Loop on the properties of the pattern.
						// If one is not found then the current ObjectName does not match
						boolean found = true;
						for (Iterator j = patternProps.entrySet().iterator(); j.hasNext();)
						{
							Map.Entry entry = (Map.Entry)j.next();
							Object patternKey = entry.getKey();
							Object patternValue = entry.getValue();
							if (patternKey.equals("*"))
							{
								continue;
							}

							// Try to see if the current ObjectName contains this entry
							if (!props.containsKey(patternKey))
							{
								// Not even the key is present
								found = false;
								break;
							}
							else
							{
								// The key is present, let's check if the values are equal
								Object value = props.get(patternKey);
								if (value == null && patternValue == null)
								{
									// Values are equal, go on with next pattern entry
									continue;
								}
								if (value != null && value.equals(patternValue))
								{
									// Values are equal, go on with next pattern entry
									continue;
								}
								// Here values are different
								found = false;
								break;
							}
						}
						if (found) set.add(name);
					}
				}
				else
				{
					if (props.entrySet().equals(patternProps.entrySet())) set.add(name);
				}
			}
		}
		return set;
	}

	/**
	 * Filters the given set of ObjectNames following the permission that client code has granted.
	 * Returns a set containing the allowed ObjectNames.
	 */
	private Set filterMBeansBySecurity(Set mbeans, boolean instances)
	{
		SecurityManager sm = System.getSecurityManager();
		if (sm == null) return mbeans;

		HashSet set = new HashSet();
		for (Iterator i = mbeans.iterator(); i.hasNext();)
		{
			ObjectName name = (ObjectName)i.next();
            try
            {
                MBeanMetaData metadata = findMBeanMetaData(name);
                String className = metadata.info.getClassName();
				sm.checkPermission(new MBeanPermission(Utils.createPermissionTargetName(className, "-", name), instances ? "queryMBeans" : "queryNames"));
				set.add(name);
            }
            catch (InstanceNotFoundException ignored)
            {
                // A concurrent thread removed this MBean, continue
				continue;
            }
			catch (SecurityException ignored)
			{
				// Don't add the name to the list, and go on.
			}
		}
		return set;
	}

	/**
	 * Filters the given set of ObjectNames following the given QueryExp.
	 * Returns a set of ObjectNames that match the given QueryExp.
	 */
	private Set filterMBeansByQuery(Set scope, QueryExp filter)
	{
		if (filter == null) return scope;

		Set set = new HashSet();
		for (Iterator i = scope.iterator(); i.hasNext();)
		{
			ObjectName name = (ObjectName)i.next();
			filter.setMBeanServer(this);
			try
			{
				if (filter.apply(name)) set.add(name);
			}
			catch (BadStringOperationException ignored)
			{
			}
			catch (BadBinaryOpValueExpException ignored)
			{
			}
			catch (BadAttributeValueExpException x)
			{
			}
			catch (InvalidApplicationException x)
			{
			}
			catch (SecurityException x)
			{
			}
		}
		return set;
	}

	/**
	 * Returns a normalized ObjectName from the given one.
	 * If an ObjectName is specified with the abbreviated notation for the default domain, that is ':key=value'
	 * this method returns an ObjectName whose domain is the default domain of this MBeanServer and with the same
	 * properties.
	 */
	private ObjectName normalizeObjectName(ObjectName name)
	{
		if (name == null) return null;

		String defaultDomain = getDefaultDomain();
		String domain = name.getDomain();

		if (domain.length() == 0 && defaultDomain.length() > 0)
		{
			// The given object name specifies the abbreviated form to indicate the default domain,
			// ie ':key=value', the empty string as domain. I must convert this abbreviated form
			// to the full one, if the default domain of this mbeanserver is not the empty string as well
			StringBuffer buffer = new StringBuffer(getDefaultDomain()).append(":").append(name.getKeyPropertyListString());
			if (name.isPropertyPattern())
			{
				if (name.getKeyPropertyList().size() > 0)
				{
					buffer.append(",*");
				}
				else
				{
					buffer.append("*");
				}
			}
			try
			{
				name = new ObjectName(buffer.toString());
			}
			catch (MalformedObjectNameException ignored)
			{
			}
		}
		return name;
	}

	/**
	 * Returns an ObjectName instance even if the provided ObjectName is a subclass.
	 * This is done to avoid security holes: a nasty ObjectName subclass can bypass security checks.
	 */
	private ObjectName secureObjectName(ObjectName name)
	{
		// I cannot trust ObjectName, since a malicious user can send a subclass that overrides equals and hashcode
		// to match another ObjectName for which it does not have permission, or returns different results from
		// ObjectName.getCanonicalName() for different calls, so that passes the security checks but in fact will
		// later refer to a different ObjectName for which it does not have permission.

		if (System.getSecurityManager() == null) return name;
		if (name == null || name.getClass() == ObjectName.class) return name;

		try
		{
			return new ObjectName(name.getCanonicalName());
		}
		catch (MalformedObjectNameException x)
		{
			 throw new SecurityException(x.toString());
		}
	}
}
