/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;

import javax.management.Attribute;
import javax.management.AttributeNotFoundException;
import javax.management.InvalidAttributeValueException;
import javax.management.JMRuntimeException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanOperationInfo;
import javax.management.MBeanParameterInfo;
import javax.management.ReflectionException;
import javax.management.RuntimeErrorException;
import javax.management.RuntimeMBeanException;
import javax.management.RuntimeOperationsException;

import mx4j.ImplementationException;
import mx4j.util.Utils;
import mx4j.util.MethodTernaryTree;

/**
 * MBeanInvoker that uses reflection to invoke on MBean instances.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.7 $
 */
public class ReflectedMBeanInvoker implements MBeanInvoker
{
	private static final String[] EMPTY_PARAMS = new String[0];
	private static final Object[] EMPTY_ARGS = new Object[0];

	private HashMap m_attributeCache = new HashMap();
	private MethodTernaryTree m_operationCache = new MethodTernaryTree();
	private MethodTernaryTree m_methodCache = new MethodTernaryTree();

	public Object invoke(MBeanMetaData metadata, String method, String[] params, Object[] args) throws MBeanException, ReflectionException
	{
		MBeanOperationInfo oper = getStandardOperationInfo(metadata, method, params);
		if (oper != null)
		{
			try
			{
				return invokeImpl(metadata, method, params, args);
			}
			catch (IllegalArgumentException x)
			{
				throw new RuntimeOperationsException(x);
			}
		}
		else
		{
			throw new ReflectionException(new NoSuchMethodException("Method " + method + " does not belong to the management interface"));
		}
	}

	public Object getAttribute(MBeanMetaData metadata, String attribute) throws MBeanException, AttributeNotFoundException, ReflectionException
	{
		MBeanAttributeInfo attr = getStandardAttributeInfo(metadata, attribute, false);
		if (attr != null)
		{
			String prefix = attr.isIs() ? "is" : "get";
			try
			{
				return invokeImpl(metadata, prefix + attribute, EMPTY_PARAMS, EMPTY_ARGS);
			}
			catch (IllegalArgumentException x)
			{
				// Never thrown, since there are no arguments
				throw new ImplementationException();
			}
		}
		else
		{
			throw new AttributeNotFoundException(attribute);
		}
	}

	public void setAttribute(MBeanMetaData metadata, Attribute attribute) throws MBeanException, AttributeNotFoundException, InvalidAttributeValueException, ReflectionException
	{
		String name = attribute.getName();
		MBeanAttributeInfo attr = getStandardAttributeInfo(metadata, name, true);
		if (attr != null)
		{
			try
			{
				invokeImpl(metadata, "set" + name, new String[] {attr.getType()}, new Object[] {attribute.getValue()});
			}
			catch (IllegalArgumentException x)
			{
				throw new InvalidAttributeValueException("Invalid value for attribute " + attribute.getName() + ": " + attribute.getValue());
			}
		}
		else
		{
			throw new AttributeNotFoundException(name);
		}
	}

	protected Object invokeImpl(MBeanMetaData metadata, String method, String[] signature, Object[] args)  throws ReflectionException, MBeanException, IllegalArgumentException
	{
		Method m = getStandardManagementMethod(metadata, method, signature);

		try
		{
			return m.invoke(metadata.mbean, args);
		}
		catch (IllegalAccessException x) {throw new ReflectionException(x);}
		catch (InvocationTargetException x)
		{
			Throwable t = x.getTargetException();
			if (t instanceof Error) {throw new RuntimeErrorException((Error)t);}
			else if (t instanceof JMRuntimeException) {throw (JMRuntimeException)t;}
			else if (t instanceof RuntimeException) {throw new RuntimeMBeanException((RuntimeException)t);}
			else {throw new MBeanException((Exception)t);}
		}
	}

	private MBeanAttributeInfo getStandardAttributeInfo(MBeanMetaData metadata, String attribute, boolean isWritable)
	{
		// Hit the cache
		MBeanAttributeInfo attr = (MBeanAttributeInfo)m_attributeCache.get(attribute);
		if (attr != null)
		{
			// Found in cache
			if (isWritable)
			{
				if (attr.isWritable()) {return attr;}
			}
			else
			{
				if (attr.isReadable()) {return attr;}
			}
		}
		else
		{
			// Not in cache, find it

			MBeanInfo info = metadata.info;

			MBeanAttributeInfo[] attrs = info.getAttributes();
			if (attrs != null)
			{
				for (int i = 0; i < attrs.length; ++i)
				{
					attr = attrs[i];
					String name = attr.getName();
					if (attribute.equals(name))
					{
						// Cache it
						m_attributeCache.put(attribute, attr);

						if (isWritable)
						{
							if (attr.isWritable()) {return attr;}
						}
						else
						{
							if (attr.isReadable()) {return attr;}
						}
					}
				}
			}
		}
		return null;
	}

	private MBeanOperationInfo getStandardOperationInfo(MBeanMetaData metadata, String method, String[] signature)
	{
		// Hit the cache
		MBeanOperationInfo oper = (MBeanOperationInfo)m_operationCache.get(method, signature);
		if (oper != null)
		{
			// Found in cache
			return oper;
		}
		else
		{
			// Not in cache, find it

			MBeanInfo info = metadata.info;

			MBeanOperationInfo[] opers = info.getOperations();
			if (opers != null)
			{
				for (int i = 0; i < opers.length; ++i)
				{
					oper = opers[i];
					String name = oper.getName();
					if (method.equals(name))
					{
						// Same method name, check number of parameters
						MBeanParameterInfo[] params = oper.getSignature();
						if (signature.length == params.length)
						{
							boolean match = true;
							for (int j = 0; j < params.length; ++j)
							{
								MBeanParameterInfo param = params[j];
								if (!signature[j].equals(param.getType()))
								{
									match = false;
									break;
								}
							}
							if (match)
							{
								m_operationCache.put(method, signature, oper);
								return oper;
							}
						}
					}
				}
			}
			return null;
		}
	}

	private Method getStandardManagementMethod(MBeanMetaData metadata, String name, String[] signature) throws ReflectionException
	{
		// Hit the cache
		Method method = (Method)m_methodCache.get(name, signature);

		if (method != null)
		{
			return method;
		}
		else
		{
			// Not in cache, find it
			try
			{
				Class[] params = Utils.loadClasses(metadata.classloader, signature);
				method = metadata.mbean.getClass().getMethod(name, params);
				m_methodCache.put(name, signature, method);
				return method;
			}
			catch (ClassNotFoundException x) {throw new ReflectionException(x);}
			catch (NoSuchMethodException x) {throw new ReflectionException(x);}
		}
	}
}
