/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server.interceptor;

import java.security.PrivilegedAction;
import java.security.AccessController;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistrationException;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ReflectionException;

import mx4j.server.MBeanMetaData;


/**
 * This interceptor sets the context class loader to the proper value for incoming calls.
 * It saves the current context class loader, set the context class loader to be the MBean's class loader for
 * the current call, and on return re-set the context class loader to the previous value
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.5 $
 */
public class ContextClassLoaderMBeanServerInterceptor extends DefaultMBeanServerInterceptor
{
	private PrivilegedAction m_get = new GetContextClassLoaderAction();

	public String getType()
	{
		return "contextclassloader";
	}

	public void addNotificationListener(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback)
	{
		if (isEnabled())
		{
			ClassLoader cl = getContextClassLoader();
			try
			{
				setContextClassLoader(metadata.classloader);
				super.addNotificationListener(metadata, listener, filter, handback);
			}
			finally {setContextClassLoader(cl);}
		}
		else
		{
			super.addNotificationListener(metadata, listener, filter, handback);
		}
	}

	public void removeNotificationListener(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback) throws ListenerNotFoundException
	{
		if (isEnabled())
		{
			ClassLoader cl = getContextClassLoader();
			try
			{
				setContextClassLoader(metadata.classloader);
				super.removeNotificationListener(metadata, listener, filter, handback);
			}
			finally {setContextClassLoader(cl);}
		}
		else
		{
			super.removeNotificationListener(metadata, listener, filter, handback);
		}
	}

	public void instantiate(MBeanMetaData metadata, String className, String[] params, Object[] args) throws ReflectionException, MBeanException
	{
		if (isEnabled())
		{
			ClassLoader cl = getContextClassLoader();
			try
			{
				setContextClassLoader(metadata.classloader);
				super.instantiate(metadata, className, params, args);
			}
			finally {setContextClassLoader(cl);}
		}
		else
		{
			super.instantiate(metadata, className, params, args);
		}
	}

	public void registration(MBeanMetaData metadata, int operation) throws MBeanRegistrationException
	{
		if (isEnabled())
		{
			ClassLoader cl = getContextClassLoader();
			try
			{
				setContextClassLoader(metadata.classloader);
				super.registration(metadata, operation);
			}
			finally {setContextClassLoader(cl);}
		}
		else
		{
			super.registration(metadata, operation);
		}
	}

	public MBeanInfo getMBeanInfo(MBeanMetaData metadata)
	{
		if (isEnabled())
		{
			ClassLoader cl = getContextClassLoader();
			try
			{
				setContextClassLoader(metadata.classloader);
				return super.getMBeanInfo(metadata);
			}
			finally {setContextClassLoader(cl);}
		}
		else
		{
			return super.getMBeanInfo(metadata);
		}
	}

	public Object invoke(MBeanMetaData metadata, String method, String[] params, Object[] args) throws MBeanException, ReflectionException
	{
		if (isEnabled())
		{
			ClassLoader cl = getContextClassLoader();
			try
			{
				setContextClassLoader(metadata.classloader);
				return super.invoke(metadata, method, params, args);
			}
			finally {setContextClassLoader(cl);}
		}
		else
		{
			return super.invoke(metadata, method, params, args);
		}
	}

	public AttributeList getAttributes(MBeanMetaData metadata, String[] attributes)
	{
		if (isEnabled())
		{
			ClassLoader cl = getContextClassLoader();
			try
			{
				setContextClassLoader(metadata.classloader);
				return super.getAttributes(metadata, attributes);
			}
			finally {setContextClassLoader(cl);}
		}
		else
		{
			return super.getAttributes(metadata, attributes);
		}
	}

	public AttributeList setAttributes(MBeanMetaData metadata, AttributeList attributes)
	{
		if (isEnabled())
		{
			ClassLoader cl = getContextClassLoader();
			try
			{
				setContextClassLoader(metadata.classloader);
				return super.setAttributes(metadata, attributes);
			}
			finally {setContextClassLoader(cl);}
		}
		else
		{
			return super.setAttributes(metadata, attributes);
		}
	}

	public Object getAttribute(MBeanMetaData metadata, String attribute) throws MBeanException, AttributeNotFoundException, ReflectionException
	{
		if (isEnabled())
		{
			ClassLoader cl = getContextClassLoader();
			try
			{
				setContextClassLoader(metadata.classloader);
				return super.getAttribute(metadata, attribute);
			}
			finally {setContextClassLoader(cl);}
		}
		else
		{
			return super.getAttribute(metadata, attribute);
		}
	}

	public void setAttribute(MBeanMetaData metadata, Attribute attribute) throws MBeanException, AttributeNotFoundException, InvalidAttributeValueException, ReflectionException
	{
		if (isEnabled())
		{
			ClassLoader cl = getContextClassLoader();
			try
			{
				setContextClassLoader(metadata.classloader);
				super.setAttribute(metadata, attribute);
			}
			finally {setContextClassLoader(cl);}
		}
		else
		{
			super.setAttribute(metadata, attribute);
		}
	}

	private ClassLoader getContextClassLoader()
	{
		return (ClassLoader)AccessController.doPrivileged(m_get);
	}

	private void setContextClassLoader(final ClassLoader cl)
	{
		AccessController.doPrivileged(new PrivilegedAction()
		{
			public Object run()
			{
				Thread.currentThread().setContextClassLoader(cl);
				return null;
			}
		});
	}

	private static class GetContextClassLoaderAction implements PrivilegedAction
	{
		public Object run()
		{
			return Thread.currentThread().getContextClassLoader();
		}
	}
}
