/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server.interceptor;

import java.util.List;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistrationException;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ReflectionException;

import mx4j.ImplementationException;
import mx4j.log.Log;
import mx4j.log.Logger;
import mx4j.server.MBeanMetaData;

/**
 * Base class for MBeanServer interceptors
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.6 $
 * @jmx:mbean description="MBeanServer interceptor"
 */
public abstract class DefaultMBeanServerInterceptor implements MBeanServerInterceptor, DefaultMBeanServerInterceptorMBean
{
	private boolean m_enabled = true;
	private String m_logCategory;
	private ThreadLocal m_chain = new ThreadLocal();

	protected DefaultMBeanServerInterceptor()
	{
		// It's amazing how setting up here this string dramatically reduces the times to get the Logger instance
		m_logCategory = getClass().getName() + "." + getType();
	}

	/**
	 * Returns whether this interceptor is enabled
	 * @see #setEnabled
	 * @jmx:managed-attribute description="The enable status of this interceptor"
	 */
	public boolean isEnabled()
	{
		return m_enabled;
	}

	/**
	 * Enables or disables this interceptor
	 * @see #isEnabled
	 * @jmx:managed-attribute
	 */
	public void setEnabled(boolean enabled)
	{
		m_enabled = enabled;
	}

	/**
	 * Returns the type of this interceptor
	 * @jmx:managed-attribute description="The type of this interceptor"
	 */
	public abstract String getType();

	protected MBeanServerInterceptor getNext()
	{
		List chain = (List)m_chain.get();
		int index = chain.indexOf(this);
		if (index < 0) throw new ImplementationException();
		MBeanServerInterceptor next = (MBeanServerInterceptor)chain.get(index + 1);
		next.setChain(chain);
		return next;
	}

	public void setChain(List chain)
	{
		// Put the chain in a thread local variable
		m_chain.set(chain);
	}

	protected Logger getLogger()
	{
		return Log.getLogger(m_logCategory);
	}

	public void addNotificationListener(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback)
	{
		getNext().addNotificationListener(metadata, listener, filter, handback);
	}

	public void removeNotificationListener(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback) throws ListenerNotFoundException
	{
		getNext().removeNotificationListener(metadata, listener, filter, handback);
	}

	public void instantiate(MBeanMetaData metadata, String className, String[] params, Object[] args) throws ReflectionException, MBeanException
	{
		getNext().instantiate(metadata, className, params, args);
	}

	public void registration(MBeanMetaData metadata, int operation) throws MBeanRegistrationException
	{
		getNext().registration(metadata, operation);
	}

	public MBeanInfo getMBeanInfo(MBeanMetaData metadata)
	{
		return getNext().getMBeanInfo(metadata);
	}

	public Object invoke(MBeanMetaData metadata, String method, String[] params, Object[] args) throws MBeanException, ReflectionException
	{
		return getNext().invoke(metadata, method, params, args);
	}

	public AttributeList getAttributes(MBeanMetaData metadata, String[] attributes)
	{
		return getNext().getAttributes(metadata, attributes);
	}

	public AttributeList setAttributes(MBeanMetaData metadata, AttributeList attributes)
	{
		return getNext().setAttributes(metadata, attributes);
	}

	public Object getAttribute(MBeanMetaData metadata, String attribute) throws MBeanException, AttributeNotFoundException, ReflectionException
	{
		return getNext().getAttribute(metadata, attribute);
	}

	public void setAttribute(MBeanMetaData metadata, Attribute attribute) throws MBeanException, AttributeNotFoundException, InvalidAttributeValueException, ReflectionException
	{
		getNext().setAttribute(metadata, attribute);
	}
}
