/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server.interceptor;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.InvalidAttributeValueException;
import javax.management.JMRuntimeException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistration;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.NotificationBroadcaster;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import javax.management.RuntimeErrorException;
import javax.management.RuntimeMBeanException;

import mx4j.ImplementationException;
import mx4j.log.Logger;
import mx4j.server.MBeanMetaData;
import mx4j.util.Utils;

/**
 * The last interceptor in the chain. <p>
 * It calls the MBean instance; if the MBean is a dynamic MBean, the call is direct, otherwise the call is delegated
 * to an {@link mx4j.server.MBeanInvoker MBeanInvoker}.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.8 $
 * @jmx:mbean description="The interceptor that invokes on the MBean instance"
 */
public class InvokerMBeanServerInterceptor extends DefaultMBeanServerInterceptor implements InvokerMBeanServerInterceptorMBean
{
	private MBeanServer m_server;

	public InvokerMBeanServerInterceptor(MBeanServer server)
	{
		m_server = server;
	}

	/**
	 * Returns the type of this interceptor
	 * @jmx:managed-attribute description="The type of this interceptor"
	 */
	public String getType()
	{
		return "invoker";
	}

	/**
	 * Returns whether this interceptor is enabled
	 * @jmx:managed-attribute description="This interceptor is always enabled and cannot be disabled"
	 */
	public boolean isEnabled()
	{
		return true;
	}

	public void addNotificationListener(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback)
	{
		((NotificationBroadcaster)metadata.mbean).addNotificationListener(listener, filter, handback);
	}

	public void removeNotificationListener(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback)
			throws ListenerNotFoundException
	{
		// PENDING: for JMX 1.2 this has to be changed
//		((NotificationBroadcaster)metadata.mbean).removeNotificationListener(listener, filter, handback);
		((NotificationBroadcaster)metadata.mbean).removeNotificationListener(listener);
	}

	public void instantiate(MBeanMetaData metadata, String className, String[] params, Object[] args) throws ReflectionException, MBeanException
	{
		try
		{
			Class cls = metadata.classloader.loadClass(className);

			Class[] signature = Utils.loadClasses(metadata.classloader, params);

			Constructor ctor = cls.getConstructor(signature);

			metadata.mbean = ctor.newInstance(args);
		}
		catch (ClassNotFoundException x)
		{
			throw new ReflectionException(x);
		}
		catch (NoSuchMethodException x)
		{
			throw new ReflectionException(x);
		}
		catch (InstantiationException x)
		{
			throw new ReflectionException(x);
		}
		catch (IllegalAccessException x)
		{
			throw new ReflectionException(x);
		}
		catch (IllegalArgumentException x)
		{
			throw new ReflectionException(x);
		}
		catch (InvocationTargetException x)
		{
			Throwable t = x.getTargetException();
			if (t instanceof Error)
			{
				throw new RuntimeErrorException((Error)t);
			}
			else if (t instanceof RuntimeException)
			{
				throw new RuntimeMBeanException((RuntimeException)t);
			}
			else
			{
				throw new MBeanException((Exception)t);
			}
		}
	}

	public void registration(MBeanMetaData metadata, int operation) throws MBeanRegistrationException
	{
		if (!(metadata.mbean instanceof MBeanRegistration)) return;

		MBeanRegistration registrable = (MBeanRegistration)metadata.mbean;

		try
		{
			switch (operation)
			{
				case PRE_REGISTER:
					ObjectName objName = registrable.preRegister(m_server, metadata.name);
					if (objName != null) metadata.name = objName;
					break;
				case POST_REGISTER_TRUE:
					registrable.postRegister(Boolean.TRUE);
					break;
				case POST_REGISTER_FALSE:
					registrable.postRegister(Boolean.FALSE);
					break;
				case PRE_DEREGISTER:
					registrable.preDeregister();
					break;
				case POST_DEREGISTER:
					registrable.postDeregister();
					break;
				default:
					throw new ImplementationException();
			}
		}
		catch (Exception x)
		{
			if (x instanceof MBeanRegistrationException)
			{
				throw (MBeanRegistrationException)x;
			}
			throw new MBeanRegistrationException(x);
		}
	}

	public MBeanInfo getMBeanInfo(MBeanMetaData metadata)
	{
		if (metadata.dynamic)
		{
			// From JMX 1.1 the MBeanInfo may be dynamically changed at every time, let's refresh it
			MBeanInfo info = ((DynamicMBean)metadata.mbean).getMBeanInfo();
			if (info == null) return null;
			metadata.info = info;

			// Refresh also ObjectInstance.getClassName(), if it's the case
			String className = info.getClassName();
			if (!metadata.instance.getClassName().equals(className))
			{
				metadata.instance = new ObjectInstance(metadata.name, className);
			}
		}

		return (MBeanInfo)metadata.info.clone();
	}

	public Object invoke(MBeanMetaData metadata, String method, String[] params, Object[] args) throws MBeanException, ReflectionException
	{
		Logger logger = getLogger();

		if (metadata.dynamic)
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Call to invoke() on dynamic MBean " + metadata.name);

			try
			{
				return ((DynamicMBean)metadata.mbean).invoke(method, args, params);
			}
			catch (JMRuntimeException x)
			{
				throw x;
			}
			catch (RuntimeException x)
			{
				throw new RuntimeMBeanException(x);
			}
			catch (Error x)
			{
				throw new RuntimeErrorException(x);
			}
		}
		else
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Call to invoke() on standard MBean " + metadata.name);

			return metadata.invoker.invoke(metadata, method, params, args);
		}
	}

	public Object getAttribute(MBeanMetaData metadata, String attribute) throws MBeanException, AttributeNotFoundException, ReflectionException
	{
		Logger logger = getLogger();

		if (metadata.dynamic)
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Call to getAttribute() on dynamic MBean " + metadata.name);

			try
			{
				return ((DynamicMBean)metadata.mbean).getAttribute(attribute);
			}
			catch (JMRuntimeException x)
			{
				throw x;
			}
			catch (RuntimeException x)
			{
				throw new RuntimeMBeanException(x);
			}
			catch (Error x)
			{
				throw new RuntimeErrorException(x);
			}
		}
		else
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Call to getAttribute() on standard MBean " + metadata.name);

			return metadata.invoker.getAttribute(metadata, attribute);
		}
	}

	public void setAttribute(MBeanMetaData metadata, Attribute attribute) throws MBeanException, AttributeNotFoundException, InvalidAttributeValueException, ReflectionException
	{
		Logger logger = getLogger();

		if (metadata.dynamic)
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Call to setAttribute() on dynamic MBean " + metadata.name);

			try
			{
				((DynamicMBean)metadata.mbean).setAttribute(attribute);
			}
			catch (JMRuntimeException x)
			{
				throw x;
			}
			catch (RuntimeException x)
			{
				throw new RuntimeMBeanException(x);
			}
			catch (Error x)
			{
				throw new RuntimeErrorException(x);
			}
		}
		else
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Call to setAttribute() on standard MBean " + metadata.name);

			metadata.invoker.setAttribute(metadata, attribute);
		}
	}

	public AttributeList getAttributes(MBeanMetaData metadata, String[] attributes)
	{
		Logger logger = getLogger();

		if (metadata.dynamic)
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Call to getAttributes() on dynamic MBean " + metadata.name);

			try
			{
				return ((DynamicMBean)metadata.mbean).getAttributes(attributes);
			}
			catch (JMRuntimeException x)
			{
				throw x;
			}
			catch (RuntimeException x)
			{
				throw new RuntimeMBeanException(x);
			}
			catch (Error x)
			{
				throw new RuntimeErrorException(x);
			}
		}
		else
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Call to getAttributes() on standard MBean " + metadata.name);

			AttributeList list = new AttributeList();
			for (int i = 0; i < attributes.length; ++i)
			{
				String name = attributes[i];
				try
				{
					Object value = getAttribute(metadata, name);
					Attribute attr = new Attribute(name, value);
					list.add(attr);
				}
				catch (Exception ignored)
				{
					if (logger.isEnabledFor(Logger.INFO)) logger.info("Exception caught from getAttributes(), ignoring attribute " + name);
				}
			}
			return list;
		}
	}

	public AttributeList setAttributes(MBeanMetaData metadata, AttributeList attributes)
	{
		Logger logger = getLogger();

		if (metadata.dynamic)
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Call to setAttributes() on dynamic MBean " + metadata.name);

			try
			{
				return ((DynamicMBean)metadata.mbean).setAttributes(attributes);
			}
			catch (JMRuntimeException x)
			{
				throw x;
			}
			catch (RuntimeException x)
			{
				throw new RuntimeMBeanException(x);
			}
			catch (Error x)
			{
				throw new RuntimeErrorException(x);
			}
		}
		else
		{
			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("Call to setAttributes() on standard MBean " + metadata.name);

			AttributeList list = new AttributeList();
			for (int i = 0; i < attributes.size(); ++i)
			{
				Attribute attr = (Attribute)attributes.get(i);
				try
				{
					setAttribute(metadata, attr);
					list.add(attr);
				}
				catch (Exception ignored)
				{
					if (logger.isEnabledFor(Logger.INFO)) logger.info("Exception caught from setAttributes(), ignoring attribute " + attr, ignored);
				}
			}
			return list;
		}
	}
}
