/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server.interceptor;

import java.util.ArrayList;

import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.ObjectName;

import mx4j.ImplementationException;

/**
 * MBean that configures the MBeanServer --> MBean interceptor chain.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.5 $
 * @jmx:mbean description="Configurator for MBeanServer to MBean interceptors"
 */
public class MBeanServerInterceptorConfigurator implements MBeanServerInterceptorConfiguratorMBean
{
	public static final String OBJECT_NAME = "JMImplementation:type=MBeanServerInterceptorConfigurator";

	private MBeanServer m_server;
	private ArrayList m_preInterceptors = new ArrayList();
	private ArrayList m_postInterceptors = new ArrayList();
	private ArrayList m_clientInterceptors = new ArrayList();
	private boolean m_running;

	/**
	 * Creates an instance of this configurator, for the given MBeanServer
	 * @jmx:managed-constructor description="Creates a new instance of MBeanServer to MBean interceptor configurator"
	 * @jmx:managed-constructor-parameter name="server" position="0" description="The MBeanServer that uses this configurator"
	 */
	public MBeanServerInterceptorConfigurator(MBeanServer server)
	{
		m_server = server;
	}

	/**
	 * Appends the given interceptor, provided by the client, to the existing interceptor chain.
	 * @see #registerInterceptor
	 * @jmx:managed-operation description="Appends an interceptor to the interceptor chain"
	 * @jmx:managed-operation-parameter name="interceptor" position="0" description="The interceptor to be appended to the interceptor chain"
	 */
	public void addInterceptor(MBeanServerInterceptor interceptor)
	{
		synchronized (m_clientInterceptors)
		{
			m_clientInterceptors.add(interceptor);
		}
	}

	/**
	 * Appends the given interceptor, provided by the client, to the existing interceptor chain and registers it as MBean.
	 * @see #addInterceptor
	 * @jmx:managed-operation description="Appends an MBean interceptor to the interceptor chain and registers it"
	 * @jmx:managed-operation-parameter name="interceptor" position="0" description="The interceptor to be appended to the interceptor chain"
	 * @jmx:managed-operation-parameter name="name" position="1" description="The ObjectName under which register the interceptor"
	 */
	public void registerInterceptor(MBeanServerInterceptor interceptor, ObjectName name) throws MBeanException
	{
		// First, try register this interceptor. The call will use the old interceptor chain
		try
		{
			m_server.registerMBean(interceptor, name);
		}
		catch (Exception x)
		{
			throw new MBeanException(x, "Could not register interceptor with name " + name);
		}

		addInterceptor(interceptor);
	}

	/**
	 * Removes all the interceptors added via {@link #addInterceptor(MBeanServerInterceptor interceptor)}.
	 * @see #addInterceptor
	 * @jmx:managed-operation description="Removes all the interceptors added via addInterceptor(MBeanServerInterceptor interceptor)"
	 */
	public void clearInterceptors()
	{
		synchronized (m_clientInterceptors)
		{
			m_clientInterceptors.clear();
		}
	}

	/**
	 * Adds the given interceptor at the beginning of the interceptor chain, before the custom interceptors that may be added
	 * via {@link #addInterceptor}.
	 * This method is called by the MBeanServer during initialization, to configure the interceptors needed to work properly.
	 */
	public void addPreInterceptor(MBeanServerInterceptor interceptor)
	{
		if (!isRunning())
		{
			m_preInterceptors.add(interceptor);
		}
		else
		{
			throw new ImplementationException();
		}
	}

	/**
	 * Adds the given interceptor at the end of the interceptor chain, after the custom interceptors that may be added
	 * via {@link #addInterceptor}.
	 * This method is called by the MBeanServer during initialization, to configure the interceptors needed to work properly.
	 */
	public void addPostInterceptor(MBeanServerInterceptor interceptor)
	{
		if (!isRunning())
		{
			m_postInterceptors.add(interceptor);
		}
		else
		{
			throw new ImplementationException();
		}
	}

	/**
	 * Returns the head interceptor of the interceptor chain.
	 * The head interceptor is always present.
	 */
	public MBeanServerInterceptor getHeadInterceptor()
	{
		if (isRunning())
		{
			// Create every time the interceptor chain, as it may change.
			int size = m_clientInterceptors.size();
			ArrayList chain = new ArrayList(m_preInterceptors.size() + size + m_postInterceptors.size());
			chain.addAll(m_preInterceptors);
			if (size > 0)
			{
				synchronized (m_clientInterceptors)
				{
					chain.addAll(m_clientInterceptors);
				}
			}
			chain.addAll(m_postInterceptors);

			// Set the chain on the first interceptor
			MBeanServerInterceptor first = (MBeanServerInterceptor)chain.get(0);
			first.setChain(chain);

			return first;
		}
		else
		{
			return null;
		}
	}

	/**
	 * Starts this configurator, so that the MBeanServer is now able to accept incoming calls.
	 * @see #stop
	 * @see #isRunning
	 * @jmx:managed-operation description="Starts the configurator so that the MBeanServer can accept incoming calls"
	 */
	public void start()
	{
		if (!isRunning())
		{
			m_running = true;
		}
	}

	/**
	 * Stops this configurator, so that the MBeanServer is not able to accept incoming calls.
	 * @see #start
	 * @jmx:managed-operation description="Stops the configurator so that the MBeanServer cannot accept incoming calls"
	 */
	public void stop()
	{
		if (isRunning())
		{
			m_running = false;
		}
	}

	/**
	 * Returns whether this configurator is running and thus if the MBeanServer can accept incoming calls
	 * @see #start
	 * @jmx:managed-attribute description="The running status of the configurator"
	 */
	public boolean isRunning()
	{
		return m_running;
	}
}
