/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server.interceptor;

import javax.management.ListenerNotFoundException;
import javax.management.Notification;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;

import mx4j.server.MBeanMetaData;

/**
 * Interceptor that takes care of replacing the source of Notifications to the
 * ObjectName of the NotificationBroadcaster that emitted it.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.4 $
 */
public class NotificationListenerMBeanServerInterceptor extends DefaultMBeanServerInterceptor
{
	private static class ListenerWrapper implements NotificationListener
	{
		private NotificationListener m_listener;
		private ObjectName m_objectName;

		private ListenerWrapper(NotificationListener listener, ObjectName name)
		{
			m_listener = listener;
			m_objectName = name;
		}

		public void handleNotification(Notification notification, Object handback)
		{
			// The JMX spec does not specify how to change the source to be the ObjectName
			// of the broadcaster. If we serialize the calls to the listeners, then it's
			// possible to change the source and restore it back to the old value before
			// calling the next listener; but if we want to support concurrent calls
			// to the listeners, this is not possible. Here I chose to support concurrent
			// calls so I change the value once and I never restore it.

			// Change the source to be the ObjectName of the notification broadcaster
			notification.setSource(m_objectName);
			// Notify the real listener
			NotificationListener listener = getTargetListener();
			listener.handleNotification(notification, handback);
		}

		private NotificationListener getTargetListener()
		{
			return m_listener;
		}

		public int hashCode()
		{
			return getTargetListener().hashCode();
		}

		public boolean equals(Object obj)
		{
			if (obj == null) return false;
			if (obj == this) return true;

			try
			{
				ListenerWrapper other = (ListenerWrapper)obj;
				return getTargetListener().equals(other.getTargetListener());
			}
			catch (ClassCastException ignored)
			{
			}
			return false;
		}
	}

	public String getType()
	{
		return "notificationlistener";
	}

	public void addNotificationListener(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback)
	{
		if (isEnabled())
		{
			ListenerWrapper wrapper = new ListenerWrapper(listener, metadata.name);
			super.addNotificationListener(metadata, wrapper, filter, handback);
		}
		else
		{
			super.addNotificationListener(metadata, listener, filter, handback);
		}
	}

	public void removeNotificationListener(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback) throws ListenerNotFoundException
	{
		if (isEnabled())
		{
			ListenerWrapper wrapper = new ListenerWrapper(listener, metadata.name);
			super.removeNotificationListener(metadata, wrapper, filter, handback);
		}
		else
		{
			super.removeNotificationListener(metadata, listener, filter, handback);
		}
	}
}
