/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.tools.naming;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;

import mx4j.log.Log;
import mx4j.log.Logger;

/**
 * An MBean that wraps tnameserv. <p>
 * Calling {@link #start} will start tnameserv in a separate process via
 * {@link java.lang.Runtime#exec(String) Runtime.exec(String command)}.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.6 $
 *
 * @jmx:mbean description="MBean that wraps tnameserv"
 */
public class CosNamingService implements CosNamingServiceMBean
{
	private int m_port;
	private boolean m_running;
	private Process m_process;
	private InputStreamConsumer m_output;
	private InputStreamConsumer m_error;
	private long m_delay;

	/**
	 * Creates a new instance of CosNamingService with the default port (900).
	 * @jmx:managed-constructor description="Creates a new instance of CosNamingService with the default port (900)"
	 */
	public CosNamingService()
	{
		this(900);
	}

	/**
	 * Creates a new instance of CosNamingService with the specified port.
	 * @jmx:managed-constructor description="Creates a new instance of CosNamingService with the specified port"
	 * @jmx:managed-constructor-parameter name="port" position="0" description="The port on which tnameserv will listen for incoming connections"
	 */
	public CosNamingService(int port)
	{
		m_port = port;
	}

	/**
	 * Sets the port on which tnameserv listens for incoming connections.
	 * @see #getPort
	 * @jmx:managed-attribute
	 */
	public void setPort(int port)
	{
		m_port = port;
	}

	/**
	 * Returns the port on which tnameserv listens for incoming connections
	 * @see #setPort
	 * @jmx:managed-attribute description="The port on which tnameserv listens for incoming connections"
	 */
	public int getPort()
	{
		return m_port;
	}

	/**
	 * Returns whether this MBean has been started and not yet stopped.
	 * @see #start
	 * @jmx:managed-attribute description="The running status of this MBean"
	 */
	public boolean isRunning()
	{
		return m_running;
	}

	/**
	 * Sets the delay (in milliseconds) that the {@link #start} and {@link #stop} methods waits before returning.
	 * This is needed to allow the tnameserv process to really start and stop.
	 * @see #getDelay
	 * @jmx:managed-attribute
	 */
	public void setDelay(long delay)
	{
		m_delay = delay;
	}

	/**
	 * Returns the delay (in milliseconds) that the {@link #start} and {@link #stop} methods waits before returning.
	 * @see #setDelay
	 * @jmx:managed-attribute description="The delay (ms) for the start() and stop() methods"
	 */
	public long getDelay()
	{
		return m_delay;
	}

	/**
	 * Starts this MBean: tnameserv can now accept incoming calls
	 * @see #stop
	 * @see #isRunning
	 * @jmx:managed-operation description="Starts tnameserv"
	 */
	public void start()
	{
		if (!isRunning())
		{
			final Logger logger = getLogger();

			final Object lock = new Object();
			Thread thread = new Thread(new Runnable()
			{
				public void run()
				{
					try
					{
						// TODO: use JAVA_HOME ?
						m_process = Runtime.getRuntime().exec("tnameserv -ORBInitialPort " + getPort());
						if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Process created: " + m_process);

						m_output = new InputStreamConsumer(m_process.getInputStream());
						m_error = new InputStreamConsumer(m_process.getErrorStream());
						m_output.start();
						m_error.start();

						Thread.sleep(getDelay());

						synchronized (lock)
						{
							lock.notifyAll();
						}
						int result = m_process.waitFor();
						if (logger.isEnabledFor(Logger.DEBUG)) logger.debug("Exit value is: " + result);

						if (isRunning()) throw new IOException("Unexpected exception (maybe the port " + getPort() + " is already in use)");
					}
					catch (IOException x)
					{
						logger.error("Could not create process", x);
					}
					catch (InterruptedException x)
					{
						if (logger.isEnabledFor(Logger.INFO)) logger.info("Interrupted", x);
					}
				}
			}, "CosNamingService Thread");

			thread.start();

			try
			{
				synchronized (lock)
				{
					lock.wait();
				}
			}
			catch (InterruptedException ignored)
			{
			}

			if (logger.isEnabledFor(Logger.TRACE)) logger.trace("CosNamingService started");

			m_running = true;
		}
	}

	/**
	 * Stops this MBean: tnameserv cannot accept anymore incoming calls
	 * @see #start
	 * @jmx:managed-operation description="Stops tnameserv"
	 */
	public void stop()
	{
		if (isRunning())
		{
			m_running = false;
			m_process.destroy();
			m_output.interrupt();
			m_error.interrupt();

			try
			{
				Thread.sleep(m_delay);
			}
			catch (InterruptedException x)
			{
			}
		}
	}

	private Logger getLogger()
	{
		return Log.getLogger(getClass().getName());
	}

	private class InputStreamConsumer extends Thread
	{
		private InputStream m_stream;
		private byte[] m_buffer = new byte[512];

		public InputStreamConsumer(InputStream stream)
		{
			super("Stream Consumer Thread");
			m_stream = new BufferedInputStream(stream);
			setDaemon(true);
		}

		public void run()
		{
			Logger logger = getLogger();
			while (!isInterrupted())
			{
				try
				{
					int read = -1;
					while ((read = m_stream.read(m_buffer)) >= 0)
					{
						if (logger.isEnabledFor(Logger.INFO)) logger.info(new String(m_buffer, 0, read));
					}
				}
				catch (InterruptedIOException x)
				{
					break;
				}
				catch (IOException x)
				{
					logger.info("Error while consuming process stream", x);
				}
			}
		}
	}
}
