/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.util;

import java.util.ArrayList;
import java.util.Collections;

/**
 * A queue that executes TimeTasks. <p>
 * Every scheduled task is executed at the specified time; if this time is in the past,
 * the task is executed immediately.
 * Just before the task is executed, the task itself is updated so that its new
 * execution time is set or it is marked as finished (if it is not forever periodic).
 *
 * @see TimeTask
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.4 $
 */
public class TimeQueue
{
	private Thread m_thread;
	private boolean m_started;
	private ArrayList m_tasks;

   /**
    * Creates a new TimeQueue
    */
	public TimeQueue()
	{
		m_tasks = new ArrayList();
	}

   /**
    * Starts this TimeQueue. <br>
    * Tasks is executed only after the queue is started.
    * @see #stop
    */
	public void start()
	{
		if (!m_started)
		{
			m_thread = new Thread(new Loop(), "MBean Timer Notification Thread");
			m_thread.setDaemon(true);
			m_thread.start();
			m_started = true;
		}
	}

   /**
    * Stops this TimeQueue. <br>
    * No task is executed when the queue is stopped; however, already scheduled tasks
    * are not removed; restarting the queue has the effect of executing the tasks remained
    * if their time has come.
    * @see #start
    */
  public void stop()
	{
		if (m_started)
		{
			m_thread.interrupt();
			try {m_thread.join();}
			catch (InterruptedException ignored) {}
			m_thread = null;
			m_started = false;
		}
  }

   /**
    * Returns the number of tasks present in this TimeQueue
    */
	public int size() {return m_tasks.size();}

   /**
    * Schedules the given task for execution.
    * @see #unschedule
    */
	public void schedule(TimeTask task)
	{
		synchronized (this)
		{
			m_tasks.add(task);
			// Using tree sets or maps does not work, since they don't rely on equals for add, but on compareTo
			Collections.sort(m_tasks);
			notifyAll();
		}
	}

   /**
    * Removes the given task from this TimeQueue
    * @see #schedule
    */
	public void unschedule(TimeTask task)
	{
		synchronized (this)
		{
			m_tasks.remove(task);
		}
	}

   /**
    * Removes all the tasks from this TimeQueue.
    */
	public void clear()
	{
		synchronized (this)
		{
			m_tasks.clear();
		}
	}

	private TimeTask getTask() throws InterruptedException
	{
		synchronized (this)
		{
			while (m_tasks.isEmpty())
			{
				wait();
			}

			// Do not remove the task from the queue
			TimeTask task = (TimeTask)m_tasks.get(0);
			return task;
		}
	}

	private class Loop implements Runnable
	{
		public void run()
		{
			while (!m_thread.isInterrupted())
			{
				try
				{
					TimeTask task = getTask();
               long now = System.currentTimeMillis();
					long executionTime = task.getNextExecutionTime();
					long timeToWait = executionTime - now;
					boolean runTask = timeToWait <= 0;
					if (!runTask)
					{
						// When a new job is scheduled, I wake up, but this job may not be the one to run
						Object lock = TimeQueue.this;
						synchronized (lock)
						{
							// timeToWait is always strictly > 0, so I don't wait forever
							lock.wait(timeToWait);
						}
					}
					else
					{
						// The task must be run, remove it from the list
						unschedule(task);

						if (task.isPeriodic())
						{
							// Compute the new execution time
							task.setNextExecutionTime(executionTime + task.getPeriod());

							// Reschedule on the new time
							schedule(task);
						}
						else
						{
							// Mark the task as finished, it will be removed also by the Timer class
							task.setFinished(true);
						}

						try
						{
							// Run it !
							task.run();
						}
						catch (Exception x)
						{
							// RuntimeException caught, for now just print it
							x.printStackTrace();
						}
					}
				}
				catch (InterruptedException x)
				{
					break;
				}
			}
		}
	}
}
